//---------------------------------------------------------------------------
// DataHandler.cpp
//
// Serial traffic handler
//
// Noptel/FN Expert (Field), NOS Pro
// C. Noptel Oy, Oulu Finland, 2009-2016
//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

//  Device object; receive frames and blows; upper class in hierarchy
#include "Device.h"
// this class
#include "DataHandler.h"
// Serial handler class; handles connection with serial port; transfers bytes from/to this class
#include "SerialHandler.h"
// Frame types
#include "FramesAndSets.h"

// constructor/destructor;
TDataHandler::TDataHandler(TDevice& ADevice)
{
	memset(this, 0, sizeof(*this));

	// link device
    FDevice = &ADevice;

	// init receive buffer to 4k; can receive a whole blow
    FDataBufferSize = 4096;
	FDataBuffer = new char[FDataBufferSize];

	FDataReceivedBytes = 0;

	FStat_BytesReceived = 0;
	FStat_BytesSkipped = 0;
	FStat_BytesOverflowed = 0;
	FStat_ChecksumErrors = 0;

	// create serial handler and link it to "this" instance
	FSerialHandler = new TSerialHandler(*this);

	FPortNumber = 0;
}

TDataHandler::~TDataHandler()
{
	// close port before deleting object; no data is accidentally received during deinitialization
	FSerialHandler->CloseCOMPort();
	delete FSerialHandler;
	delete [] FDataBuffer;
}

// open serial port
void TDataHandler::OpenComPort(int APort)
{
	FSerialHandler->OpenCOMPort(APort);
	FPortNumber = APort;
	// we may be waiting for things after port is opened so we try to catch the first byte
	FStartupByteReceived = false;
}

void TDataHandler::CloseComPort()
{
	FPortNumber = 0;
	FSerialHandler->CloseCOMPort();
}

// recreate port after "catastrophic" shutdown (disconnect)
void TDataHandler::RecreatePort()
{
	int LPort = FPortNumber;
	FSerialHandler->CreateSerialPort();
	FSerialHandler->OpenCOMPort(LPort);
	FSerialHandler->SetBaudRate(115200);
	FSerialHandler->SetDTRRTS(false, false);
}

// initialize device; read parameters sets and status
bool TDataHandler::Initialize()
{
	// check if com port set; otherwise return false
	if (!FSerialHandler->IsPortOpen()) return false;

	ResetStats();
	//FStartupByteReceived = false;

    // do we handle downloading sets or is it the upper class?

	return true;
}

unsigned char TDataHandler::GetChecksum(void* APtrToStart, int ALengthWithoutCheckByte)
{
	int LSum = 0;
    unsigned char* LPtr = (unsigned char*)APtrToStart;
    for (int i=0;i < ALengthWithoutCheckByte; i++)
    {
    	LSum += LPtr[i];
    }

    // return low 8 bits of sum
    return LSum & 0xff;
}

void TDataHandler::SendRaw(void* APtrToStart, int ALength)
{
 	// write to serial port
	FSerialHandler->Write(APtrToStart, ALength);
	memcpy(LLastSentCommand, APtrToStart, ALength);

}

void TDataHandler::SendShowMessage(WideString AMessage, char AControlData, int AFirstRow)
{
	TCommandFrame_Field_ShowMessage LCommand;
	memset(&LCommand, 0, sizeof(LCommand) );
	LCommand.COMMANDBYTE = CCMD_FieldShowMessage;

	LCommand.CONTROL_DATA = AControlData;

	TStringList* LList = new TStringList();
	LList->Text = AMessage;

	int LMax = LList->Count < 6 ? LList->Count : 6;
	// start message from mid row if first row explicitly given
	int LStartRow = (AFirstRow == -1) ? 3 - LMax / 2 : AFirstRow;

	for (int i=0; i < LMax; i++)
	{
		// this is on purpose ansistring, i.e. force one-byte-per-character
		AnsiString LStr = LList->Strings[i];

		memset(LCommand.CHARACTER, 0, sizeof(LCommand.CHARACTER));
		int Len = LStr.Length() < 10 ? LStr.Length() : 10;

		memcpy(LCommand.CHARACTER, LStr.c_str(), Len);

		// clear screen when first one is sent
		LCommand.ClearBufferFirst = (i == 0);
		// .. but show all only when last row is sent
		LCommand.ShowMessageMode = (i != LMax - 1) ? 0 : 3;

		//LCommand.TextBlinking = i & 1; // blinking is not supported by Expert
		//LCommand.TextInversed = i & 2;

		LCommand.LINE_NUMBER = LStartRow + i;
		LCommand.CHECKBYTE = GetChecksum(&LCommand, sizeof(LCommand) - 1);

		FSerialHandler->Write(&LCommand, sizeof(LCommand));
		memcpy(LLastSentCommand, &LCommand, sizeof(LCommand));

		// TODO wait for ackn, if needed. However, this will do as well
		Sleep(200);
	}

	delete LList;

}

void TDataHandler::SendFieldGetParameterFrame(int AParameterSet)
{
	TCommandFrame_Field_GetParameterSet LCommand;
	LCommand.COMMANDBYTE = CCMD_FieldGetParameterSet;
	LCommand.PARAMETERSETNUMBERBYTE = (unsigned char)AParameterSet;
	// calculate checksum
	LCommand.CHECKBYTE = GetChecksum(&LCommand, sizeof(LCommand) - 1);
	// write to serial port
	FSerialHandler->Write(&LCommand, sizeof(LCommand));
	memcpy(LLastSentCommand, &LCommand, sizeof(LCommand));
}

void TDataHandler::SendFieldGetDeviceInfo(int ADeviceInfoPage)
{
    // we use the same struct at this time (identical)
    TCommandFrame_Field_GetParameterSet LCommand;
	LCommand.COMMANDBYTE = CCMD_FieldGetDeviceInfo;
    LCommand.PARAMETERSETNUMBERBYTE = (unsigned char)ADeviceInfoPage;
    // calculate checksum
    LCommand.CHECKBYTE = GetChecksum(&LCommand, sizeof(LCommand) - 1);
    // write to serial port
	FSerialHandler->Write(&LCommand, sizeof(LCommand));
	memcpy(LLastSentCommand, &LCommand, sizeof(LCommand));
}

void TDataHandler::SendFieldGetShot(int ABlock, int AIndex)
{
    // build command frame
	TCommandFrame_Field_GetShot LCommand;
	LCommand.COMMANDBYTE = CCMD_FieldGetShot;
    LCommand.MEMORY_PAGE_NUMBER = ABlock;
    LCommand.SHOT_INDEX = AIndex;
    // calculate checksum
    LCommand.CHECKBYTE = GetChecksum(&LCommand, sizeof(LCommand) - 1);
    // write to serial port
	FSerialHandler->Write(&LCommand, sizeof(LCommand));
	memcpy(LLastSentCommand, &LCommand, sizeof(LCommand));
}

void TDataHandler::SendFieldGetSilhouetteTargetPiece(int ATarget, bool ABitmaskFrame, int ABlockNumber)
{
	// build command frame
	TCommandFrame_Field_DownloadSilhouetteBitmap LCommand;
	memset(&LCommand, 0, sizeof(LCommand));
	LCommand.COMMANDBYTE = CCMD_DownloadSilhouette_Bitmap;
	LCommand.SilhouetteNumber = ATarget;
	LCommand.BitmaskFrame = ABitmaskFrame;
	LCommand.BlockNumber = ABlockNumber;
	// calculate checksum
	LCommand.CHECKBYTE = GetChecksum(&LCommand, sizeof(LCommand) - 1);
	// write to serial port
	FSerialHandler->Write(&LCommand, sizeof(LCommand));
	memcpy(LLastSentCommand, &LCommand, sizeof(LCommand));
}

void TDataHandler::SendFieldGetRingTarget(int ATarget, int ABlock)
{
	// build command frame
	TCommandFrame_Field_GetRingTarget LCommand;
	LCommand.COMMANDBYTE = CCMD_FieldGetRingTarget;
	LCommand.RING_TARGET_NUMBER = ATarget;
	LCommand.RING_TARGET_FRAME = ABlock;
	// calculate checksum
	LCommand.CHECKBYTE = GetChecksum(&LCommand, sizeof(LCommand) - 1);
	// write to serial port
	FSerialHandler->Write(&LCommand, sizeof(LCommand));
	memcpy(LLastSentCommand, &LCommand, sizeof(LCommand));
}

void TDataHandler::SendFieldSetRingTarget(int ATarget, int ABlock, void* ABytes, bool AUpdateCheckBit)
{
    // build command frame
    TCommandFrame_Field_SetRingTarget LCommand;
	LCommand.COMMANDBYTE = CCMD_FieldSetRingTarget;
    LCommand.RING_TARGET_NUMBER = ATarget;
    LCommand.TargetBlock = ABlock;
    LCommand.UpdateCheckBit = AUpdateCheckBit;
    memcpy(LCommand.SETBYTE, ABytes, sizeof(LCommand.SETBYTE));
    // calculate checksum
    LCommand.CHECKBYTE = GetChecksum(&LCommand, sizeof(LCommand) - 1);
    // write to serial port
	FSerialHandler->Write(&LCommand, sizeof(LCommand));
	memcpy(LLastSentCommand, &LCommand, sizeof(LCommand));
}

void TDataHandler::SendFieldEraseBlock(int ABlock)
{
    // build command frame
    TCommandFrame_Field_GetShot LCommand;
    LCommand.COMMANDBYTE = CCMD_FieldGetShot;
    LCommand.MEMORY_PAGE_NUMBER = ABlock;
    // sending 0xff as shot index erases block
    LCommand.SHOT_INDEX = 0xff;
    // calculate checksum
    LCommand.CHECKBYTE = GetChecksum(&LCommand, sizeof(LCommand) - 1);
    // write to serial port
	FSerialHandler->Write(&LCommand, sizeof(LCommand));
	memcpy(LLastSentCommand, &LCommand, sizeof(LCommand));
}

void TDataHandler::SendFieldSetParameterFrame(int AParameterSet, char* APtr, bool ASetCheck)
{
    // set calculate checksum bit
    TCommandFrame_Field_SetParameterSet LCommand;
	LCommand.COMMANDBYTE = CCMD_FieldSetParameterSet;
    LCommand.ParameterSetNumber = (unsigned char)AParameterSet;
    LCommand.UpdateCheckBit = ASetCheck;
    memcpy(LCommand.SETBYTE, APtr, sizeof(LCommand.SETBYTE));
    // calculate checksum
    LCommand.CHECKBYTE = GetChecksum(&LCommand, sizeof(LCommand) - 1);
    // write to serial port
	FSerialHandler->Write(&LCommand, sizeof(LCommand));
	memcpy(LLastSentCommand, &LCommand, sizeof(LCommand));
}

void TDataHandler::SendStartProgramming()
{
	SendFieldSetFunctionMode(15, 0, 0x20);
}

void TDataHandler::SendFieldSetFunctionMode(char AMode, char AReserved2, char AReserved3)
{
    // create command frame
	TCommandFrame_Field_SetFunctionMode LCommand;
	LCommand.COMMANDBYTE = CCMD_FieldEnterFunctionMode;
    LCommand.OPTICALUNITSTATE = AMode;
    LCommand.RESERVED2 = AReserved2;
    LCommand.RESERVED3 = AReserved3;
	// calculate checksum
	LCommand.CHECKBYTE = GetChecksum(&LCommand, sizeof(LCommand) - 1);
	// write to serial port
	FSerialHandler->Write(&LCommand, sizeof(LCommand));
	memcpy(LLastSentCommand, &LCommand, sizeof(LCommand));
}

void TDataHandler::SendStartDC()
{

}

void TDataHandler::SendStop()
{
}

void TDataHandler::SetPortBaudRate(int ABaud, bool DTR, bool RTS)
{
    if (FSerialHandler)
    {
        FSerialHandler->SetBaudRate(ABaud);        
        FSerialHandler->SetDTRRTS(DTR,RTS);
    }
}

void TDataHandler::BreakOnNextReceive()
{
    FBreakOnNextReceive = true;
}

// buffer bytes and seek for frames. send frames forward
void TDataHandler::ReceiveIncomingBytes(char* APtrToData, int AByteCount)
{
	FLastLatency = FSerialHandler->FLastLatency;

	// put startup byte received up when first 0x2d is received
	if (FStartupByteReceived == false && APtrToData[0] == 0x2d)
	{
		FStartupByteReceived = true;
		//Beep();
		HandleReceivedFrame(APtrToData, KFID_StartupByte);
	}

	// check that we have enough space to store bytes
	if (FDataBufferSize-FDataReceivedBytes <= AByteCount)
    {
        FStat_BytesOverflowed += AByteCount;
        return;
    }

    FStat_BytesReceived += AByteCount;
    FStat_LargestReceivedChunk = AByteCount > FStat_LargestReceivedChunk ? AByteCount : FStat_LargestReceivedChunk;

    // store data to back end of receive buffer
    memcpy(FDataBuffer+FDataReceivedBytes, APtrToData, AByteCount);
	FDataReceivedBytes += AByteCount;

	// we can "force" a break point to check response for a certain command using this
	if (FBreakOnNextReceive)
	{
		int i=0; i++; // Insert break point here
		FBreakOnNextReceive = false;
	}

	// we handle the received bytes in flight; this could be done 'asyncronously' if some reason requires to do so.
    // ** then we should change the data buffer to ring buffer for thread safety reasons

    int LBytesUsed = SearchForFramesInDataBuffer();

    // if bytes spent (i.e. that can be discarded now)...
    if (LBytesUsed != 0)
    {
		// move the unspent data to front of buffer and reduce bytes to be later handled
    	memcpy(FDataBuffer, FDataBuffer + LBytesUsed, FDataReceivedBytes - LBytesUsed);
        FDataReceivedBytes -= LBytesUsed; 
    }
}

int TDataHandler::SearchForFramesInDataBuffer()
{
	// go through received buffer until got
    // a) <2 bytes in buffer -> no frame identifier can be found
    // b) frame identifier found but not enough bytes

    // how many bytes to scroll the receive buffer
	int LBytesUsed = 0;
    // how many bytes still left to check
    int LBytesAvailable = FDataReceivedBytes;
    // where are we going in receive buffer
    char* LCurrentPtr = FDataBuffer;

    // at least 2 bytes to even bother check
    bool GoOn = (LBytesAvailable >= 2);

    // do until need to stop for some reason
    while (GoOn)
    {
		EFRAMEID_FrameTypes LRetFrameTypeDetected;
        int LRetFrameLength = 0;
    	bool IsOk = CheckForFrame(LCurrentPtr, LBytesAvailable, LRetFrameTypeDetected, LRetFrameLength);

        // detected frame, look for next
		if (IsOk)
        {
        	HandleReceivedFrame(LCurrentPtr, LRetFrameTypeDetected);
			FStat_FramesReceived++;

            // log on terminal if selected
			//if (FormTerminal->CheckBoxReceive->Checked)
			//    FormTerminal->AddBytesToMemo(LCurrentPtr, LRetFrameLength, false, true);

        	// move ahead the length of detected frame; increase used count and reduce available count
        	LCurrentPtr += LRetFrameLength;
            LBytesUsed += LRetFrameLength;
            LBytesAvailable -= LRetFrameLength;
        }
        else
        {
        	// frame found but not enough bytes... go out and wait for more
        	if (LRetFrameLength >= LBytesAvailable)
            {
            	// just get out
            	GoOn = false;
            }
            // otherwise no frame was detected; actually a data stream error
            else
            {
                // log on terminal if selected
				//if (FormTerminal->CheckBoxReceive->Checked)
                //    FormTerminal->AddBytesToMemo(LCurrentPtr, 1, false, false);

            	// jump to next byte
	        	LCurrentPtr ++;
    	        LBytesUsed ++;
        	    LBytesAvailable --;

				FStat_BytesSkipped++;
            }
        }

        // if we were left with less than 2 bytes, get out
        if (LBytesAvailable < 2)
        {
        	GoOn = false;
        }
    } // outest loop

    return LBytesUsed;
}

bool TDataHandler::CheckForFrame(char* AFramePtr, int ABytesAvailable, EFRAMEID_FrameTypes& AFrameTypeDetected, int& AFrameLength)
{
	// default: no frame detected
	AFrameTypeDetected = KFID_NotIdentified;
    AFrameLength = 0;

	// too little bytes to contain event frame identifier
	if (ABytesAvailable<2) return false;

    unsigned char LByte0 = AFramePtr[0];
    unsigned char LByte1 = AFramePtr[1];

    // frames start with either 0x59 or 0xCA; other start are not supported
	if ( LByte0 == CFRAMEID_AckFrame )
    {
		// Ack Frame = 0x59; the frame is a result to an request or command

        // There are four CA XX frames; parameter, info, diagnostics and 'ack' frame
        // SyncByte 2 is known only for first three so we must check if first if we are receiving
        // any those; otherwise we are dealing with an ack frame.

		// frame is some ack frame if none other identified;
		AFrameTypeDetected = KFID_AckFrame;
        AFrameLength = sizeof(TResponseFrame_Field_AckFrame);

		if (LByte1 == CFRAMEID_Field_DownloadSilhouette_1)
        {
            AFrameTypeDetected = KFID_SilhouetteBitmapFrame;
            AFrameLength = sizeof(TResponseFrame_DownloadSilhouetteBitmapFrame);
        }
        // check if received a parameter frame;
		else if (LByte1 == CFRAMEID_Field_ParameterSet_1)
        {
            AFrameTypeDetected = KFID_FieldParameterFrame;
            AFrameLength = sizeof(TResponseFrame_Field_ParameterSetFrame);
        }
        // check if received a device info frame;
		else if (LByte1 == CFRAMEID_Field_DeviceInfo_1)
        {
            AFrameTypeDetected = KFID_FieldDeviceInfoFrame;
            AFrameLength = sizeof(TResponseFrame_Field_DeviceInfoFrame);

            // shot count frame is received as a device info frame.
            // third byte is 0x20. special treatment
            if ( (ABytesAvailable > 2) && (AFramePtr[2] == 0x20) )
			{
				AFrameTypeDetected = KFID_FieldShotCountFrame;
				AFrameLength = sizeof(TResponseFrame_Field_ShotCountFrame);
			}
		}
		else if (LByte1 == CFRAMEID_Field_GetRingTarget_1)
		{
			AFrameTypeDetected = KFID_RingTargetFrame;
			AFrameLength = sizeof(TResponseFrame_Field_RingTargetFrame);
		}
	}
	else if (LByte0 == CFRAMEID_SampleFrameField_0)
	{
		if (LByte1 == CFRAMEID_SampleFrameField_1)
		{
			AFrameTypeDetected = KFID_SampleFrameField;
			AFrameLength = sizeof(TResponseFrame_FieldSampleFrame);
		}
		if (LByte1 == CFRAMEID_ResultFrameField_1)
		{
			AFrameTypeDetected = KFID_ResultFrameField;
			AFrameLength = sizeof(TResponseFrame_FieldResultFrame);
		}
		else if (LByte1 == CFRAMEID_InfoErrorFrame_1)
		{
            AFrameTypeDetected = KFID_InfoErrorFrame;
            AFrameLength = sizeof(TResponseFrame_Field_InfoErrorFrame);
        }
    }

	// check avail length and checksum
	bool isOk = false;
    if ( (AFrameLength <= ABytesAvailable) && (AFrameTypeDetected != KFID_NotIdentified) )
    {
    	// enough bytes, check checksum
        isOk = CheckChecksum(AFramePtr, AFrameLength);

        if (!isOk)
        {
        	FStat_ChecksumErrors++;

            // if checksum is wrong the frame looker moves to next byte;
            //  so we don't need to drop frame length to 1 byte (which could be done)
        }
    }

    return isOk;
}

bool TDataHandler::CheckChecksum(char* AFirstBytePtr, int AFrameByteLengthWithChecksum)
{
    // if no frame length
    if (AFrameByteLengthWithChecksum == 0)
    {
        throw new Exception("CheckChecksum: Invalid frame length");
    }
	// sum of bytes
	unsigned char AChecksum = 0;
    // current byte
    unsigned char* ACurPtr = AFirstBytePtr;
    // checksum byte, first not to include in checksum
    unsigned char* AChkPtr = AFirstBytePtr + AFrameByteLengthWithChecksum - 1;
	// sum until pointers meet
    while (ACurPtr != AChkPtr)
    	AChecksum += *(ACurPtr++);

    bool ACheckOk = (AChecksum == *AChkPtr);

	/*if (!ACheckOk)
		int idebug=0;*/

	return ACheckOk;
}

// handle the detected (and checksummed) frame
void TDataHandler::HandleReceivedFrame(char* AFramePtr, EFRAMEID_FrameTypes AFrameTypeDetected)
{
	// store received frame in the 'last received struct' in  device.
	// at the moment no other handling is required for anything else but sample frame at this level
	FDevice->StoreReceivedFrame( (void*)AFramePtr, AFrameTypeDetected);
}

// return pointer to read log first entry to allow easy reading etc
TLogRow* TDataHandler::GetReadLogStart()
{
	return FSerialHandler->FReadLogEntryList;
}

// return pointer to write log first entry to allow easy reading etc
TLogRow* TDataHandler::GetWriteLogStart()
{
	return FSerialHandler->FWriteLogEntryList;
}

// enable or disable logging
void TDataHandler::SetLogEnabled(bool AEnabled)
{
	FSerialHandler->SetLogEnabled(AEnabled);
}
