//---------------------------------------------------------------------------

#ifndef DataHandlerH
#define DataHandlerH
//---------------------------------------------------------------------------

#include "FramesAndSets.h"

// forward declarations; don't need to include the headers
class TSerialHandler;
class TDevice;

struct TLogRow;

//---------------------------------------------------------------------------
class TDataHandler
{

// PKy 2017. Getter to get serial port data logs. Maybe this is nice enough.
public:
	TLogRow* GetReadLogStart();
	TLogRow* GetWriteLogStart();
    void SetLogEnabled(bool AEnabled);

private:
	TSerialHandler* FSerialHandler;
	TDevice* FDevice;

    // byte array for incoming data, updated by com port events. Frames packed 'in flight'
    char* FDataBuffer;
    int FDataBufferSize; // how big the buffer is
    int FDataReceivedBytes; // how many bytes are currently buffered

public:
    // constructor/destructor;
    TDataHandler(TDevice& ADevice);
    ~TDataHandler();

	// ** COM Related pass-through functions

    unsigned char LLastSentCommand[20];

	// create and open serial port
    void OpenComPort(int APort);
    void CloseComPort();
    void SetPortBaudRate(int ABaud, bool DTR, bool RTS);
    // 0 = closed
    int FPortNumber;

	// recreate port after "catastrophic" shutdown (disconnect)
    void RecreatePort();

	// device sends 0x2D
	bool FStartupByteReceived;

    int FStat_BytesReceived;
    int FStat_FramesReceived;
    int FStat_BytesSkipped;
    int FStat_ChecksumErrors;
    int FStat_BytesOverflowed;
    int FStat_LargestReceivedChunk;
    void ResetStats() { FStat_BytesReceived = FStat_FramesReceived = FStat_BytesSkipped = FStat_ChecksumErrors = FStat_BytesOverflowed = 0; }

	bool FBreakOnNextReceive;
	void BreakOnNextReceive();

    // ** Device state handling

    // initialize device; read parameters sets and status
    bool Initialize();

	// calculate and return checksum
	unsigned char GetChecksum(void* APtrToStart, int ALengthWithoutCheckByte);

    // st2000 tester functions
	void SendRaw(void* APtrToStart, int ALength);

    // commands, field
    void SendFieldGetParameterFrame(int AParameterSet);
    void SendFieldGetDeviceInfo(int ADeviceInfoPage);
    void SendFieldSetParameterFrame(int AParameterSet, char* APtr, bool ASetCheck = false);
    void SendFieldSetFunctionMode(char AMode, char AReserved2 = 0, char AReserved3 = 0);
    void SendFieldGetShot(int ABlock, int AIndex);
    void SendFieldEraseBlock(int ABlock);
    void SendFieldGetRingTarget(int ATarget, int ABlock);
    void SendFieldSetRingTarget(int ATarget, int ABlock, void* ABytes, bool AUpdateCheckBit = false);

	void SendFieldGetSilhouetteTargetPiece(int ATarget, bool ABitmaskFrame, int ABlockNumber);

	void SendFieldSetFunctionMode(char AMode, char AReserved2 = 0);
	void SendStop();
	void SendStartProgramming();
	void SendStartDC();

    void SendShowMessage(WideString AMessage, char AControlData = 0, int AFirstRow = -1);

	// ** Data buffer handling

	double FLastLatency;

	// buffer incoming bytes; and check for frames
	void ReceiveIncomingBytes(char* APtrToBuffer, int AByteCount);
	// check data buffer for received frames and transmit them to be handled
	// return bytes used from start of receive buffer
	int SearchForFramesInDataBuffer();
	// check for a frame at given address (and data length);
	// returns found frame type, frame length and true if ok to parse at given address.
	// returns false if no usable frame in address, and frame length != 0 if not enough bytes
	bool CheckForFrame(char* AFramePtr, int ABytesAvailable, EFRAMEID_FrameTypes& ARetFrameTypeDetected, int& ARetFrameLength);
	// check for check sum in a frame; check sum byte is always last byte in frame
	bool CheckChecksum(char* AFirstBytePtr, int AFrameByteLengthWithChecksum);

	// ** Received frame handling

	// handle the detected (and checksummed) frame
	void HandleReceivedFrame(char* AFramePtr, EFRAMEID_FrameTypes AFrameTypeDetected);

	// a sample frame received; handle according to selected state and check if it's a final sample
	//void HandleReceivedSampleFrame(TResponseFrame_CombinatorSampleFrame* AFramePtr);
	// deliver received sample to be handled by UI; called when in continuous mode (no buffering done)
	//void HandleReceivedContinuousSample(TResponseFrame_CombinatorSampleFrame* AFramePtr);
	// deliver blow to be handled by UI; called when data handler has a full blow buffered
	void HandleReceivedBlow();

	//
	//MHi 17.5.2016 obsolete frame removed
	//void HandleReceivedDiagnosticsFrame(TResponseFrame_DiagnosticsFrame* AFramePtr);
	//
	void HandleReceivedErrorFrame(TResponseFrame_ErrorFrame* AFramePtr);
	//
	void HandleReceivedAckFrame(TResponseFrame_Field_AckFrame* AFramePtr);


};
//---------------------------------------------------------------------------

#endif
 