//---------------------------------------------------------------------------
// Device.cpp
//
// Device object
//
// Noptel/FN Expert (Field), NOS Pro
// C. Noptel Oy, Oulu Finland, 2009-2017
//
// The name of the device during R&D was "Field", which is used in this source code.
// Marketing name of the product is "Expert", which is used in UI and User's Guide.
//---------------------------------------------------------------------------


#include <vcl.h>
#pragma hdrstop

// this class
#include "Device.h"

// serial traffic handler
#include "DataHandler.h"

// Frame types
#include "FramesAndSets.h"

// TShooter, to add a shot
#include "ShotData.h"


#include "Globals.h"

// arc tan&sin, calculating sector or tilt/rotation
#include "math.h"

// min max
#include <Math.hpp>

AnsiString GetStringAsHex(void* APtr, int ANumber)
{
    AnsiString LStr;
    unsigned char* LPtr = (unsigned char*)APtr;
    while (ANumber>0)
    {
        LStr.cat_printf("%02X", *LPtr);
        LPtr++;
		ANumber--;
    }
    return LStr;
}

// extern defined in FramesAndSets.h
const double ONE_NAU_to_mm_in_one_meter = 1.0 / 300;

// constructor/destructor
TDevice::TDevice()
{
	memset(this, 0, sizeof(TDevice));

    // construct data handler
	FDataHandler = new TDataHandler(*this);

	// to save shots in a file (for debug)
    FStringListToSave = new TStringList();

	// default for resetting aim buffer
    FFailureSamplesUntilReset = 6;
	FFailureSamplesNow = 0;

	// in realtime mode, how many sample to ignore after trigger
	FIgnoreSamplesAfterTrigger = 20;
	FIgnoreSamplesNow = 0;

	// allocate frame buffer; constructor clears frames
    FNextFrameIndex = 0;
    FFrameBufferSize = 200; // 10.6.2010. 3 seconds
	FFrames = new TResponseFrame_FieldSampleFrame[FFrameBufferSize];

	FDeviceTimer = new Extctrls::TTimer(NULL);
	FDeviceTimer->OnTimer = InitializeTimerEvent;
	FDeviceTimer->Enabled = false;

	// get this value only once
	QueryPerformanceFrequency(&FPerformanceTimerFreq);
    // reset last frames
    memset(&FLastFrames, 0, sizeof(FLastFrames));

	FShooter = new TShooter;
	// add shooter to session
	GSession->FShooters->Add(FShooter);
	FCOMPort = 0;
}

TDevice::~TDevice()
{
	// delete "working set"; destructor of set frees stored objects
	//delete FSetOfBlows;

	// close com port before deleting data handler; assume measuring and other ops are stopped when destructing
	FDataHandler->CloseComPort();
	delete FDataHandler;

	delete FStringListToSave;

	delete [] FFrames;

	// TODO Device should NOT own a shooter object, as device can "vanish" from computer while shooter remains
	// avoid memory leak
	if (FShooter)
	{
		//delete FShooter;
	}

	delete FDeviceTimer;
}

void TDevice::OpenComPort(int APort, bool DTRRTS)
{
	// ignore boot byte
	FCOMPort = APort;
	FAsyncInfo.FPortOpened = true;
	FDataHandler->OpenComPort(APort);

	if (DTRRTS) // default baud rate of Expert device is 115200 bps
	{
		FDataHandler->SetPortBaudRate(115200, DTRRTS, DTRRTS);
	}
}

void TDevice::CloseComPort()
{
	FCOMPort = 0;
	FDataHandler->CloseComPort();
}

bool TDevice::CheckAsyncTimeout()
{

	// TODO Add other asyncs here as well, if needed
	//if ( (GetTickCount() - FAsyncInfo.FLastAsyncProcessed > 30000) && FAsyncInfo.IsAsyncGoing() )
    // PKy 21.3.2017. Upped to 6000ms in case of delay in stopping.
	if ( (GetTickCount() - FAsyncInfo.FLastAsyncProcessed > 6000) && (FAsyncInfo.FUploadTargetsGoing && !FAsyncInfo.FUploadTargetsDone ) )
		return true;
	else
		return false;
}

void TDevice::StartAsyncInitialize()
{
	FAsyncInfo.FLastAsyncProcessed = GetTickCount();
	FInitializing = true;

	try
	{

		// clear received stuff, set baud rates and dtr/rts up
		memset(&FLastFrames, 0, sizeof(FLastFrames));
		FDataHandler->SetPortBaudRate(115200, true, true);

		// reset last receive frame
		FLastFrames.FLastReceivedFrameType = KFID_None;

		// set scan on, data handler then fired "initialize timer event" when startup byte is received
		FAsyncInfo.FScanGoing = true;
		FAsyncInfo.FScanIndex = 0;

		// first we ask for a device info frame.ProcessAsyncInitialize is fired on ack. If no ack is received, there is no device in port
		StopMeasuring();
	}
	catch (...)
	{
		// on exception... i.e. can't write to port
		// no device found, scanning is complete
		FInitializing = false;
		FInitDeviceFound = false;
		FAsyncInfo.FScanGoing = false;
		FAsyncInfo.FScanDone = true;

		FDeviceTimer->Enabled = false;

		// don't start timer if exception raised
		return;
	}

	// start wait timer, tag 0 means powered up
	FDeviceTimer->Tag = 0;
	FDeviceTimer->Interval = 3000;
	FDeviceTimer->Enabled = true;


}

void __fastcall TDevice::InitializeTimerEvent(TObject* Sender)
{

	// if tag is 0, this is the first event after RTS up. check if we got the startup bytes.
	if (FDeviceTimer->Tag < 3)
	{
		FDeviceTimer->Tag = FDeviceTimer->Tag + 1;

		// first we ask for a device info frame.ProcessAsyncInitialize is fired on ack. If no ack is received, there is no device in port
		try
		{
			StopMeasuring();
		}
		catch (...)
		{
            // on exception... i.e. can't write to port
			// no device found, scanning is complete
			FInitializing = false;
			FInitDeviceFound = false;
			FAsyncInfo.FScanGoing = false;
			FAsyncInfo.FScanDone = true;

			FDeviceTimer->Enabled = false;
		}

		// if not.. we update tag to 1 and send get device info
		if (FLastFrames.FLastReceivedFrameType == KFID_None)
		{
			// just 1 second to receive ack for stop measuring
			FDeviceTimer->Interval = 2000;
		}
		else if (FLastFrames.FLastReceivedFrameType == KFID_AckFrame || FLastFrames.FLastReceivedFrameType == KFID_StartupByte)
		{
			// we received have either waited enough, or received the startup byte now stuff goes on
			FDeviceTimer->Enabled = false;
		}
		// first we ask for a device info frame.ProcessAsyncInitialize is fired on ack. If no ack is received, there is no device in port
		//StopMeasuring();
	}
	else if (FDeviceTimer->Tag == 3)
	{
		// if still haven't received a frame, we don't have a device in port and scan is done.
		if (FLastFrames.FLastReceivedFrameType == KFID_None)
		{
			FDeviceTimer->Enabled = false;

			//CloseComPort();

			// no device found, scanning is complete
			FInitializing = false;
			FInitDeviceFound = false;
			FAsyncInfo.FScanGoing = false;
			FAsyncInfo.FScanDone = true;
		}
	}
}

void TDevice::ProcessAsyncInitialize()
{
	FAsyncInfo.FLastAsyncProcessed = GetTickCount();

	if (!FAsyncInfo.FScanGoing) throw new Exception("No async scan going");

	int LI = FAsyncInfo.FScanIndex++;

	FInitializing = true;
	//int LStartTickCount = GetTickCount();

	if (LI==0)
	{
		FDataHandler->SendFieldGetDeviceInfo(0);
		// disable "timer" after first received message
		FDeviceTimer->Enabled = false;
		//FInitDeviceFound = true;
	}


	if (LI==1) FDataHandler->SendFieldGetDeviceInfo(0x20);
	// request device info frames 0,2,3,4,5,8
	if (LI==2) FDataHandler->SendFieldGetDeviceInfo(0);
	if (LI==3) FDataHandler->SendFieldGetDeviceInfo(2);
	if (LI==4) FDataHandler->SendFieldGetDeviceInfo(3);
	if (LI==5) FDataHandler->SendFieldGetDeviceInfo(4);
	if (LI==6) FDataHandler->SendFieldGetDeviceInfo(5);
	// usage info
	if (LI==7) FDataHandler->SendFieldGetDeviceInfo(0x10);
	// error counters
	if (LI==8) FDataHandler->SendFieldGetDeviceInfo(0x19);

	// request parameter sets 1-4, i.e. blocks 0,1,2,3,8,9,16,17,18,19,56
	if (LI==9) FDataHandler->SendFieldGetParameterFrame(0);
	if (LI==10) FDataHandler->SendFieldGetParameterFrame(1);
	if (LI==11) FDataHandler->SendFieldGetParameterFrame(2);
	if (LI==12) FDataHandler->SendFieldGetParameterFrame(3);
	if (LI==13) FDataHandler->SendFieldGetParameterFrame(8);
	if (LI==14) FDataHandler->SendFieldGetParameterFrame(9);
	if (LI==15) FDataHandler->SendFieldGetParameterFrame(16);
	if (LI==16) FDataHandler->SendFieldGetParameterFrame(17);
	if (LI==17) FDataHandler->SendFieldGetParameterFrame(18);
	if (LI==18) FDataHandler->SendFieldGetParameterFrame(19);
	// check parameter checksum (|0x80)
	if (LI==19) FDataHandler->SendFieldGetParameterFrame(56 | 0x80);

	// download shot count page
	if (LI==20) FDataHandler->SendFieldGetDeviceInfo(0x20);

	if (LI==21)
	{
		FDeviceParameterCheckFailed = !FLastFrames.FLastFieldParameterSetFrame.UpdateCheckBit;

		int LCount = FLastFrames.FLastShotCountFrame.GetNumberOfShots();
		FDownloadableShotCount = LCount;

		// get 10 chars from the device info frame to make serial number
		FSerialNumber = AnsiString((char*)FLastFrames.FFieldParameterSet4.SERIAL_NUMBER, 10);
		FShooter->FSerialNumber = FSerialNumber;

		// get device type from serial number
		if (FSerialNumber.Pos("FLD") == 1)
		{
			FDeviceType = KTypeExpert;
		}
		else if (FSerialNumber.Pos("XMA") == 1)
		{
			FDeviceType = KTypeXMA;
		}

		FLastFrames.FFieldUsageInfoFrame;
		FLastFrames.FFieldErrorCounterFrame;

		FInitializing = false;
		FInitDeviceFound = true;
		FInitDeviceReady = true;

		FAsyncInfo.FScanDone = true;
		FAsyncInfo.FScanGoing = false;

		// copy the use scenario set to shooter
		FShooter->FUseScenario = FLastFrames.FFieldParameterSet1;

		// indicate that a conflict check can be made
        GScanCompletedAfterLastCheck = true;

		FDataHandler->FLastLatency;

		// proceed downloading shots...
		if (FAsyncInfo.FDownloadShotsAfterScan)
		{
			StartAsyncDownloadShots();
		}
	}
}

// we wish to process the target uploads "asyncronously" i.e. so that the main thread is not blocking
// and we can show a neat progress bar (TODO). this is especially true if we upload to many devices/targets
void TDevice::ProcessAsyncUploadTargets()
{
	FAsyncInfo.FLastAsyncProcessed = GetTickCount();
	FAsyncInfo.FProgressNow = FAsyncInfo.FProgressNow + 1;

	if (FAsyncInfo.FUploadTargetsGoing && !FAsyncInfo.FUploadTargetsDone)
	{
		// check if we need to enter programming mode
		if (!FAsyncInfo.InProgrammingMode && (FAsyncInfo.FUploadRingTarget |
			FAsyncInfo.FUploadSilhouetteTarget | FAsyncInfo.FUploadUserScenarioAndErase) )
		{
			FDataHandler->SendStartProgramming();
			FAsyncInfo.InProgrammingMode = true;
			// reset counter just in case
			FAsyncInfo.FUploadRingTargetBlock = 0;
			FAsyncInfo.FUploadSilhouetteTargetBlock = 0;
			FAsyncInfo.FUploadUserScenarioBlock = 0;
			return;
		}

		// ring, silhouette, user scenario each do one thing that is followed by an ack from device
		// after each ack, next command is issued.
		// for each "upload" there is a boolean that indicates whether it should be done/is going
		// each "upload" resets it's going after upload done (no command sent) and then next "upload" starts with a command
		// after every "upload" is done, and we are in programming mode, we return from programming mode, wait for ack and then upload is DONE.
		// TODO check for timeouts

		if (FAsyncInfo.FUploadRingTarget)
		{
			TRingTarget& LTarget = FAllTargets.GetRingTarget(FAsyncInfo.FUploadRingTargetIndex);
			switch (FAsyncInfo.FUploadRingTargetBlock)
			{
				case 0:
					FDataHandler->SendFieldSetRingTarget(FAsyncInfo.FUploadRingTargetIndex, 0, &LTarget.FRingSet1, false);
					break;
				case 1:
					FDataHandler->SendFieldSetRingTarget(FAsyncInfo.FUploadRingTargetIndex, 1, &LTarget.FRingSet2, false);
					break;
				case 2:
					FDataHandler->SendFieldSetRingTarget(FAsyncInfo.FUploadRingTargetIndex, 2, &LTarget.FRingSet3, false);
					break;
				case 3:
				    Sleep(100);
					FDataHandler->SendFieldSetRingTarget(FAsyncInfo.FUploadRingTargetIndex, 3, &LTarget.FTargetName, true);

					//FAsyncInfo.FUploadRingTarget = false;
					//Beep();
					// if we needed to upload multiple ring target at once, we would pick the next one here
					// .. and recursively call this function again to start it (and return).
					break;

					break;

				case 4:
					// now we are done with ring target, next command goes for silhouette target or user scenario, OR out of programming mode
					FAsyncInfo.FUploadRingTarget = false;
					//Beep();
					// if we needed to upload multiple ring target at once, we would pick the next one here
					// .. and recursively call this function again to start it (and return).
					break;

				default:
					throw new Exception("Illegate state async upload ring");
			}

			// so on ack we handle the next block
			FAsyncInfo.FUploadRingTargetBlock++;
		}
		// upload silhouette (when ring target upload not going)
		if (!FAsyncInfo.FUploadRingTarget && FAsyncInfo.FUploadSilhouetteTarget)
		{
			TSilhouetteTarget& LTarget = FAllTargets.GetSilhouetteTarget(FAsyncInfo.FUploadSilhouetteTargetIndex);
			TSilhouetteBitmapMatrix& LBitmap = FAllTargets.GetSilhouetteBitmapMatrix(FAsyncInfo.FUploadSilhouetteTargetIndex);

			// TODO refactor command send to datahandler
			TCommandFrame_Field_UploadSilhouetteBitmap LStruct;
			memset(&LStruct, 0, sizeof(LStruct));

			LStruct.COMMANDBYTE = CCMD_UploadSilhouette_Bitmap;
			LStruct.SilhouetteNumber = FAsyncInfo.FUploadSilhouetteTargetIndex;
			LStruct.BitmaskFrame = 0;

			switch (FAsyncInfo.FUploadSilhouetteTargetBlock)
			{
				// first we upload the matrix rows
				case 0:
					LStruct.SilhouetteTargetNameBlock = 0; //
					memcpy(LStruct.DATA, &LTarget.FSilhouetteSet, sizeof(LStruct.DATA));
					LStruct.CHECKBYTE = FDataHandler->GetChecksum(&LStruct, sizeof(LStruct)-1);
					FDataHandler->SendRaw(&LStruct, sizeof(LStruct));
					// so on ack we handle the next block.. this updated in a different manner than ring upload
					FAsyncInfo.FUploadSilhouetteTargetBlock++;
					break;
				case 1:
					LStruct.SilhouetteTargetNameBlock = 1;
					memcpy(LStruct.DATA, &LTarget.FTargetName, sizeof(LStruct.DATA));
					LStruct.CHECKBYTE = FDataHandler->GetChecksum(&LStruct, sizeof(LStruct)-1);
					FDataHandler->SendRaw(&LStruct, sizeof(LStruct));
					// so on ack we handle the next block
					FAsyncInfo.FUploadSilhouetteTargetBlock++;
					// reset row counter before going to upload rows
					FAsyncInfo.FUploadSilhouetteTargetRow = 0;
					break;
				case 2:
				{
					LStruct.BitmaskFrame = 1;
					int LRow = FAsyncInfo.FUploadSilhouetteTargetRow;
					FAsyncInfo.FUploadSilhouetteTargetRow++;
					// 4 pixels make a byte
					unsigned int LByteNumber = 0;
					unsigned char LVal/* = 0*/;
					for (int LCol = 0; LCol<64; LCol+=4)
					{
						// left most pixel to bits 6,7, 2nd left 4,5; 3rd 2,3; 4th 0,1. LVal is zerong on first line
						LVal = LBitmap.FPixelState [LCol][LRow] << 6;
						LVal |= LBitmap.FPixelState[LCol+1][LRow] << 4;
						LVal |= LBitmap.FPixelState[LCol+2][LRow] << 2;
						LVal |= LBitmap.FPixelState[LCol+3][LRow];
						LStruct.DATA[LByteNumber] = LVal;
						LByteNumber++;
					}

					// row = 'frame number'
					LStruct.SH1_FRAME_NUMBER = LRow;

					// update check bit for last frame
					if (LRow==63) LStruct.UpdateCheckBit = 1;
					LStruct.CHECKBYTE = FDataHandler->GetChecksum(&LStruct, sizeof(LStruct)-1);

					// TODO to use data handler to build command struct
					FDataHandler->SendRaw(&LStruct, sizeof(LStruct));

					// to next phase after last row
					if (LRow==63) FAsyncInfo.FUploadSilhouetteTargetBlock++;

					break;
				}
				case 3:
					// target is done uploaded
					FAsyncInfo.FUploadSilhouetteTarget = false;
					//Beep();
					// if we needed to upload more sil target at once, we would pick the next one here
					// .. and recursively call this function again to start it (and return).
					break;

				default:
					throw new Exception("Illegate state async upload silhouette");
			}
		}
		// ring/silhouette not going and have to update user scenario set
		if (FAsyncInfo.FUploadUserScenarioAndErase && !(FAsyncInfo.FUploadSilhouetteTarget | FAsyncInfo.FUploadRingTarget) )
		{
			int L = FAsyncInfo.FUploadUserScenarioBlock;
			if (L == 0) FDataHandler->SendStartProgramming();
			if (L == 1) FDataHandler->SendFieldSetParameterFrame(0,GetPointerToFieldParameterFrame(0));
			if (L == 2) FDataHandler->SendFieldSetParameterFrame(1,GetPointerToFieldParameterFrame(1));
			if (L == 3) FDataHandler->SendFieldSetParameterFrame(2,GetPointerToFieldParameterFrame(2));
			if (L == 4) FDataHandler->SendFieldEraseBlock(6);

			if (L == 5) FDataHandler->SendFieldEraseBlock(0);
			if (L == 6) FDataHandler->SendFieldEraseBlock(1);
			if (L == 7) FDataHandler->SendFieldEraseBlock(2);
			if (L == 8) FDataHandler->SendFieldEraseBlock(3);
			if (L == 9) FDataHandler->SendFieldEraseBlock(4);
			if (L == 10) FDataHandler->SendFieldEraseBlock(5);
			// update check bit =(true)
			if (L == 11) FDataHandler->SendFieldSetParameterFrame(3,GetPointerToFieldParameterFrame(3), true);

			if (L == 12)
			{
				// update shot count counters
				FDataHandler->SendFieldGetDeviceInfo(0x20);
				FAsyncInfo.FUploadUserScenarioAndErase = false;
				//Beep();
			}

			FAsyncInfo.FUploadUserScenarioBlock++;









		}
		// if all uploads are done, return from programming mode
		if ( !(FAsyncInfo.FUploadRingTarget | FAsyncInfo.FUploadSilhouetteTarget | FAsyncInfo.FUploadUserScenarioAndErase) )
		{
        	// no uploads going, but we are in programming mode -> send leave programming mode
			if (FAsyncInfo.InProgrammingMode)
			{
				FDataHandler->SendFieldSetFunctionMode(0x01, 0x00, 0x21);
				FAsyncInfo.InProgrammingMode = false;
			}
			else
			{
            	// now we are done uploading targets
            	FAsyncInfo.FUploadTargetsDone = true;
            }
		}
	}

	// update done bit

}

void TDevice::StartAsyncDownloadShots()
{
	if (FLastFrames.FLastShotCountFrame.GetNumberOfShots() == 0) return;

	FAsyncInfo.FDownloadShotsDone = false;
	FAsyncInfo.FDownloadShotsGoing = true;

	FAsyncInfo.FDownloadRequestedShots = 0;

	// blocks and shot-in-block start at index 1
	int i=0;

	// seek for first block that contains shots
	while (i < 7 && FLastFrames.FLastShotCountFrame.MemoryPages[i].BLOCK_SHOT_COUNT == 0)
	{
		i++;
    }

	FAsyncInfo.FDownloadShotIndexInBlock = 0;
	FAsyncInfo.FDownloadShotBlock = i;

	FAsyncInfo.FLastAsyncProcessed = GetTickCount();
	FDataHandler->SendFieldGetShot(FAsyncInfo.FDownloadShotBlock, FAsyncInfo.FDownloadShotIndexInBlock);
	FAsyncInfo.FDownloadRequestedShots++;
}

void TDevice::ProcessAsyncDownloadShots()
{
	// update timeout ticker
	FAsyncInfo.FLastAsyncProcessed = GetTickCount();

	if (!FAsyncInfo.FDownloadShotsGoing)
	{
		// no async in progress, we shouldnt be here.
		// dont throw exceptions as we are on another thread
		FAsyncInfo.FAsyncError = true;
		return;
	}

	if (FAsyncInfo.FAsyncError)
	{
		return;
	}

	int LBlock = FAsyncInfo.FDownloadShotBlock;
	int LShot  = FAsyncInfo.FDownloadShotIndexInBlock;

    FShooter->FShotList->Count;

	// if current block has more shots to be downloaded
	if (FLastFrames.FLastShotCountFrame.MemoryPages[LBlock].BLOCK_SHOT_COUNT > LShot+1)
	{
		FAsyncInfo.FDownloadShotIndexInBlock++;
		FDataHandler->SendFieldGetShot(FAsyncInfo.FDownloadShotBlock, FAsyncInfo.FDownloadShotIndexInBlock);
		FAsyncInfo.FDownloadRequestedShots++;
	}
	else
	{
		LBlock++;
		// seek for next block that contains shots
		while (FLastFrames.FLastShotCountFrame.MemoryPages[LBlock].BLOCK_SHOT_COUNT == 0)
		{
			LBlock++;
			// if we are out of blocks...
			if (LBlock > 6)
			{
				FAsyncInfo.FDownloadShotsGoing = true;
				FAsyncInfo.FDownloadTargetsGoing = true;

				// TODO get necessary targets & update error
				//FErrorDownloading = !DownloadRequiredSilhouettesAndTargets();
				//ConvertNAUToMMAfterDownload();

				// now download is complete
				FAsyncInfo.FDownloadShotsDone = true;
				//FAsyncInfo.FDownloadTargetsDone = true;

				return;
			}
		}

		FAsyncInfo.FDownloadShotBlock = LBlock;
		FAsyncInfo.FDownloadShotIndexInBlock = 0;
		FDataHandler->SendFieldGetShot(FAsyncInfo.FDownloadShotBlock, FAsyncInfo.FDownloadShotIndexInBlock);
		FAsyncInfo.FDownloadRequestedShots++;
	}

	// remove (comment) this to allow debug, if needed
	FAsyncInfo.FLastAsyncProcessed = GetTickCount();
}

// this is called from (any) thread to start the async process
void TDevice::StartAsyncDownloadTargets()
{
	// setup async download info
	FAsyncInfo.FDownloadTargetsGoing = true;
	FAsyncInfo.FDownloadTargetIndex = 0;
	FAsyncInfo.FDownloadTargetIndexFirst = 0;
	// 0..15 ring targets, 16..23 silhouette targets
	FAsyncInfo.FDownloadTargetIndexFinal = 24;
	FAsyncInfo.FDownloadTargetBlock = 0;
	// start downloading first block
	FDataHandler->SendFieldGetRingTarget(FAsyncInfo.FDownloadTargetIndex, FAsyncInfo.FDownloadTargetBlock);
}

// this is called by serial port thread when one block is received,
void TDevice::ProcessAsyncDownloadTargets()
{
	// 16.5.2016 MHi Code removed, because "async download targets" is no longer in use
	return;
}

// initialize device; read parameters sets and statusbool TDevice::Initialize()
bool TDevice::Initialize()
{
	// check com port open
	if (!FDataHandler->Initialize()) return false;

	FInitializing = true;
	//int LStartTickCount = GetTickCount();

	// we do do like this:
	// startup on 115200. wait 1500ms.
	// Send get device infos. timeout 1000ms
	// Send get parameter sets. timeout 1000ms
	// All must be ok, otherwise scan failed.

	FLastFrames.ResetReceivedHandled();
	memset(&FLastFrames, 0, sizeof(FLastFrames));

	// request all parameter sets from field device (no sleep on bluetooth)
	if (!FIsBluetooth) Sleep(1000);
	FDataHandler->SetPortBaudRate(115200, true, true);

	// we don't have an event handler to launch this when the startup byte is given so we just check every 1ms if we received it
	// don't wait on bluetooth devices
	bool LCont = true && !FIsBluetooth;
	int i; // we can measure how long it takes for the startup byte to arrive after power up
	for (i = 0; (i < 2000) && LCont; i++)
	{
		Sleep(1);
		// in practice we would settle for any byte
		LCont = !(FDataHandler->FStartupByteReceived || FDataHandler->FStat_BytesReceived);
	}

	// send stop measuring and wait for ack
	StopMeasuring();
	//if (WaitUntilResponseType(KFID_AckFrame, 4000) != KFID_AckFrame)
	if (WaitUntilResponseType(KFID_AckFrame, 2000) != KFID_AckFrame)
	{
		AddToLog("No stop ack");
		// if serial mode, skip at first fail
		if (!FIsBluetooth)
		{
			return false;
		}
		else
		{
			// retry open if bluetooth fails

			int LPort = FCOMPort;
			AddToLog("Retrying #1 BT COM"+AnsiString(LPort));
			//CloseComPort();
			//Sleep(3000);
			//OpenComPort(LPort, true);

			// send stop measuring and wait for ack
			AddToLog("Retrying #1 Send stop");
			StopMeasuring();
			AddToLog("Retrying #1 Wait ack");
			//if (WaitUntilResponseType(KFID_AckFrame, 4000) != KFID_AckFrame)
			if (WaitUntilResponseType(KFID_AckFrame, 2000) != KFID_AckFrame)
			{
				// retry open if bluetooth fails
				AddToLog("Retrying #2 BT COM"+AnsiString(LPort));
				//CloseComPort();
				//Sleep(3000);
				//OpenComPort(LPort, true);

				// send stop measuring and wait for ack
				AddToLog("Retrying #2 Send stop");
				StopMeasuring();
				AddToLog("Retrying #2 Wait ack");
				//if (WaitUntilResponseType(KFID_AckFrame, 4000) != KFID_AckFrame)
				//	return false;
				if (WaitUntilResponseType(KFID_AckFrame, 2000) != KFID_AckFrame)
					return false;

			}
			//return false;
		}
	}

	FDataHandler->FLastLatency;
	AddToLog("Stop ackd " + AnsiString(int(1000*FDataHandler->FLastLatency)) +"ms" );

	// this is "Get Shot Count" and is received as shot count frame
	FDataHandler->SendFieldGetDeviceInfo(0x20);
	if (WaitUntilResponseType(KFID_FieldShotCountFrame, 2000) != KFID_FieldShotCountFrame) return false;
	Application->ProcessMessages();
	// now we believe a device is present
	FInitDeviceFound = true;

	FDataHandler->FLastLatency;

	// request device info frames 0,2,3,4,5,8
	FDataHandler->SendFieldGetDeviceInfo(0);
	if (WaitUntilResponseType(KFID_FieldDeviceInfoFrame, 2000) != KFID_FieldDeviceInfoFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldGetDeviceInfo(2);
	if (WaitUntilResponseType(KFID_FieldDeviceInfoFrame, 2000) != KFID_FieldDeviceInfoFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldGetDeviceInfo(3);
	if (WaitUntilResponseType(KFID_FieldDeviceInfoFrame, 2000) != KFID_FieldDeviceInfoFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldGetDeviceInfo(4);
	if (WaitUntilResponseType(KFID_FieldDeviceInfoFrame, 2000) != KFID_FieldDeviceInfoFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldGetDeviceInfo(5);
	if (WaitUntilResponseType(KFID_FieldDeviceInfoFrame, 2000) != KFID_FieldDeviceInfoFrame) return false;
	Application->ProcessMessages();

	// usage info
	FDataHandler->SendFieldGetDeviceInfo(0x10);
	if (WaitUntilResponseType(KFID_FieldDeviceInfoFrame, 2000) != KFID_FieldDeviceInfoFrame) return false;
	Application->ProcessMessages();
	// error counters
	FDataHandler->SendFieldGetDeviceInfo(0x19);
	if (WaitUntilResponseType(KFID_FieldDeviceInfoFrame, 2000) != KFID_FieldDeviceInfoFrame) return false;
	Application->ProcessMessages();

	FDataHandler->FLastLatency;

	// request parameter sets 1-4, i.e. blocks 0,1,2,3,8,9,16,17,18,19,56
	FDataHandler->SendFieldGetParameterFrame(0);
	if (WaitUntilResponseType(KFID_FieldParameterFrame, 2000) != KFID_FieldParameterFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldGetParameterFrame(1);
	if (WaitUntilResponseType(KFID_FieldParameterFrame, 2000) != KFID_FieldParameterFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldGetParameterFrame(2);
	if (WaitUntilResponseType(KFID_FieldParameterFrame, 2000) != KFID_FieldParameterFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldGetParameterFrame(3);
	if (WaitUntilResponseType(KFID_FieldParameterFrame, 2000) != KFID_FieldParameterFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldGetParameterFrame(8);
	if (WaitUntilResponseType(KFID_FieldParameterFrame, 2000) != KFID_FieldParameterFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldGetParameterFrame(9);
	if (WaitUntilResponseType(KFID_FieldParameterFrame, 2000) != KFID_FieldParameterFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldGetParameterFrame(16);
	if (WaitUntilResponseType(KFID_FieldParameterFrame, 2000) != KFID_FieldParameterFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldGetParameterFrame(17);
	if (WaitUntilResponseType(KFID_FieldParameterFrame, 2000) != KFID_FieldParameterFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldGetParameterFrame(18);
	if (WaitUntilResponseType(KFID_FieldParameterFrame, 2000) != KFID_FieldParameterFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldGetParameterFrame(19);
	if (WaitUntilResponseType(KFID_FieldParameterFrame, 2000) != KFID_FieldParameterFrame) return false;
	Application->ProcessMessages();

	// check parameter checksum (|0x80)
	FDataHandler->SendFieldGetParameterFrame(56 | 0x80);
	if (WaitUntilResponseType(KFID_FieldParameterFrame, 2000) != KFID_FieldParameterFrame) return false;
	FDeviceParameterCheckFailed = !FLastFrames.FLastFieldParameterSetFrame.UpdateCheckBit;

	// download shot count page
	CheckForShots();

	FDataHandler->FLastLatency;

	/*
	int LRingTarget = GetUserScenarioSet().CLA_RING_TARGET_NUMBER;
	if (LRingTarget < 16) DownloadRingTarget(LRingTarget);

	LRingTarget = GetUserScenarioSet().DYN_RING_TARGET_NUMBER;
	if (LRingTarget < 16) DownloadRingTarget(LRingTarget);

	int LSilTarget = GetUserScenarioSet().CLA_SILH_TARGET_NUMBER;
	if (LSilTarget < 16) DownloadSilhouetteTarget(LSilTarget);

	LSilTarget = GetUserScenarioSet().DYN_SILH_TARGET_NUMBER;
	if (LSilTarget < 16) DownloadSilhouetteTarget(LSilTarget);
	*/

	// debug
	//LStartTickCount = GetTickCount()-LStartTickCount;

	FLastFrames.FFieldParameterSet4;

	// get 10 chars from the device info frame to make serial number
	FSerialNumber = AnsiString((char*)FLastFrames.FFieldParameterSet4.SERIAL_NUMBER, 10);
	FShooter->FSerialNumber = FSerialNumber;

	FLastFrames.FFieldUsageInfoFrame;
	FLastFrames.FFieldErrorCounterFrame;

	FInitializing = false;
	FInitDeviceReady = true;

	// copy the use scenario set to shooter
	FShooter->FUseScenario = FLastFrames.FFieldParameterSet1;

	FDataHandler->FLastLatency;

	//AddToLog("Found COM"+AnsiString(FCOMPort));
	//AddToLog("Last latency "+AnsiString(int(1000*FDataHandler->FLastLatency))+" ms");
	//AddToLog("");

	return true;
}

TParameterSet_Field_1& TDevice::GetUserScenarioSet()
{
	return FLastFrames.FFieldParameterSet1;
}

bool TDevice::UploadUseScenario(TParameterSet_Field_1& AUserScenarioSet)
{
	// update parameter set here and upload it
	FLastFrames.FFieldParameterSet1 = AUserScenarioSet;

	FDataHandler->SendStartProgramming();
	if (WaitUntilResponseType(KFID_AckFrame, 10000) != KFID_AckFrame) return false;

	FDataHandler->SendFieldSetParameterFrame(0,GetPointerToFieldParameterFrame(0));
	if (WaitUntilResponseType(KFID_AckFrame, 2000) != KFID_AckFrame) return false;

	FDataHandler->SendFieldSetParameterFrame(1,GetPointerToFieldParameterFrame(1));
	if (WaitUntilResponseType(KFID_AckFrame, 2000) != KFID_AckFrame) return false;

	FDataHandler->SendFieldSetParameterFrame(2,GetPointerToFieldParameterFrame(2));
	if (WaitUntilResponseType(KFID_AckFrame, 2000) != KFID_AckFrame) return false;

	FDataHandler->SendFieldSetParameterFrame(3,GetPointerToFieldParameterFrame(3), true);
	if (WaitUntilResponseType(KFID_AckFrame, 2000) != KFID_AckFrame) return false;

    // TODO function mode "0,0,0x20", if needed
    FDataHandler->SendFieldSetFunctionMode(2, 0, 0x21);
    if (WaitUntilResponseType(KFID_AckFrame, 3000) != KFID_AckFrame) return false;

	return true;
	
}

bool TDevice::UploadParameterSets()
{
	FDataHandler->SendStartProgramming();
	if (WaitUntilResponseType(KFID_AckFrame, 10000) != KFID_AckFrame) return false;
	Application->ProcessMessages();

	FDataHandler->SendFieldSetParameterFrame(0,GetPointerToFieldParameterFrame(0));
	if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldSetParameterFrame(1,GetPointerToFieldParameterFrame(1));
	if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame) return false;
    Application->ProcessMessages();
	FDataHandler->SendFieldSetParameterFrame(2,GetPointerToFieldParameterFrame(2));
	if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame) return false;
	Application->ProcessMessages();
	FDataHandler->SendFieldSetParameterFrame(3,GetPointerToFieldParameterFrame(3));
    if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame) return false;
    Application->ProcessMessages();
	FDataHandler->SendFieldSetParameterFrame(8,GetPointerToFieldParameterFrame(8));
    if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame) return false;
    Application->ProcessMessages();
	FDataHandler->SendFieldSetParameterFrame(9,GetPointerToFieldParameterFrame(9));
    if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame) return false;
    Application->ProcessMessages();
    FDataHandler->SendFieldSetParameterFrame(16,GetPointerToFieldParameterFrame(16));
    if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame) return false;
    Application->ProcessMessages();
    FDataHandler->SendFieldSetParameterFrame(17,GetPointerToFieldParameterFrame(17));
    if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame) return false;
    Application->ProcessMessages();
    FDataHandler->SendFieldSetParameterFrame(18,GetPointerToFieldParameterFrame(18));
    if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame) return false;
    Application->ProcessMessages();
    FDataHandler->SendFieldSetParameterFrame(19,GetPointerToFieldParameterFrame(19));
    if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame) return false;
    Application->ProcessMessages();

	// TODO don't update serial number
	FDataHandler->SendFieldSetParameterFrame(56,GetPointerToFieldParameterFrame(56), true);
	if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame) return false;
	Application->ProcessMessages();

	// TODO function mode "0,0,0x20", if needed
    FDataHandler->SendFieldSetFunctionMode(2, 0, 0x21);
    if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame) return false;

	return true;
}

void TDevice::ClearZeroing()
{
	FSoftZeroingX = 0;
	FSoftZeroingY = 0;
	FSoftZeroingXPrevious = 0;
	FSoftZeroingYPrevious = 0;
}

//
void TDevice::StartMeasuring(int AMode)
{
	// reset zeroing when starting dynamic
	if (AMode == 2)
	{
		ClearZeroing();
	}

	FMeasuringMode = AMode;
	FDataHandler->SendFieldSetFunctionMode(AMode /*0x02*/, 0x00, 0x20);
	// TODO Wait for ack
	FReceivedSampleCounter = 0;

	FMeasuring = true;
}

// stop measurement
void TDevice::StopMeasuring()
{
	//FMeasuringMode = 0;
	FDataHandler->SendFieldSetFunctionMode(0x02, 0x00, 0x21);

	FMeasuring = false;
}

// return unknown frame/no frame
EFRAMEID_FrameTypes TDevice::WaitUntilResponse(unsigned int ATimeout_ms)
{
    // reset received bits
	FLastFrames.ResetReceivedHandled();

    unsigned int LWaitTime = 0;
    bool LCont = true;
    // sleep in 10ms until ack received
    while (LCont)
	{
    	// none of these don't seem to force thread change (allow serial thread to do its job)
        Sleep(10);
        Sleep(0);
        SwitchToThread();
        LWaitTime += 10;
        if (FLastFrames.AnyFrameReceived())
        {
            return FLastFrames.FLastReceivedFrameType;
        }
        // if waited for timeout time; go out
        if (LWaitTime > ATimeout_ms) LCont = false;
    }
    // none received if timeout
    return KFID_None;
}

// return error/requested frame or timeout
EFRAMEID_FrameTypes TDevice::WaitUntilResponseType(EFRAMEID_FrameTypes AFrameWanted, unsigned int ATimeout_ms)
{
    // reset received bits
	FLastFrames.ResetReceivedHandled();
    FLastFrames.FLastReceivedFrameType = KFID_None;

    unsigned int LWaitTime = 0;
	bool LCont = true;
	// sleep in 10ms until ack received
    while (LCont)
    {
    	// none of these don't seem to force thread change (allow serial thread to do its job)
		Sleep(10);
        Sleep(0);
		SwitchToThread();
		LWaitTime += 10;
        if (FLastFrames.FLastReceivedFrameType == AFrameWanted ||
			FLastFrames.FLastReceivedFrameType == KFID_ErrorFrame)
        {
            return FLastFrames.FLastReceivedFrameType;
        }
        // if waited for timeout time; go out
        if (LWaitTime > ATimeout_ms) LCont = false;
    }
    // none received if timeout
	return KFID_None;
}

// create string list if not created already; set file name
void TDevice::SetSaveFilename(AnsiString& AFilename)
{
	FSaveFilename = AFilename;
	FSaveShotNumber = 0;
}

// add strings to string list of this sample
void TDevice::SaveShotAsString()
{
	// String list is created when a save file name is given
	if (FStringListToSave)
	{
		AnsiString str;
		str = "*** "+Now();
		FStringListToSave->Clear();
		FStringListToSave->Add(str);

		// result frame
		str = "Result: ";
		str.cat_printf("X: %d Y: %d Tilt: %d Rota: %d ", FLastResultFrame.GetX(), FLastResultFrame.GetY(), FLastResultFrame.GetTilt(), FLastResultFrame.GetRotation());
		str += GetStringAsHex(FLastResultFrame.RAW_BYTE, sizeof(TResponseFrame_FieldResultFrame));
		FStringListToSave->Add(str);

		for (int i=FNextFrameIndex-1; i > FNextFrameIndex-20; i--)
		{
			// if we go below 0, continue from the last frame downwards
			int LFrameIndex = i >= 0 ? i : i + FFrameBufferSize;

			// if not enough stuff in buffer..
			if (i < 0 && !FRingBufferGoneAround) break;
			FStringListToSave->Add("");

			TResponseFrame_FieldSampleFrame* LFrame = &FFrames[LFrameIndex];
			if (LFrame->SYNCBYTE1 != 0)
			{
				str.printf("%3d ", i-(FNextFrameIndex-1));
				str.cat_printf("X: %d Y: %d Tilt: %d Rota: %d ", LFrame->GetX(), LFrame->GetY(), LFrame->GetTilt(), LFrame->GetRotation());
				str += GetStringAsHex(LFrame->RAW_BYTE, sizeof(TResponseFrame_FieldSampleFrame));
				FStringListToSave->Add(str);
			}
		}

		int LPos = FSaveFilename.LastDelimiter(".");
		str = FSaveFilename.SubString(1, LPos-1)+"_"+FSaveShotNumber+".txt";
		FStringListToSave->SaveToFile(str);
		FSaveLastFilename = str;
		FSaveShotNumber++;
	}
}

int TDevice::GetShotCount()
{
	if (FShooter)
		return FShooter->GetShotCount();
	else
		return 0;
}

TShot* TDevice::GetShot(int AShotIndex)
{
    return NULL;
}

void TDevice::CheckAndAddShotToShooter(TShooter* AShooter, TResponseFrame_FieldResultFrame* AFramePtr)
{
	int AShotAlreadyExists = AShooter->GetShotIndexWithResultFrame(*AFramePtr);

    // -1 = not exists, only in that case we create and store the shot
	if (AShotAlreadyExists > -1)
	{
		return;
    }

    // create new shot; allocate samples according to how many we currently have for this shot
    // if ring buffer gone around: we have as many sample as buffer can take; otherwise "next index"
	TShot* LNewShot = new TShot();
	LNewShot->FShooterIndex = FShooter->FShooterIndex;
    int LSampleCount = FRingBufferGoneAround ? FFrameBufferSize : FNextFrameIndex;
    LNewShot->SetSampleCount(LSampleCount);
	LNewShot->FAimingTime = FAimingTime;

    // copy result frame
    LNewShot->FResultSample = *AFramePtr;
    LNewShot->FShotIndexInList = AShooter->FShotList->Count;

	// if gone around, the first sample to copy is the one which would be next replaced
    int LFirstSample = FRingBufferGoneAround ? FNextFrameIndex : 0;
    for (int i=0; i<LSampleCount; i++)
    {
        LNewShot->FSamples[i] = FFrames[(LFirstSample + i) % FFrameBufferSize];
    }

	// TODO copy series, shot number and scores
	//LNewShot->FSeriesNumber = AFramePtr->

	AShooter->AddShot(*LNewShot);

}

// stop measuring and restart measuring in zeroing mode
void TDevice::EnterZeroingMode()
{
    // THIS IS NOT USED ATM
	// first send stop, a sympathic little pause and send start zeroing mode
	//FDataHandler->SendFieldSetFunctionMode(0x00, 0x00, 0x21);
	//Sleep(50);
	//FDataHandler->SendFieldSetFunctionMode(0x05, 0x00, 0x20);
	FZeroingMode = true;
}

void TDevice::StoreZeroing()
{
	FSoftZeroingXPrevious = FSoftZeroingX;
	FSoftZeroingYPrevious = FSoftZeroingY;
}

bool TDevice::PerformZeroing()
{
	bool LOK = true;
	double LX = 0;
	double LY = 0;
	int LCount = 0;
	for (int i=0; i < 3; i++)
	{
		// abandon shot if a failure bit is set
		if (FZeroingShots[i].FAILUREBITS || FZeroingShots[i].SYNCBYTE1 == 0) continue;

		LX += FZeroingShots[i].GetX();
		LY += FZeroingShots[i].GetY();
		LCount++;
	}

	// "clear" zeroing shots after calculate
	FZeroingShots[0].SYNCBYTE1 = 0;
	FZeroingShots[1].SYNCBYTE1 = 0;
	FZeroingShots[2].SYNCBYTE1 = 0;

	// take average
	if (LCount > 0)
	{
		LX /= LCount;
		LY /= LCount;

		// check zeroing limit, 3750 radius NAU
		if (LX * LX + LY * LY > 3750*3750)
		{
			// restore previous zeroing
			FSoftZeroingX = FSoftZeroingXPrevious;
			FSoftZeroingY = FSoftZeroingYPrevious;
		}
		else
		{
			// store values
			FSoftZeroingX = LX;
			FSoftZeroingY = LY;
		}
	}
	else
	{
        // Reset is now performed by caller
		// NO LONGER RESTORE, BUT RESET ON NO SHOTS. restore previous zeroing
		FSoftZeroingX = 0;//FSoftZeroingXPrevious;
		FSoftZeroingY = 0,//FSoftZeroingYPrevious;
		LOK = true;//false;
	}

	// THIS IS NOT USED ATM
	// enter previous operating mode via a short sleep
	//FDataHandler->SendFieldSetFunctionMode(0, 0x00, 0x21);
	//Sleep(50);
	//FDataHandler->SendFieldSetFunctionMode(FMeasuringMode, 0x00, 0x20);

	return LOK;
}

void TDevice::AddZeroingShot(TResponseFrame_FieldResultFrame& AFrame)
{
	// scroll the 2 first shots forward in the 3 shot buffer
	if (!AFrame.FAILUREBITS)
	{
		FZeroingShots[2] = FZeroingShots[1];
		FZeroingShots[1] = FZeroingShots[0];
		FZeroingShots[0] = AFrame;
	}
}

void TDevice::UpdateSilhouetteScore(TResponseFrame_FieldResultFrame& AResultFrame)
{
	if (!FCurrentTarget.FUsingSilhouette)
	{
		throw new Exception("Calculating silhouette score for non-silhouette target");
	}

	if (AResultFrame.FUNCTION_MODE != 1)
	{
		throw new Exception("Calculating silhouette score for non-classic target");
	}

	TSilhouetteTarget& LSilTarget = FCurrentTarget.FSilTarget;

	// Distance with
	double LX;
	double LY;
	if (AResultFrame.FUNCTION_MODE == 1)
	{
		LX = AResultFrame.GetX() * ONE_NAU_to_mm_in_one_meter * (LSilTarget.FSilhouetteSet.GetNominalDistance_m() - 0.75);
		LY = AResultFrame.GetY() * ONE_NAU_to_mm_in_one_meter * (LSilTarget.FSilhouetteSet.GetNominalDistance_m() - 0.75);
	}
//	else
//	{
//		LX = AResultFrame.GetX() * ONE_NAU_to_mm_in_one_meter * 0.1 * AResultFrame.GetDist();
//		LY = AResultFrame.GetY() * ONE_NAU_to_mm_in_one_meter * 0.1 * AResultFrame.GetDist();
//	}

	// 32 would be the center (XY=0,0). +- 0.5 * size gives -1..1 => -32..+32 = 0..63
	double LCX = (32.0 + 32.0 * (LX / (0.5 * LSilTarget.FSilhouetteSet.GetSilhouetteSize_mm() ) ) );
	double LCY = (32.0 - 32.0 * (LY / (0.5 * LSilTarget.FSilhouetteSet.GetSilhouetteSize_mm() ) ) );

	int LScore = 0;
	// if we are on target, get the "pixel" color. otherwise it is miss
	if (LCX >= 0 && LCX < 64 && LCY >= 0 && LCY < 64)
	{
		LScore = FCurrentTarget.FSilhouette.FPixelState[(int)LCX][(int)LCY];
	}

	AResultFrame.SILHOUETTE_RESULT = LScore;

}

void TDevice::UpdateScoreAndSector(TResponseFrame_FieldResultFrame& AResultFrame)
{
	if(!FCurrentTarget.FUsingRing)
	{
		throw new Exception("Calculating score and sector for non-ring target");
	}

	TRingTarget& LRingTarget = FCurrentTarget.FRingTarget;

	// Distance with
	double LX;
	double LY;
	if (AResultFrame.FUNCTION_MODE == 1)
    {
		LX = AResultFrame.GetX() * ONE_NAU_to_mm_in_one_meter * LRingTarget.FRingSet1.GetNominalDistance_m_calc();
		LY = AResultFrame.GetY() * ONE_NAU_to_mm_in_one_meter * LRingTarget.FRingSet1.GetNominalDistance_m_calc();
	}
	else
	{
		LX = AResultFrame.GetX() * ONE_NAU_to_mm_in_one_meter * 0.1 * AResultFrame.GetDist();
		LY = AResultFrame.GetY() * ONE_NAU_to_mm_in_one_meter * 0.1 * AResultFrame.GetDist();
	}

	// pythagoras to get distance from center
	double LR = Sqrt(LX * LX + LY * LY);
	double LCal = LRingTarget.FRingSet1.GetCaliber_mm();

	int LHighestRing = LRingTarget.FRingSet1.HIGHEST_HIT_RING;
    double LHighestRadius = LRingTarget.GetRadiusForRingmm(LHighestRing);
    int LLowestRing = LRingTarget.FRingSet1.MINIMUM_HIT_RING;
	double LXRadius = LRingTarget.FRingSet2.GetXDiameter() * 0.05;
	double LScore = 0;
    int LSector = 0;

	// are we within highest ring?
	if (LR <= LRingTarget.GetRadiusForRingmm(LHighestRing) + LCal * 0.5)
	{
		// Interpolate between.... 10.0 - 10.9
		double LDecimals = 0.99 - (0.99 * LR / (LHighestRadius + LCal * 0.5) );
		LScore = LDecimals + LHighestRing;
	}
	else
	{
		// which ring .. ring+1 are we checking against
		int LNowEstimating = LHighestRing - 1;
        bool LCont = true;

		// we go on until we get the right rings to interpolate between, or we get out of target.
		while (LCont)
		{
        	// we are trying to find the correct place between these two
			double LUpperRing = LRingTarget.GetRadiusForRingmm(LNowEstimating + 1);
			double LLowerRing =	LRingTarget.GetRadiusForRingmm(LNowEstimating);
			double LRingWidth = LLowerRing - LUpperRing;

			// if we are farther away than caliber corrected lower check ring width, go for next ring
			if (LR > LLowerRing + LCal * 0.5)
			{
				LNowEstimating--;
				// check if we are still on target (not out of rings...)
				if (LNowEstimating < LLowestRing)
				{
					// score is 0, don't go on
					LCont = false;
					LScore = 0;					                    
				}
			}
			else
			{
				// let's calculate the place between the rings
				double LDistFromLower = (LLowerRing - LR) + LCal * 0.5;
				double LDecimals = LDistFromLower / LRingWidth;
				LScore = LDecimals + LNowEstimating;
				LCont = false;
			}
		}
	}

	// calculate sector
	if (LY == 0)
	{
		LSector = LX > 0 ? 3 : 9;
	}
	else
	{
		double LAtan = atan2(-LX, -LY);
		LSector = 0.5 + 6.0 + LAtan * 6.0 / M_PI;
		if (LSector < 1) LSector = 12;
	}

	AResultFrame.RING_RESULT = 10 * LScore;
	AResultFrame.RingSector = LSector;
	AResultFrame.XResult = (LR <= LXRadius); 
}

// ** Would be wiser if was on main form itself.
// Add measured blow to blow list. Called by com port when received a 'blow'. Creates a copy of the blow instance.
void TDevice::HandleReceivedBlow(TResponseFrame_FieldResultFrame* AFramePtr)
{
	if (FShooter==NULL)
	{
		//int Idebug;
		//Idebug++;
	}

	// Close range hit special handling.
	// 27.1.2011. Close range hit handling testing. Short distance but not oversignal,
	bool LAllowFailure = AFramePtr->ReducedTXPower && (AFramePtr->MeasureErrorFOV || AFramePtr->FSIError);
	if (LAllowFailure)
	{
		AFramePtr->FAILUREBITS &= ~(0x10 | 0x40); // MFOVerror bit4, FSIerror bit6
	}


	// store last result frame for later usage
	FLastResultFrame = *AFramePtr;

	// check if to save hits to files
    if (FSaveFilename.Length()>0)
    {
		SaveShotAsString();
	}

	// rotation to degrees
	double LToDeg = Max(-1.0, Min(1.0, 0.001 * AFramePtr->GetRotation() ) );
	int LDegrees = 360.0 / 6.28 * asin(LToDeg);
	AFramePtr->ROTATION_MSB = (LDegrees & 0xff00) >> 8;
	AFramePtr->ROTATION_LSB = LDegrees & 0xff;

	// tilt
	LToDeg = Max(-1.0, Min(1.0, 0.001 * AFramePtr->GetTilt() ) );
	LDegrees = 360.0 / 6.28 * asin(LToDeg);
	AFramePtr->TILT_MSB = (LDegrees & 0xff00) >> 8;
	AFramePtr->TILT_LSB = LDegrees & 0xff;

	// connected? if so, set used target number to 0
	if (FConnectedShooting)
	{
    	// set to ignore next aim samples... (recoil)
    	FIgnoreSamplesNow = FIgnoreSamplesAfterTrigger;

		// this is a hack. when connected, we always use target number 0 for both ring and silhouette.
		AFramePtr->RING_TARGET = FShooter->FTargetDefinition.FUsingRing ? 0 : 255;
		AFramePtr->SILHOUETTE_TARGET = FShooter->FTargetDefinition.FUsingSilhouette ? 0 : 255;

		// handle zeroing mode: store shot in 3-shot buffer.
		if (FZeroingMode)
		{
			AddZeroingShot(*AFramePtr);
		}
		else
		{
			int LXNAU = AFramePtr->GetX();
			int LYNAU = AFramePtr->GetY();

			double LCoeff;
			if (FMeasuringMode == 1)
			{
				// Also apply nominal distance in classic, also hand length (0.75, in m_calc)
				LCoeff = ONE_NAU_to_mm_in_one_meter * (FCurrentTarget.GetDist_m() - 0.75);
				// Apply zeroing in classic mode
				LXNAU -= FSoftZeroingX;
				LYNAU -= FSoftZeroingY;
			}
			else
			{
				// translate curve to mm using measured distance
				LCoeff = ONE_NAU_to_mm_in_one_meter * 0.1 * AFramePtr->GetDist();
			}

			// NAU to mm
			double LXmm10 = 10 * LXNAU * LCoeff;
			double LYmm10 = 10 * LYNAU * LCoeff;

			AFramePtr->SetX_raw(LXNAU);
			AFramePtr->SetY_raw(LYNAU);

			AFramePtr->SetXmm_raw(LXmm10);
			AFramePtr->SetYmm_raw(LYmm10);

			if (FShooter->FTargetDefinition.FUsingRing)
			{
				UpdateScoreAndSector(*AFramePtr);
			}

			if (FShooter->FTargetDefinition.FUsingSilhouette && FMeasuringMode == 1)
			{
            	UpdateSilhouetteScore(*AFramePtr);
			}

			// change curve to mm
			for (int i = 0; i < FFrameBufferSize; i++)
			{
				 TResponseFrame_FieldSampleFrame& LFrame = FFrames[i];

				 // NAU to 0.1mm
				 int L = LFrame.GetX() * LCoeff * 10;
				 // check if we dont fit within 16 bits (put a failure on)
				 bool LBad = ((short)L) != L;
				 LFrame.SetX(L);

				 L = LFrame.GetY() * LCoeff * 10;
				 LBad |= ((short)L) != L;
				 LFrame.SetY(L);

				 // if we cant fit in 16 bits, put measure error on
				 if (LBad)
				 {
					LFrame.MeasureErrorFOV = true;
				 }
			}
		}
	}
	else // downloading, save mm info to NAU info
	{

	}

	// either connected or download
	if (!FZeroingMode)
	{
    	// if both ring and silhouette, AND silhouette score = 0 -> ring score = 0
        // 8.6.2010. No longer zero ring score if miss silhouette
		//if ((AFramePtr->RING_TARGET < 16) & (AFramePtr->SILHOUETTE_TARGET < 16) & (AFramePtr->SILHOUETTE_RESULT == 0))
		//{
		//	AFramePtr->RING_RESULT = 0;
		//}

		CheckAndAddShotToShooter(FShooter, AFramePtr);

		// received shot -> ask whether to save on close
		GShotsReceivedAfterSave = true;
	}

	// reset buffer and counters
	ClearSampleBuffer();
	FAimingTime = 0;

	// inform listener that shot happened and is available in shot list
	if (FEventListener)
	{
		FEventListener->EventShotReceived();
	}
}

void TDevice::ConvertNAUToMMAfterDownload()
{
	for (int i = 0; i < FShooter->GetShotCount(); i++)
	{
    	TShot& LShot = FShooter->GetShot(i);

		// Convert classic mode shots' NAU coords to mm
		if (LShot.FResultSample.FUNCTION_MODE == 1)
		{
        	double LCoeff;

			if (LShot.FResultSample.RING_TARGET < 16 /*!= 0xff*/)
			{
				// Also apply nominal distance in classic, also hand length (0.75, in m_calc)
				LCoeff = ONE_NAU_to_mm_in_one_meter * (FAllTargets.GetRingTarget(LShot.FResultSample.RING_TARGET).FRingSet1.GetNominalDistance_m_calc());
			}
			else if (LShot.FResultSample.SILHOUETTE_TARGET < 16 /*!= 0xff*/)
			{
				LCoeff = ONE_NAU_to_mm_in_one_meter * (FAllTargets.GetSilhouetteTarget(LShot.FResultSample.SILHOUETTE_TARGET).FSilhouetteSet.GetNominalDistance_m() - 0.75) ;
			}
			else throw new Exception("Bad target number");

			// copy mm to XY
			for (int i = 0; i < LShot.FSampleCount; i++)
			{
				TResponseFrame_FieldSampleFrame& LFrame = LShot.FSamples[i];

				// NAU to 0.1mm
				int L = 10 * LCoeff * LFrame.GetX();
            	// assert we fit in 16 bits, otherwise make sample bad
				bool LBad = ( (short)L ) != L;
				LFrame.SetX(L);

				L = 10 * LCoeff * LFrame.GetY();
                LBad |= ( (short)L ) != L;
				LFrame.SetY(L);

				// if either sample did not fit within 16 bits, put Field Of View error bit on
				if (LBad)
				{
					LFrame.MeasureErrorFOV = true;
				}
			}

			double LXmm10 = 10 * LCoeff * LShot.FResultSample.GetX();
			double LYmm10 = 10 * LCoeff * LShot.FResultSample.GetY();

			LShot.FResultSample.SetXmm_raw(LXmm10);
			LShot.FResultSample.SetYmm_raw(LYmm10);

			// we need to update this in order to get correct calcuations...
			FCurrentTarget.FRingTarget = FAllTargets.GetRingTarget(LShot.FResultSample.RING_TARGET);
			FCurrentTarget.FUsingRing = true;
			// TODO check that this works
			FCurrentTarget.FUsingSilhouette;// = false;
			UpdateScoreAndSector(LShot.FResultSample);

		}
		
	}

}

void TDevice::HandleReceivedAimSample(TResponseFrame_FieldSampleFrame* AFramePtr)
{
	// apply zeroing to sample in classic (zeroing only available in realtime shooting, not downloads)
	if ( FMeasuringMode == 1 && (FSoftZeroingX || FSoftZeroingY) )
	{
		int LX = AFramePtr->GetX() - FSoftZeroingX;
		int LY = AFramePtr->GetY() - FSoftZeroingY;
		AFramePtr->SetX(LX);
		AFramePtr->SetY(LY);
	}

	// rotation to degrees
	double LToDeg = Max(-1.0, Min(1.0, 0.001 * AFramePtr->GetRotation() ) );
	int LDegrees = 360.0 / 6.28 * asin(LToDeg);
	AFramePtr->ROTATION_MSB = (LDegrees & 0xff00) >> 8;
	AFramePtr->ROTATION_LSB = LDegrees & 0xff;

	// tilt
	LToDeg = Max(-1.0, Min(1.0, 0.001 * AFramePtr->GetTilt() ) );
	LDegrees = 360.0 / 6.28 * asin(LToDeg);
	AFramePtr->TILT_MSB = (LDegrees & 0xff00) >> 8;
	AFramePtr->TILT_LSB = LDegrees & 0xff;
	
	FLastSampleFrame = *AFramePtr;
	AddSampleToBuffer(AFramePtr);

	if (AFramePtr->FAILUREBITS)
		FAimingTime = 0;
	else
		FAimingTime += 15; // milliseconds

	// inform listener that a new aim point is available
	if (FEventListener)
	{
		FEventListener->EventAimUpdate();
	}
}

void TDevice::AddSampleToBuffer(TResponseFrame_FieldSampleFrame* AFramePtr)
{
	// if we are measuring realtime and want to ignore recoil samples..
	if (FIgnoreSamplesNow > 0)
	{
		FIgnoreSamplesNow--;
		return;
	}

	// clear sample frame(s)
	memset(&FFrames[FNextFrameIndex], 0, sizeof(TResponseFrame_FieldSampleFrame));

	if (AFramePtr)
		FFrames[FNextFrameIndex] = *AFramePtr;

	FNextFrameIndex++;
	if (FNextFrameIndex == FFrameBufferSize)
	{
		FRingBufferGoneAround = true;
		FNextFrameIndex = 0;
	}

	// count for continuous failure samples, reset buffer if needed
	if (AFramePtr && AFramePtr->FAILUREBITS)
	{
		FFailureSamplesNow++;

		if (FFailureSamplesNow >= FFailureSamplesUntilReset)
		{
			FRingBufferGoneAround = false;
			FNextFrameIndex = 0;			
		}
	}
	else
	{
    	// reset counter
        FFailureSamplesNow = 0;
    }

}

void TDevice::ClearSampleBuffer()
{
	memset(FFrames, 0, FFrameBufferSize * sizeof(TResponseFrame_FieldSampleFrame));

	FRingBufferGoneAround = false;
	FNextFrameIndex = 0;
}

double TDevice::GetCurrentSampleRate()
{
	_LARGE_INTEGER LNow;
	QueryPerformanceCounter(&LNow);
	// get 'tick' count
	__int64 LTicks = LNow.QuadPart - FSampleFirstTime.QuadPart;
	// to milliseconds
	LTicks = (LTicks * 1000) / FPerformanceTimerFreq.QuadPart;

	double rate = 0;
	if (LTicks > 0)
	{
		LTicks;
		rate = double(__int64(1000 * FReceivedSampleCounter)) / LTicks;
	}
	return rate;
}

char* TDevice::GetPointerToFieldParameterFrame(int AFrameNumber)
{
	// parameter sets are transferred from/to device by 16 byte "frames",
	// i.e. a 64 byte set is built of 4 16 byte sets.
	// we get the pointer to the correct place in wanted set
	void* LPtr = NULL;
	if (AFrameNumber >= 0 && AFrameNumber <= 3)  // use case params
	{
		LPtr = &FLastFrames.FFieldParameterSet1.RAW_BYTE[16*AFrameNumber];
	}
	else if (AFrameNumber >= 8 && AFrameNumber <= 11)  // technical params
	{
		LPtr = &FLastFrames.FFieldParameterSet2.RAW_BYTE[16*(AFrameNumber-8)];
	}
	else if (AFrameNumber >= 16 && AFrameNumber <= 19)  // reserved for factory
	{
		//LPtr = &FLastFrames.FFieldParameterSet3.RAW_BYTE[16*(AFrameNumber-16)];
	}
	else if (AFrameNumber == 56)  // device serial number
	{
		LPtr = &FLastFrames.FFieldParameterSet4;
	}
	return (char*)LPtr;
}

void TDevice::CheckAndStoreFieldParameterFrame(TResponseFrame_Field_ParameterSetFrame* APtr)
{
	// 7 low bith of parameter set byte, upmost bit is check bit
	int AFrameNumber = APtr->ParameterSetNumber;

	// parameter sets are transferred from/to device by 16 byte "frames",
	// i.e. a 64 byte set is built of 4 16 byte sets.
	// we get the pointer to the correct place in wanted set
	char* LPtr = GetPointerToFieldParameterFrame(AFrameNumber);

	// frame number was ok...
	if (LPtr)
	{
		memcpy(LPtr, APtr->PARAMETERBYTE, 16);
		FLastFramesUpdated = true;
	}
}

void TDevice::StoreReceivedFrame(void* APtr, EFRAMEID_FrameTypes AFrameTypeDetected)
{
	FLastFrames.FLastReceivedFrameType = AFrameTypeDetected;

	// device sends "i am ready" when dtr is set
	if (AFrameTypeDetected == KFID_StartupByte)
	{
		// fire the "timer" event to note that initialize can proceed
		if (FAsyncInfo.FScanGoing)
		{
			InitializeTimerEvent(NULL);
		}
	}
	// if we receive a result frame, store it and add the shot to buffers
	else if (AFrameTypeDetected == KFID_ResultFrameField)
	{
		FLastFrames.FLastResultFrameField = *(TResponseFrame_FieldResultFrame*)APtr;
		FLastFrames.FResultFrameReceived = true;

		// add to shot list if shot sample
		HandleReceivedBlow(&FLastFrames.FLastResultFrameField);

		// if async download, request next shot.
		// THIS IS MOVED TO ACK AS DOWNLOAD COMMAND IS ACKD AFTER RESULT
		if (FAsyncInfo.FDownloadShotsGoing)
		{
			//ProcessAsyncDownloadShots();
		}
	}
	// if we received a sample frame, pass handling to dedicated function
	else if (AFrameTypeDetected == KFID_SampleFrameField)
	{
		FLastFrames.FLastSampleFrameField = *(TResponseFrame_FieldSampleFrame*)APtr;
		FLastFrames.FSampleFrameReceived = true;
		FLastFrames.FSampleFrameHandled = false;

		// check if we are storing samples (sigma)
		if (FIgnoreNextSamples > 0)
		{
			FIgnoreNextSamples--;
			return;
		}
		// this is just a sample counter for sigma;;; we could also use other counters
		if (FStoreNextSamples > 0)
			FStoreNextSamples--;

		// do the things
		HandleReceivedAimSample((TResponseFrame_FieldSampleFrame*)APtr);

		// maybe could be in data handler...
		if (FReceivedSampleCounter==0)
			QueryPerformanceCounter(&FSampleFirstTime);
		FReceivedSampleCounter++;
	}
	else if (AFrameTypeDetected == KFID_ErrorFrame)
	{
		FLastFrames.FLastErrorFrame = *(TResponseFrame_ErrorFrame*)APtr;
		FLastFrames.FErrorFrameReceived = true;
		FLastFrames.FErrorFrameHandled = false;

		// if async download, request next shot
		if (FAsyncInfo.IsAsyncGoing())
		{
			FAsyncInfo.FAsyncError = true;
			// Callback?

        	//attempt to go on anyway
			//ProcessAsyncDownloadShots();
		}

	}
	else if (AFrameTypeDetected == KFID_InfoErrorFrame)
	{
		FLastFrames.FLastInfoErrorFrame = *(TResponseFrame_Field_InfoErrorFrame*)APtr;
		FLastFrames.FErrorFrameReceived = true;
		FLastFrames.FErrorFrameHandled = false;

		// if async download, request next shot
		//26.1.2017 MHi commented out because causes frequent errors when starting the session
		if (FAsyncInfo.IsAsyncGoing())
		{
			FAsyncInfo.FAsyncError = true; // 26.1.2017 MHi causes frequent errors when starting the session
			// Callback?

			//attempt to go on anyway
			//ProcessAsyncDownloadShots();
		}

	}
	else if (AFrameTypeDetected == KFID_AckFrame)
	{
		FLastFrames.FLastAckFrame = *(TResponseFrame_Field_AckFrame*)APtr;
		FLastFrames.FAckFrameReceived = true;
		FLastFrames.FAckFrameHandled = false;

		// if async download, request next shot, unless we are already loading targets (non-async atm)
		if (FAsyncInfo.FDownloadShotsGoing && !FAsyncInfo.FDownloadTargetsGoing)
		{
			ProcessAsyncDownloadShots();
		}

		// if we are updating targets (async), process next block
		//if (FAsyncInfo.FUploadTargetsGoing && !FAsyncInfo.FUploadTargetsDone)
		if (FAsyncInfo.FUploadTargetsGoing && !FAsyncInfo.FUploadTargetsDone &&
			!(FLastFrames.FLastAckFrame.SYNCBYTE2 == 0x61 && FLastFrames.FLastAckFrame.ACKBYTE == 0x3c && FLastFrames.FLastAckFrame.INFOBYTE == 0x02 )
		)
		{
			ProcessAsyncUploadTargets();
		}

		// if we are scanning a port...
		if (FAsyncInfo.FScanGoing)
		{
            ProcessAsyncInitialize();
        }
	}
	else if (AFrameTypeDetected == KFID_FieldParameterFrame)
	{
		FLastFrames.FLastFieldParameterSetFrame = *((TResponseFrame_Field_ParameterSetFrame*)APtr);
		CheckAndStoreFieldParameterFrame((TResponseFrame_Field_ParameterSetFrame*)APtr);

		// if we are scanning a port...
		if (FAsyncInfo.FScanGoing)
		{
			ProcessAsyncInitialize();
		}
	}
	else if (AFrameTypeDetected == KFID_FieldDeviceInfoFrame)
	{
		FLastFrames.FLastFieldDeviceInfoFrame = *(TResponseFrame_Field_DeviceInfoFrame*)APtr;
		if (FLastFrames.FLastFieldDeviceInfoFrame.PAGE_NUMBER == 0x10)
			FLastFrames.FFieldUsageInfoFrame = FLastFrames.FLastFieldDeviceInfoFrame;
		else if (FLastFrames.FLastFieldDeviceInfoFrame.PAGE_NUMBER == 0x19)
			FLastFrames.FFieldErrorCounterFrame = FLastFrames.FLastFieldDeviceInfoFrame;
		else if (FLastFrames.FLastFieldDeviceInfoFrame.PAGE_NUMBER <= 7)
			FLastFrames.FFieldDeviceInfoPages[FLastFrames.FLastFieldDeviceInfoFrame.PAGE_NUMBER] = FLastFrames.FLastFieldDeviceInfoFrame;

		// if we are scanning a port...
		if (FAsyncInfo.FScanGoing)
		{
            ProcessAsyncInitialize();
        }
	}
	else if (AFrameTypeDetected == KFID_SilhouetteBitmapFrame)
	{
		FLastFrames.FLastSilhouetteBitmapFrame = *(TResponseFrame_DownloadSilhouetteBitmapFrame*)APtr;

		FLastFrames.FLastSilhouetteBitmapFrame.SH1_FRAME_NUMBER == 0;

		if (FAsyncInfo.FDownloadTargetsGoing)
		{
			ProcessAsyncDownloadTargets();
		}

	}
	else if (AFrameTypeDetected == KFID_FieldShotCountFrame)
	{
		FLastFrames.FLastShotCountFrame = *(TResponseFrame_Field_ShotCountFrame*)APtr;
		FLastFrames.FShotCountFrameReceived = true;

		// if we are scanning a port...
		if (FAsyncInfo.FScanGoing)
		{
            ProcessAsyncInitialize();
        }

	}
	else if (AFrameTypeDetected == KFID_RingTargetFrame)
	{
		FLastFrames.FLastRingTargetFrame = *(TResponseFrame_Field_RingTargetFrame*)APtr;

		if (FAsyncInfo.FDownloadTargetsGoing)
		{
			ProcessAsyncDownloadTargets();
		}
	}
	else
	{
		// invalid frame; do something
		//int i=0; i++;
	}

	// so the state machine knows that something has happened
	FLastFramesUpdated = true;
}

bool TDevice::DownloadSilhouetteTarget(int ASilhouetteTargetNumber)
{
	if (ASilhouetteTargetNumber < 0 || ASilhouetteTargetNumber > 255)
		return false;

	// set to true when download is completed succesfully
	FAllTargets.FSilhouetteSet[ASilhouetteTargetNumber] = false;

	TCommandFrame_Field_DownloadSilhouetteBitmap LCMD;
	memset(&LCMD, 0, sizeof(LCMD));
	TResponseFrame_DownloadSilhouetteBitmapFrame LStruct;
	memset(&LStruct, 0, sizeof(LStruct));

	// build the command frame
	LCMD.COMMANDBYTE = CCMD_DownloadSilhouette_Bitmap;
	LCMD.SH1_FRAME_CONTROL = ASilhouetteTargetNumber;

	TSilhouetteTarget& LTarget = FAllTargets.GetSilhouetteTarget(ASilhouetteTargetNumber); /*FSilhouetteTargets[ASilhouetteTargetNumber]*/;
	memset(&LTarget, 0, sizeof(TSilhouetteTarget));

	TSilhouetteBitmapMatrix& LBitmap = FAllTargets.GetSilhouetteBitmapMatrix(ASilhouetteTargetNumber); /*FSilhouetteBitmaps[ASilhouetteTargetNumber]*/;
	memset(&LBitmap, 0, sizeof(LBitmap));

	// first we query for the silhouette info & name..
	LCMD.BitmaskFrame = 0;
	LCMD.SH1_FRAME_NUMBER = 0; // information
	LCMD.CHECKBYTE = FDataHandler->GetChecksum(&LCMD, sizeof(LCMD)-1);
	FDataHandler->SendRaw(&LCMD, sizeof(LCMD));

	// wait for ack
	if (WaitUntilResponseType(KFID_SilhouetteBitmapFrame, 5000) != KFID_SilhouetteBitmapFrame)
	{
		return false;
	}

	// update current silhouette
	LStruct = FLastFrames.FLastSilhouetteBitmapFrame;
	sizeof(LStruct.DATA);
	sizeof(LTarget.FSilhouetteSet);
	sizeof(TParameterSet_Field_SilhouetteSettings);

	LTarget.FSilhouetteSet = *(TParameterSet_Field_SilhouetteSettings *)(LStruct.DATA);

	LCMD.SH1_FRAME_NUMBER = 1; // name
	LCMD.CHECKBYTE = FDataHandler->GetChecksum(&LCMD, sizeof(LCMD)-1);
	FDataHandler->SendRaw(&LCMD, sizeof(LCMD));

	// wait for ack
	if (WaitUntilResponseType(KFID_SilhouetteBitmapFrame, 5000) != KFID_SilhouetteBitmapFrame)
	{
		return false;
	}

	LStruct = FLastFrames.FLastSilhouetteBitmapFrame;
	// this is 16 bytes max, we should make a check for it
	memcpy(LTarget.FTargetName.RAW_BYTE, LStruct.DATA, sizeof(LStruct.DATA));

	// then the actual bitmask
	LCMD.BitmaskFrame = 1;

	for (int LRow = 0; LRow<64; LRow++)
	{
		// request the row
		LCMD.SH1_FRAME_NUMBER = LRow;
		LCMD.CHECKBYTE = FDataHandler->GetChecksum(&LCMD, sizeof(LCMD)-1);
		FDataHandler->SendRaw(&LCMD, sizeof(LCMD));

		// wait for ack
		if (WaitUntilResponseType(KFID_SilhouetteBitmapFrame, 5000) != KFID_SilhouetteBitmapFrame)
		{
			return false;
		}

		LStruct = FLastFrames.FLastSilhouetteBitmapFrame;

		// count bytes as we go process the frame
		unsigned int LByteNumber = 0;

		for (int LCol = 0; LCol<64; LCol+=4)
		{
			unsigned char LVal = LStruct.DATA[LByteNumber];

			// left most pixel to bits 6,7, 2nd left 4,5; 3rd 2,3; 4th 0,1. LVal is zerong on first line
			LBitmap.FPixelState[LCol][LRow] = (LVal & 0xc0) >> 6;
			LBitmap.FPixelState[LCol+1][LRow] = (LVal & 0x30) >> 4;
			LBitmap.FPixelState[LCol+2][LRow] = (LVal & 0x0c) >> 2;
			LBitmap.FPixelState[LCol+3][LRow] = LVal & 0x03;
			LByteNumber++;
		}
	}

	FAllTargets.FSilhouetteSet[ASilhouetteTargetNumber] = true;

	return true;
}

bool TDevice::DownloadRingTarget(int ARingTargetNumber)
{
	TRingTarget LTarget;
	memset(&LTarget,0,sizeof(LTarget));

	// first basic block, then 2 supplemental blocks and finally name block
	FDataHandler->SendFieldGetRingTarget(ARingTargetNumber,0);
	if (WaitUntilResponseType(KFID_RingTargetFrame, 10000) != KFID_RingTargetFrame)
		return false;
	LTarget.FRingSet1 = *FLastFrames.FLastRingTargetFrame.GetAsRingTarget();

	FLastFrames.ResetReceivedHandled();

	FDataHandler->SendFieldGetRingTarget(ARingTargetNumber,1);
	if (WaitUntilResponseType(KFID_RingTargetFrame, 10000) != KFID_RingTargetFrame)
		return false;
	LTarget.FRingSet2 = *FLastFrames.FLastRingTargetFrame.GetAsRingTarget_block2();

	FLastFrames.ResetReceivedHandled();

	FDataHandler->SendFieldGetRingTarget(ARingTargetNumber,2);
	if (WaitUntilResponseType(KFID_RingTargetFrame, 10000) != KFID_RingTargetFrame)
		return false;
	LTarget.FRingSet3 = *FLastFrames.FLastRingTargetFrame.GetAsRingTarget_block3();

	FLastFrames.ResetReceivedHandled();

	FDataHandler->SendFieldGetRingTarget(ARingTargetNumber,3 | 0x80); // verift check bit
	if (WaitUntilResponseType(KFID_RingTargetFrame, 10000) != KFID_RingTargetFrame)
		return false;
	LTarget.FTargetName = *FLastFrames.FLastRingTargetFrame.GetAsRingTarget_Name();

	if (!FLastFrames.FLastRingTargetFrame.UpdateCheckBit)
		return false;

	if (ARingTargetNumber < 16)
	{
		//FRingTargets[ARingTargetNumber] = LTarget;
		FAllTargets.SetRingTarget(ARingTargetNumber, LTarget);
		FAllTargets.FRingTargetSet[ARingTargetNumber] = true;
	}
	else
	{
		FAllTargets.FFixedRingTarget = LTarget;
	}

	return true;
}

bool TDevice::UploadRingTarget(int ARingTargetNumber, TRingTarget& ARingTarget)
{
	if (ARingTargetNumber < 0 || ARingTargetNumber > 255) throw new Exception("Bad ring target number "+ARingTargetNumber);

	TParameterSet_Field_RingTarget& FRingSet1 = ARingTarget.FRingSet1;
	TParameterSet_Field_RingTarget_block2& FRingSet2 = ARingTarget.FRingSet2;
	TParameterSet_Field_RingTarget_block3& FRingSet3 = ARingTarget.FRingSet3;
	TParameterSet_Any16bytes& FTargetName = ARingTarget.FTargetName;

	FDataHandler->SendStartProgramming();
	if (WaitUntilResponseType(KFID_AckFrame, 10000) != KFID_AckFrame) return false;

	FDataHandler->SendFieldSetRingTarget(ARingTargetNumber, 0, &FRingSet1, false);
	if (WaitUntilResponseType(KFID_AckFrame, 2000) != KFID_AckFrame) return false;

	FDataHandler->SendFieldSetRingTarget(ARingTargetNumber, 1, &FRingSet2, false);
	if (WaitUntilResponseType(KFID_AckFrame, 2000) != KFID_AckFrame) return false;

	FDataHandler->SendFieldSetRingTarget(ARingTargetNumber, 2, &FRingSet3, false);
	if (WaitUntilResponseType(KFID_AckFrame, 2000) != KFID_AckFrame) return false;

	FDataHandler->SendFieldSetRingTarget(ARingTargetNumber, 3, &FTargetName, true);
	if (WaitUntilResponseType(KFID_AckFrame, 2000) != KFID_AckFrame) return false;


	FDataHandler->SendFieldSetFunctionMode(0x01, 0x00, 0x21);
	if (WaitUntilResponseType(KFID_AckFrame, 10000) != KFID_AckFrame) return false;

	return true;
}

bool TDevice::UploadSilhouetteTarget(int ASilTargetNumber, TSilhouetteTarget& ASilTarget, TSilhouetteBitmapMatrix& ABitmap)
{
	FDataHandler->SendStartProgramming();

	if (WaitUntilResponseType(KFID_AckFrame, 10000) != KFID_AckFrame)
	{
		return false;
	}

	TCommandFrame_Field_UploadSilhouetteBitmap LStruct;
	memset(&LStruct, 0, sizeof(LStruct));

	LStruct.COMMANDBYTE = CCMD_UploadSilhouette_Bitmap;
	LStruct.SilhouetteNumber = ASilTargetNumber;
	LStruct.BitmaskFrame = 0;

	// first we query for the silhouette info & name..
	LStruct.SilhouetteTargetNameBlock = 0; //
	memcpy(LStruct.DATA, &ASilTarget.FSilhouetteSet, sizeof(LStruct.DATA));
	LStruct.CHECKBYTE = FDataHandler->GetChecksum(&LStruct, sizeof(LStruct)-1);
	FDataHandler->SendRaw(&LStruct, sizeof(LStruct));

	// wait for ack
	if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame)
	{
		FDataHandler->SendFieldSetFunctionMode(2, 0, 0x21);
		return false;
	}

	// update name
	LStruct.SilhouetteTargetNameBlock = 1;
	memcpy(LStruct.DATA, &ASilTarget.FTargetName, sizeof(LStruct.DATA));
	LStruct.CHECKBYTE = FDataHandler->GetChecksum(&LStruct, sizeof(LStruct)-1);
	FDataHandler->SendRaw(&LStruct, sizeof(LStruct));

	// wait for ack
	if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame)
	{
		FDataHandler->SendFieldSetFunctionMode(2, 0, 0x21);
		return false;
	}

	LStruct.BitmaskFrame = 1;

	for (int LRow = 0; LRow<64; LRow++)
	{
		// 4 pixels make a byte
		unsigned int LByteNumber = 0;
		unsigned char LVal/* = 0*/;
		for (int LCol = 0; LCol<64; LCol+=4)
		{
			// left most pixel to bits 6,7, 2nd left 4,5; 3rd 2,3; 4th 0,1. LVal is zerong on first line
			LVal = ABitmap.FPixelState [LCol][LRow] << 6;
			LVal |= ABitmap.FPixelState[LCol+1][LRow] << 4;
			LVal |= ABitmap.FPixelState[LCol+2][LRow] << 2;
			LVal |= ABitmap.FPixelState[LCol+3][LRow];
			LStruct.DATA[LByteNumber] = LVal;
			LByteNumber++;
		}

		// row = 'frame number'
		LStruct.SH1_FRAME_NUMBER = LRow;
		// update check bit for last frame
		if (LRow==63) LStruct.UpdateCheckBit = 1;
		LStruct.CHECKBYTE = FDataHandler->GetChecksum(&LStruct, sizeof(LStruct)-1);
		FDataHandler->SendRaw(&LStruct, sizeof(LStruct));
		// wait for ack
		if (WaitUntilResponseType(KFID_AckFrame, 1000) != KFID_AckFrame)
        {
            FDataHandler->SendFieldSetFunctionMode(2, 0, 0x21);
            return false;
		}

    }

	// TODO function mode "0,0,0x20", if needed
	FDataHandler->SendFieldSetFunctionMode(2, 0, 0x21);
	if (WaitUntilResponseType(KFID_AckFrame, 10000) != KFID_AckFrame) return false;

	return true;
}

bool TDevice::CheckForShots()
{
	// get last shot count frame and return true if there are any shots to download
	FLastFrames.FShotCountFrameReceived = false;
	FDataHandler->SendFieldGetDeviceInfo(0x20);
	if (WaitUntilResponseType(KFID_FieldShotCountFrame, 1000) != KFID_FieldShotCountFrame)
		return false;

	int LCount = FLastFrames.FLastShotCountFrame.GetNumberOfShots();
	FDownloadableShotCount = LCount;

	return (LCount > 0);
}

bool TDevice::DownloadShot(int ABlock, int AShotIndex)
{
    ClearSampleBuffer();
    FDataHandler->SendFieldGetShot(ABlock, AShotIndex);
    bool LContWait = true;
	bool LIsOK = false;

    //
    while (LContWait && !LIsOK)
    {
        EFRAMEID_FrameTypes LType = WaitUntilResponse(1000);
        // sample frame receive, just wait for more
        if (LType == KFID_SampleFrameField) continue;
		// ack frame comes after samples, nothing more to receive
        // received shot is automatically added to shot list by receive "HandleShotFrame"
        if (LType == KFID_AckFrame) LIsOK = true;
        // nothing received, something went wrong.
        if (LType == KFID_None) LContWait = false;
        // allow update UI
        Application->ProcessMessages();
    }

    return LIsOK;
}

bool TDevice::DownloadAllShots()
{
    int LCount /* = FLastFrames.FLastShotCountFrame.GetNumberOfShots() */;
	//int LDownloaded = 0;
    bool isOK = false;
    
    for (int i=0;i<7;i++)
    {
        // check if supported mode
        int LMode = FLastFrames.FLastShotCountFrame.MemoryPages[i].BLOCK_FUNCTION_MODE;
        if (LMode == 2 || LMode == 3)
        {
			// check if shots present
            int LCount = FLastFrames.FLastShotCountFrame.MemoryPages[i].BLOCK_SHOT_COUNT;
            if (LCount > 0 && LCount != 0xff)
            {
                // download each shot
                for (int LIndex = 0; LIndex < LCount; LIndex++)
                {
					// download shot; 1 = first block index
                    isOK = DownloadShot(i,LIndex);
                    // let time to update UI
                    Application->ProcessMessages();
					//LDownloaded++;
                    if (!isOK)
                    {
                        Beep();
                        AnsiString str;
                        str.printf("Download fail: block %d shot %d", i, LIndex);
                    }
                }
            }
        }
    }

    return isOK;
}

bool TDevice::EraseShotsInAllBlocks()
{
	// goto programming mode
	FDataHandler->SendStartProgramming();
	if (WaitUntilResponseType(KFID_AckFrame, 10000) != KFID_AckFrame) return false;
	Application->ProcessMessages();

	// erase blocks 0..6
	for (int i=0;i<7;i++)
    {
        // if block not empty...
		//int LMode = FLastFrames.FLastShotCountFrame.MemoryPages[i].BLOCK_FUNCTION_MODE;
		//int LCount = FLastFrames.FLastShotCountFrame.MemoryPages[i].BLOCK_SHOT_COUNT;
		//if (LMode != 0xff && LCount > 0)
        {
			 FDataHandler->SendFieldEraseBlock(i);
			 if (WaitUntilResponseType(KFID_AckFrame, 2000) != KFID_AckFrame) ; // return false;
        }
	}

    // back to normal mode
	// TODO function mode "0,0,0x20", if needed
    FDataHandler->SendFieldSetFunctionMode(2, 0, 0x21);
    if (WaitUntilResponseType(KFID_AckFrame, 2000) != KFID_AckFrame) return false;

    return true;
}

bool TDevice::DownloadRequiredSilhouettesAndTargets()
{
	// go through the targets and check which targets are needed
	bool FNeedRing[16];
	bool FNeedSilhouette[16];

	memset(FNeedRing, 0, sizeof(FNeedRing));
	memset(FNeedSilhouette, 0, sizeof(FNeedSilhouette));

	for (int i=0; i < GetShotCount(); i++)
	{
		TShot& LShot = FShooter->GetShot(i);

		int LRing = LShot.FResultSample.RING_TARGET;
		int LSil = LShot.FResultSample.SILHOUETTE_TARGET;
		if (LRing < 16)
		{
			FNeedRing[LRing] = true;
		}

		if (LSil < 16)
		{
			FNeedSilhouette[LSil] = true;
		}
	}

	// go through rings and silhouettes and download them
	for (int i=0; i < 16; i++)
	{
		// this ring target is used.. and it's not downloaded...
		if (FNeedRing[i] && !FAllTargets.FRingTargetSet[i])
		{
			DownloadRingTarget(i);
		}
		if (FNeedSilhouette[i] && !FAllTargets.FSilhouetteSet[i])
		{
            DownloadSilhouetteTarget(i);
		}
	}

	return true;
}

bool TDevice::UploadPictureBitmap(int AType, int AToIndex, bool AGoProgramming = true)
{
	// go programming mode unless specifically requested not to
	if (AGoProgramming)
	{
		FDataHandler->SendStartProgramming();

		if (WaitUntilResponseType(KFID_AckFrame, 5000) != KFID_AckFrame)
		{
			return false;
		}
	}

	TCommandFrame_Field_UploadPictureBitmap LStruct;
	memset(&LStruct, 0, sizeof(LStruct));

	LStruct.COMMANDBYTE = CCMD_UploadPicture_Bitmap;
	LStruct.PictureNumber = AToIndex;

	// index 1: Bar view port, index 2: Square view port, index 3: Full view port
	// happens to match item indexes
	LStruct.PictureSizeGroup = AType;

	// which portion of the picture is being uploaded.. how many 8bit slices per row
	unsigned int LXCount = 2;
	switch(AType)
	{
		case 1: LXCount=2;break; // 16x48
		case 2: LXCount=6;break; // 48x48
		case 3: LXCount=8;break; // 64x48
	}
	// next piece to take the next 8
	unsigned int LRow = 0;
	unsigned int LCol = 0;

	// block doesn't match necessarily with rows so we go on independently from "row"
	// 48 rows, LXCount bytes per row, 16 byte per frame
	int LTotalCount = 48*LXCount/16;
	for (int LBlock = 0; LBlock < LTotalCount; LBlock++)
	{
		// how many slices of 8 pixels we take
		for (int LByteNumber = 0; LByteNumber<16; LByteNumber++)
		{
			//if(LRow>47) int idebug=0;

			// fill group of 8 pixels
			unsigned char LVal = ((bool)(FPicture[LCol][LRow])) << 7;
			LVal |= ((bool)(FPicture[LCol+1][LRow])) << 6;
			LVal |= ((bool)(FPicture[LCol+2][LRow])) << 5;
			LVal |= ((bool)(FPicture[LCol+3][LRow])) << 4;
			LVal |= ((bool)(FPicture[LCol+4][LRow])) << 3;
			LVal |= ((bool)(FPicture[LCol+5][LRow])) << 2;
			LVal |= ((bool)(FPicture[LCol+6][LRow])) << 1;
			LVal |= ((bool)(FPicture[LCol+7][LRow]));
			LStruct.DATA[LByteNumber] = LVal;

			LCol += 8;
			// check if it's time to pick next row
			if (LCol >= LXCount * 8)
			{
				LCol = 0;
				LRow++;
			}
		}

		// block's don't necessarily go with rows
		LStruct.SH1_FRAME_NUMBER = LBlock;
		// update check bit for last frame, not applicable at this moment
		//if (LRow==63) LStruct.UpdateCheckBit = 1;
		LStruct.CHECKBYTE = FDataHandler->GetChecksum(&LStruct, sizeof(LStruct)-1);
		FDataHandler->SendRaw(&LStruct, sizeof(LStruct));

		// we "skip" waiting for acks for each frame because we are lazy...
		if (LBlock < LTotalCount)
		{
			Sleep(100);
			continue;
		}

		// wait for ack
		if (WaitUntilResponseType(KFID_AckFrame, 3000) != KFID_AckFrame)
		{
			FDataHandler->SendFieldSetFunctionMode(2, 0, 0x21);
			return false;
		}
	}

	// If went to programming mode, leave
	if (AGoProgramming)
	{
		StopMeasuring();
		Sleep(100);
	}

	return true;
}

void TDevice::ShowMessage(UnicodeString AMessage)
{
    FLastFrames.Reset();

	FDataHandler->BreakOnNextReceive();
	FDataHandler->SendShowMessage(AMessage);
	FLastFrames;
	int i = 0;
}


