//---------------------------------------------------------------------------
// FramesAndSets.h
//
// The parameter sets and control frames of Expert device are defined in this file
//
// Noptel/FN Expert (Field), NOS Pro
// C. Noptel Oy, Oulu Finland, 2009-2016
//
// Version: 2.2
// Reference: Interface Control Document (ICD),
//            Noptel doc#: OS42944AE v.1.0
//---------------------------------------------------------------------------

#ifndef ParameterSetsFramesH
#define ParameterSetsFramesH

#include <vcl.h>

// The 'Structs' are defined as a union to allow indexed byte access
// to data elements. Also, each 'bit' within byte can be accessed
// with the its name.

// Types and structs start with CAPITAL T-LETTER

// FULL BYTES IN STRUCTS ARE WRITTEN IN CAPITAL LETTERS
// Individual bit information inside BYTE are in lower case
// RAW_BYTE is in union with the actual struct, allows indexed byte access

// Note that individual bits can be accessed:
// - directly, for example Store_to_local_memory
// - through CONTROL_BITS & 0x01
// - through RAW_BYTE[0] & 0x01

// The structs are presented in following order:
// - Parameter sets stored in optical unit's non-volatile memory
// - Command frames sent to optical unit
// - Response frames received from optical unit

extern unsigned char GetHexAsInt(const AnsiString& AHex8bit);
extern AnsiString GetIntAsHex(int ANumber);
extern bool CheckAndGetHexValueToByte(char* AHex, unsigned char& AWhereTo); // In UnitParameterSetForm.cpp

class TAppController;
extern TAppController* GAppController; // the "unique" app controller. Defined in AppController.cpp

extern const double ONE_NAU_to_mm_in_one_meter; // one NAU = 1mm in 300m (Noptel Angle Unit)

// introduced in device.cpp
extern double XChannelFactor;
extern double YChannelFactor;

extern unsigned short GetU16FromMSBLSB(char* AMSB);
extern short GetS16FromMSBLSB(char* AMSB);
extern void Set16ToMSBLSB(char* APMSB, int Any16);

// ** Parameter sets
// NOTE: Reserved Bytes are numbered as in ICD doc i.e. 1->16

struct TParameterSet_Field_SilhouetteSettings
{
 union
 {
  struct
  {
    // Byte2, CONTROL_BITS. Defined also as individual bits.
   	union
    {
		unsigned char CTRL_BYTE;
		struct
		{
			unsigned char Direction_bit0:1; // 1 = +90
			unsigned char Direction_bit1:1; // 1 = +180, both enabled (+270)
			unsigned char Reserved_1_bit2:1;
			unsigned char Scalable:1; // 1 = enable
			unsigned char DistanceUnit_01yd:1;
			unsigned char Reserved_1_bit5:1;
			unsigned char SizeImperial:1;
			unsigned char Reserved_1_bit7:1;
		};
	};

	union
	{
		unsigned char CTRL_BYTE_2;
		struct
		{
			unsigned char ButtonSelectable:1;
			unsigned char Reserved1:1;
			unsigned char Reserved2:1;
			unsigned char Reserved3:1;
			unsigned char Reserved4:1;
			unsigned char Reserved5:1;
			unsigned char Reserved6:1;
			unsigned char Reserved7:1;
		};
	};


	unsigned char NOMINAL_DISTANCE_MSB;
    unsigned char NOMINAL_DISTANCE_LSB;
    unsigned char RESERVED_4;
    unsigned char RESERVED_5;
    unsigned char RESERVED_6;
    unsigned char RESERVED_7;
    unsigned char RESERVED_8;
    unsigned char RESERVED_9;
	unsigned char SILHOUETTE_SIZE_MSB;  // mm
    unsigned char SILHOUETTE_SIZE_LSB;  // mm
    unsigned char PRISM_PLACE_X_MSB; // mm
    unsigned char PRISM_PLACE_X_LSB; // mm
    unsigned char PRISM_PLACE_Y_MSB; // mm
    unsigned char PRISM_PLACE_Y_LSB; // mm
  };
  unsigned char RAW_BYTE[16]; // The array can be accessed without char* cast using this
 };

 unsigned short GetSilhouetteSize_mm()
 {
     // adjust to unit if necessary
     return GetU16FromMSBLSB(&SILHOUETTE_SIZE_MSB) * (SizeImperial ? 2.54 : 1.0);
 }

 unsigned short GetSilhouetteSizeRaw() { return GetU16FromMSBLSB(&SILHOUETTE_SIZE_MSB); }
 unsigned short GetNominalDistance_raw() { return GetU16FromMSBLSB(&NOMINAL_DISTANCE_MSB); }
 double GetNominalDistance_m() { return GetNominalDistance_raw() * ( DistanceUnit_01yd ? 0.09144 : 0.1 ) ; }
 short GetPrismPlaceX() { return GetS16FromMSBLSB(&PRISM_PLACE_X_MSB); }
 short GetPrismPlaceY() { return GetS16FromMSBLSB(&PRISM_PLACE_Y_MSB); }

 void SetSilhouetteSizeRaw(unsigned short A) { Set16ToMSBLSB(&SILHOUETTE_SIZE_MSB, A); }
 void SetNominalDistance(unsigned short A) { Set16ToMSBLSB(&NOMINAL_DISTANCE_MSB, A); }
 void SetPrismPlaceX(short A) { Set16ToMSBLSB(&PRISM_PLACE_X_MSB, A); }
 void SetPrismPlaceY(short A) { Set16ToMSBLSB(&PRISM_PLACE_Y_MSB, A); }
};

// parameter set 1 actually contains 2 of these: for classic and another for dynamic
struct TModeSettings
{
	//TModeSettings() {memset(this,0,sizeof(TModeSettings));}
	union
	{
		unsigned char CONTROL_BYTE_1; // 0
		struct
		{
			unsigned char EvaluateSilhouette:1;
			unsigned char CountRingScore:1;
			unsigned char RegisterAllTriggers:1;
			unsigned char TargetReceiverHitIndication:1;
			unsigned char SleepWakeup:1;
			unsigned char Reserved5:1;
			unsigned char Reserved6:1;
			unsigned char Reserved7:1;
		};
	};

	union
	{
		unsigned char CONTROL_BYTE_2; // 1
		struct
		{
			unsigned char GraphicsDisplayEnable:1;
			unsigned char Reserved2_1:1;
			unsigned char Reserved2_2:1;
			unsigned char Reserved2_3:1;
			unsigned char Reserved2_4:1;
			unsigned char HitSoundEnable:1;
			unsigned char Reserved2_6:1;
			unsigned char Reserved2_7:1;
		};
	};

	unsigned char RING_TARGET_NUMBER; // 2
	unsigned char SILH_TARGET_NUMBER; // 3
	unsigned char NUMBER_OF_PRISMS; // 4 : 1,2,4,16
	unsigned char NUMBER_OF_BULLETS; // 5, per session, 0 = unlimited
	unsigned char SLEEP_WAKEUP_ANGLE; // 6, degrees, 5-90
	unsigned char SHOT_REJECTION_ANGLE; // 7, degrees, 5-90
	unsigned char SLEEP_TIMEOUT; // 8, seconds
	unsigned char TRIGGER_SENSITIVITY; // 9 : (10mV)
	unsigned char COCKING_DELAY; // 10, x 0.1s
	unsigned char RESERVED[5]; // 11..15
};

// Parameter set 1. Use scenario. 64 bytes.
struct TParameterSet_Field_1
{
 TParameterSet_Field_1(){memset(this,0,sizeof(TParameterSet_Field_1));}
 union
 {
  struct
  {
    // Byte2, CONTROL_BITS. Defined also as individual bits.
   	union
    {
		unsigned char CONTROL_BYTE_1;  // 0
	    struct
        {
    		unsigned char DeviceMounting:2; // 00=top,11=bottom,10=left,01=right
			unsigned char Reserved_1_bit2:1; // 1=enable
			unsigned char Reserved_1_bit3:1; // 1=enable
    		unsigned char Arrow_buttons:1; // 1 = enable
	    	unsigned char Reserved_1_bit5:1;
    		unsigned char Reserved_1_bit6:1;
	    	unsigned char Reserved_1_bit7:1;
        };
	};

    // Byte2, CONTROL_BITS. Defined also as individual bits.
   	union
    {
		unsigned char CONTROL_BYTE_2;  // 1
	    struct
        {
			unsigned char Reserved_2_bit0:1;
			unsigned char Reserved_2_bit1:1;
			unsigned char Reserved_2_bit2:1;
			unsigned char Reserved_2_bit3:1;
    		unsigned char Reserved_2_bit4:1; 
	    	unsigned char Reserved_2_bit5:1;
			unsigned char Reserved_2_bit6:1;
	    	unsigned char Reserved_2_bit7:1;
        };
	};

	unsigned char STARTUP_FUNCTION_MODE; // 2
	unsigned char RESERVED_3_11[9]; // 3..11
	unsigned char ZEROING_X_MSB; // 12
	unsigned char ZEROING_X_LSB; // 13
	unsigned char ZEROING_Y_MSB; // 14
	unsigned char ZEROING_Y_LSB; // 15

	union
	{
		TModeSettings CLASSIC_MODE_SETTINGS;
		struct
		{
			unsigned char CLASSIC_CONTROL_BYTE_1; // 16
			unsigned char CLASSIC_CONTROL_BYTE_2; // 17
			unsigned char CLA_RING_TARGET_NUMBER; // 18
			unsigned char CLA_SILH_TARGET_NUMBER; // 19
			unsigned char CLA_NUMBER_OF_PRISMS; // 20 : 1,2,4,16
			unsigned char CLA_NUMBER_OF_BULLETS; // 21, per session, 0 = unlimited
			unsigned char CLA_SLEEP_WAKEUP_ANGLE; // 22, degrees, 5-90
			unsigned char CLA_SHOT_REJECTION_ANGLE; // 23, degrees, 5-90
			unsigned char CLA_SLEEP_TIMEOUT; // 24, seconds
			unsigned char CLA_TRIGGER_SENSITIVITY; // 25 : (10mV)
			unsigned char CLA_COCKING_DELAY; // 26, x 0.1s
			unsigned char CLA_RESERVED_27_31[5]; // 27..31
		};
	};

	union
	{
		TModeSettings DYNAMIC_MODE_SETTINGS;
		struct
		{
			unsigned char DYNAMIC_CONTROL_BYTE_1; // 32
			unsigned char DYNAMIC_CONTROL_BYTE_2; // 33
			unsigned char DYN_RING_TARGET_NUMBER; // 34
			unsigned char DYN_SILH_TARGET_NUMBER; // 35
			unsigned char DYN_NUMBER_OF_PRISMS; // 36 : 1,2,4,16
			unsigned char DYN_NUMBER_OF_BULLETS; // 37, per session, 0 = unlimited
			unsigned char DYN_SLEEP_WAKEUP_ANGLE; // 38, degrees, 5-90
			unsigned char DYN_SHOT_REJECTION_ANGLE; // 39, degrees, 5-90
			unsigned char DYN_SLEEP_TIMEOUT; // 40, seconds
			unsigned char DYN_TRIGGER_SENSITIVITY; // 41 : (10mV)
			unsigned char DYN_COCKING_DELAY; // 42, x 0.1s
			unsigned char DYN_RESERVED_43_47[5]; // 43..47
		};
	};

	unsigned char RESERVED_48_63[16];
  };
  unsigned char RAW_BYTE[64]; // The array can be accessed without char* cast using this
 };

 short GetZeroingX(){ return (short)((ZEROING_X_MSB << 8) | ZEROING_X_LSB);}
 short GetZeroingY(){ return (short)((ZEROING_Y_MSB << 8) | ZEROING_Y_LSB);}
 void SetZeroingX(short A) { ZEROING_X_MSB = (A & 0xff00) >> 8; ZEROING_X_LSB = A & 0x00ff; }
 void SetZeroingY(short A) { ZEROING_Y_MSB = (A & 0xff00) >> 8; ZEROING_Y_LSB = A & 0x00ff; }
};

// Parameter set 2. Technical settings. 16 bytes.
struct TParameterSet_Field_2
{
 TParameterSet_Field_2(){memset(this,0,sizeof(TParameterSet_Field_2));}
 union
 {
  struct
  {
    // Byte1, CONTROL_BITS. Defined also as individual bits.
   	union
    {
    	unsigned char PRM2_CTRL_BYTE_1;
	    struct
		{
    		unsigned char Local_summing_mode:1; // 1=enable
			unsigned char Store_tracking:1; // 1=enable
			unsigned char TX_power_reduction:1; // 1=enable
			unsigned char Range_evaluation_correction_table:1; // 1=enable
			unsigned char Range_evaluation:1; // 1=enable
			unsigned char Angle_measurements:1; // 1=enable
			unsigned char Background_reflection_check:1; // 1=enable
			unsigned char Power_monitor:1; // 1=enable
		};
	};

	// Byte2, CONTROL_BITS. Defined also as individual bits.
	union
	{
		unsigned char PRM2_CTRL_BYTE_2;
		struct
		{
			unsigned char Reserved_2_bit0:1; // 1=enable
			unsigned char Reserved_2_bit1:1;
			unsigned char Reserved_2_bit2:1;
			unsigned char Reserved_2_bit3:1;
			unsigned char Reserved_2_bit4:1;
			unsigned char Reserved_2_bit5:1;
    		unsigned char Reserved_2_bit6:1;
	    	unsigned char Reserved_2_bit7:1;
        };
	};

    unsigned char ZEROING_SOUND_LIMIT;          // 3, "unit id", 0-255
    unsigned char AIMING_SAMPLES;       // 4, stored aim samples, 0-255
    unsigned char ACC_CHECK_INTERVAL;   // 5, angle detector interval, x 0.1s, 1-255
    unsigned char SHOT_DETECTION_DELAY;   // 6, samples, 1-16
    unsigned char FOLLOW_THROUGH_SAMPLES; // 7,
    unsigned char SIGNAL_LEVEL_VALIDITY; // 8, high ADC, sample max level, x10, 10-2550
    unsigned char RESERVED_9;
    unsigned char RESERVED_10;
    unsigned char RESERVED_11;
	unsigned char RESERVED_12;
    unsigned char RESERVED_13;
    unsigned char RESERVED_14;
    unsigned char RESERVED_15;
    unsigned char RESERVED_16;

	unsigned char RESERVED_17_32[16];
	unsigned char RESERVED_33_64[32];
  };
  unsigned char RAW_BYTE[64]; // The array can be accessed without char* cast using this
 };
};

// Parameter set 3. R&D or factory settings. 64 bytes.
// struct TParameterSet_Field_3
// 16.5.2016 MHi Structure removed, because it's not used in this application.

// Parameter set for device serial number
struct TParameterSet_Field_4
{
 TParameterSet_Field_4(){memset(this,0,sizeof(TParameterSet_Field_4));}
 union
 {
  struct
  {
	unsigned char SERIAL_NUMBER[10];
	unsigned char RESERVED_11;
	unsigned char RESERVED_12;
	unsigned char RESERVED_13;
	unsigned char RESERVED_14;
	unsigned char RESERVED_15;
	unsigned char RESERVED_16;
	unsigned char RESERVED_17_64[48];
  };
  unsigned char RAW_BYTE[64]; // The array can be accessed without char* cast using this
 };
 AnsiString GetSerialNumber() { AnsiString x; x.printf("%10c", SERIAL_NUMBER); return x;}
};


// Parameter set for discipline. Ring target settings. 16 bytes.
struct TParameterSet_Field_RingTarget
{
 TParameterSet_Field_RingTarget(){memset(this,0,sizeof(TParameterSet_Field_RingTarget));}
 union
 {
  struct
  {
    // Byte1, CONTROL_BITS. Defined also as individual bits.
   	union
    {
    	unsigned char CTRL_BYTE_1;
	    struct
        {
    		unsigned char Gun_type:2; // 0 = pistol, 1 = rifle, 23 reserved
			unsigned char Decimal_scores:1;
			unsigned char Reserved_bit3:1;
	    	unsigned char DistanceUnit_01yd:1; // 0 = 0.1m, 1 = 1ft
    		unsigned char CaliberUnit_1per1000inch:1; // 0 = 0.01mm, 1 = 0.001"
    		unsigned char RingUnit_1per100inch:1; // 0 = 0.1mm, 1 = 0.01"
    		unsigned char VariableRingGap:1; // 0=Fixed, 1=Use table
        };
	};

	union
	{
		unsigned char CTRL_BYTE_2;
		struct
		{
			unsigned char ButtonSelectable:1;
			unsigned char Reserved1:1;
			unsigned char Reserved2:1;
			unsigned char Reserved3:1;
			unsigned char Reserved4:1;
			unsigned char Reserved5:1;
			unsigned char Reserved6:1;
			unsigned char Reserved7:1;
		};
	};

	unsigned char NOMINAL_SHOOTING_DISTANCE_MSB; // 2: each 5m
    unsigned char NOMINAL_SHOOTING_DISTANCE_LSB; // 3: each 5m
    unsigned char CALIBER_MSB_BYTE; // 4.
    unsigned char CALIBER_LSB_BYTE; // 5: low 8 bits on caliber,
    unsigned char HIGHEST_HIT_RING;
    unsigned char MINIMUM_HIT_RING;
	unsigned char HIB_MIN_SCORE;   // x0.1
	unsigned char RESERVED9;
    unsigned char TARGET_WIDTH;  // 1cm / 1"
    unsigned char TARGET_HEIGHT; // 1cm / 1"
    unsigned char PRISM_PLACE_X_MSB; // ring gap unit
    unsigned char PRISM_PLACE_X_LSB; //
    unsigned char PRISM_PLACE_Y_MSB; //
    unsigned char PRISM_PLACE_Y_LSB; //
  };
  unsigned char RAW_BYTE[16]; // The array can be accessed without char* cast using this
 };

 short GetCaliber_raw() { return (short)( (CALIBER_MSB_BYTE<<8) | CALIBER_LSB_BYTE);}
 void SetCaliber_raw(short AVal) { CALIBER_MSB_BYTE = (AVal & 0xff00) >> 8; CALIBER_LSB_BYTE = AVal & 0xff;}
 double GetCaliber_mm() { return GetCaliber_raw() * ( CaliberUnit_1per1000inch ? 25.4 * 0.001 : 0.01 ); }
 void SetCaliber_mm(double AVal) { if (CaliberUnit_1per1000inch) SetCaliber_raw( AVal / (25.4 * 0.001) ); else SetCaliber_raw( AVal * 100 ); }

 void SetNominalDistance_raw(short AVal) { Set16ToMSBLSB(&NOMINAL_SHOOTING_DISTANCE_MSB, AVal); }
 unsigned short GetNominalDistance_raw() { return GetU16FromMSBLSB(&NOMINAL_SHOOTING_DISTANCE_MSB); }
 double GetNominalDistance_m() { return GetNominalDistance_raw() * ( DistanceUnit_01yd ? 0.09144 : 0.1 ) ; }
 double GetNominalDistance_m_calc() { return GetNominalDistance_m() - 0.75 ; }
 unsigned short GetPrismPlaceX_raw() { return GetU16FromMSBLSB(&PRISM_PLACE_X_MSB); }
 unsigned short GetPrismPlaceY_raw() { return GetU16FromMSBLSB(&PRISM_PLACE_Y_MSB); }

};

// Parameter set for discipline. Ring target settings. 16 bytes.
struct TParameterSet_Field_RingTarget_block2
{
 TParameterSet_Field_RingTarget_block2(){memset(this,0,sizeof(TParameterSet_Field_RingTarget_block2));}
 union
 {
  struct
  {
	unsigned char WHITE_INNER_DIAMETER_MSB; // pair 0
    unsigned char WHITE_INNER_DIAMETER_LSB;
    unsigned char BLACK_DIAMETER_MSB; // pair 1
    unsigned char BLACK_DIAMETER_LSB;
    unsigned char HIT_RING_GAP_MSB; // pair 2
    unsigned char HIT_RING_GAP_LSB;
    unsigned char X_DIAMETER_MSB; // pair 3
    unsigned char X_DIAMETER_LSB;
    unsigned char HIGHEST_RING_MSB; // pair 4
    unsigned char HIGHEST_RING_LSB;
	unsigned char RING_H1_DIAMETER_MSB; // pair 5
	unsigned char RING_H1_DIAMETER_LSB;
	unsigned char RING_H2_DIAMETER_MSB; // pair 6
	unsigned char RING_H2_DIAMETER_LSB;
	unsigned char RING_H3_DIAMETER_MSB; // pair 7
    unsigned char RING_H3_DIAMETER_LSB;
  };
  unsigned char RAW_BYTE[16]; // The array can be accessed without char* cast using this
 };
 
 unsigned short GetHighestRingDiameter() { return GetU16FromMSBLSB(&HIGHEST_RING_MSB); }
 unsigned short GetXDiameter() { return GetU16FromMSBLSB(&X_DIAMETER_MSB); }
 unsigned short GetHitRingGap() { return GetU16FromMSBLSB(&HIT_RING_GAP_MSB); }
 unsigned short GetBlackDiameter() { return GetU16FromMSBLSB(&BLACK_DIAMETER_MSB); }
 unsigned short GetInnerWhiteDiameter() { return GetU16FromMSBLSB(&WHITE_INNER_DIAMETER_MSB); }
 unsigned short GetRing9to7Diameter(int A) { return Get16bIndex( 14-A ); } // TODO: remove this
 unsigned short GetHighestMinusRings(int ARing0to3, int AValue) { return Get16bIndex(4+ARing0to3); }
 void SetHighestMinusRings(ARing0to3, int AValue) { Set16ToMSBLSB( &HIGHEST_RING_MSB + 2*ARing0to3, AValue ); }

 unsigned short Get16bIndex(int A) { return GetU16FromMSBLSB((char *)&RAW_BYTE[2*A]); }
 void Set16bIndex(int ANum, short AValue) { Set16ToMSBLSB(&RAW_BYTE[2*ANum], AValue); }

 void SetInnerWhiteDiameter(int A) { Set16ToMSBLSB(&WHITE_INNER_DIAMETER_MSB, A); }
 void SetBlackDiameter(int A) { Set16ToMSBLSB(&BLACK_DIAMETER_MSB, A); }
 void SetXDiameter(int A) { Set16ToMSBLSB(&X_DIAMETER_MSB, A); }
 void SetHighestRingDiameter(int A) { Set16ToMSBLSB(&HIGHEST_RING_MSB, A); }
 void SetHitRingGap(int A) { Set16ToMSBLSB(&HIT_RING_GAP_MSB, A); }
};

// Parameter set for discipline. Ring target settings. 16 bytes.
struct TParameterSet_Field_RingTarget_block3
{
 TParameterSet_Field_RingTarget_block3(){memset(this,0,sizeof(TParameterSet_Field_RingTarget_block3));}
 union
 {
  struct
  {
	unsigned char RING_H4_DIAMETER_MSB; // pair 0
	unsigned char RING_H4_DIAMETER_LSB;
	unsigned char RING_H5_DIAMETER_MSB; // pair 1
	unsigned char RING_H5_DIAMETER_LSB;
	unsigned char RING_H6_DIAMETER_MSB; // pair 2
	unsigned char RING_H6_DIAMETER_LSB;
	unsigned char RING_H7_DIAMETER_MSB; // pair 3
	unsigned char RING_H7_DIAMETER_LSB;
	unsigned char RING_H8_DIAMETER_MSB; // pair 4
	unsigned char RING_H8_DIAMETER_LSB;
	unsigned char RING_H9_DIAMETER_MSB; // pair 5
	unsigned char RING_H9_DIAMETER_LSB;
	unsigned char RING_H10_DIAMETER_MSB; // pair 6
	unsigned char RING_H10_DIAMETER_LSB;
	unsigned char RING_H11_DIAMETER_MSB; // pair 7
	unsigned char RING_H11_DIAMETER_LSB;
  };
  unsigned char RAW_BYTE[16]; // The array can be accessed without char* cast using this
 };

 unsigned short GetRing6tom1Diameter(int A) { return Get16bIndex( 6-A ); }
 unsigned short GetHighestMinusRings(int ARing4to11) { return Get16bIndex( ARing4to11-4 ); }

 void SetHighestMinusRings(ARing4to11, int AValue) { Set16ToMSBLSB( &RING_H4_DIAMETER_MSB + 2*(ARing4to11-4), AValue ); }

 unsigned short Get16bIndex(int A) { return GetU16FromMSBLSB((char *)&RAW_BYTE[2*A]); }
 void Set16bIndex(int ANum, short AValue) { Set16ToMSBLSB(&RAW_BYTE[2*ANum], AValue); }
};

struct TParameterSet_Any16bytes
{
 TParameterSet_Any16bytes(){memset(this,0,sizeof(TParameterSet_Any16bytes));}
 unsigned char RAW_BYTE[16];
 void SetString(AnsiString AStr){memset(RAW_BYTE,0,sizeof(RAW_BYTE));int a = AStr.Length(); a = a > 16 ? 16 : a; memcpy(RAW_BYTE,AStr.c_str(), a);}
 AnsiString GetString() { AnsiString A;for (int i=0;i<16 && RAW_BYTE[i] ;i++) A.cat_printf("%c", RAW_BYTE[i]); return A; }
};

struct TRingTarget
{
    TRingTarget() { memset(this,0,sizeof(TRingTarget)); }
    TParameterSet_Any16bytes FTargetName;
    TParameterSet_Field_RingTarget FRingSet1;
	TParameterSet_Field_RingTarget_block2 FRingSet2;
	TParameterSet_Field_RingTarget_block3 FRingSet3;

    unsigned short GetCustomRingDiameterRaw(int A)
    {
        // highest than highest ring -> highest
		// between -> check whether in 3 down of highest or even more
        // lower than lowest -> lowest
		if (A >= FRingSet1.HIGHEST_HIT_RING)
			return FRingSet2.GetHighestRingDiameter();
		if (A >= FRingSet1.HIGHEST_HIT_RING - 3)
			return FRingSet2.GetRing9to7Diameter(10 - (FRingSet1.HIGHEST_HIT_RING - A) );
		if (A >= FRingSet1.HIGHEST_HIT_RING - 11)
            return FRingSet3.GetRing6tom1Diameter(10 - (FRingSet1.HIGHEST_HIT_RING - A) );

		return FRingSet3.GetRing6tom1Diameter(FRingSet1.MINIMUM_HIT_RING);
	}

	double GetRadiusForRingmm(double ARing)
	{
		TParameterSet_Field_RingTarget& LRS1 = FRingSet1;
		TParameterSet_Field_RingTarget_block2& LRS2 = FRingSet2;
		//TParameterSet_Field_RingTarget_block3& LRS3 = FRingSet3;

		// make this class-wide
		double LUnit = LRS1.RingUnit_1per100inch ? 0.01*25.4 : 0.1; // 0.01" / 0.1mm

		// pick value to this
		double LVal = 0;

		if (!LRS1.VariableRingGap)
		{
			LVal = 0.5 * LRS2.GetHighestRingDiameter() * LUnit;

			if (ARing >= LRS1.HIGHEST_HIT_RING)
			{
				// highest ring is default
			}
			else
			{
				// how many ring gap down we need to go
				LVal += (LRS1.HIGHEST_HIT_RING - ARing) * LRS2.GetHitRingGap() * LUnit;
			}
		}
		else
		{
			// diameter to radius, adjust to unit
			LVal = 0.5 * LUnit * GetCustomRingDiameterRaw(ARing);
		}

    return LVal;
}


	double GetHitRingGap_mm()
	{
		return ((double)FRingSet2.GetHitRingGap() ) * (FRingSet1.RingUnit_1per100inch ? 0.254 : .1);
	}

};

// hold sets that define a silhouette target, except the actual bit mask
struct TSilhouetteTarget
{
    TSilhouetteTarget() { memset(this,0,sizeof(TSilhouetteTarget)); }

    TParameterSet_Any16bytes FTargetName;
	TParameterSet_Field_SilhouetteSettings FSilhouetteSet;

	double FWidth;  // from file
	double FHeight; // from file
};

struct TSilhouetteBitmapMatrix
{
	TSilhouetteBitmapMatrix() { memset(this, 0, sizeof(TSilhouetteBitmapMatrix)); }
	int X,Y;
	// at this time we use only 2 bits per pixel.
	// could optimize by creating 2bit data type or dual bool array
	// maybe create getter setter to hide array type
    char FPixelState[64][64];
};

// ** Target definition
struct TTargetDefinition
{
	int FTargetNumber; // in file/device
	bool FUsingSilhouette;
	bool FUsingRing; // both true -> ring is drawn on silhouette

	TRingTarget FRingTarget;
	TSilhouetteTarget FSilTarget;
	TSilhouetteBitmapMatrix FSilhouette;

	AnsiString GetTargetName() { if (FUsingRing) return FRingTarget.FTargetName.GetString(); else return FSilTarget.FTargetName.GetString(); }

	double GetDist_m() { return FUsingRing ? FRingTarget.FRingSet1.GetNominalDistance_m() : FSilTarget.FSilhouetteSet.GetNominalDistance_m(); }

	double GetWidth_mm()
	{
		double LRingW = FUsingRing ? 2*FRingTarget.GetRadiusForRingmm(FRingTarget.FRingSet1.MINIMUM_HIT_RING) : 0;
		double LSilW = FUsingSilhouette ? FSilTarget.FSilhouetteSet.GetSilhouetteSize_mm() : 0;
		return LRingW > LSilW ? LRingW : LSilW;		
	}
};

// This works if just swaps are done (pointers stay, no duplication)
// (Copy the constructor IF needed to make copies of FDescription or FCustomSettings)
struct TTargetSetup
{
	// construct, destruct
	TTargetSetup(){ memset(this, 0, sizeof(TTargetSetup));}
	~TTargetSetup(){ if (FCustomSettings) delete FCustomSettings; if (FDescription) delete FDescription; }

	// copy constructor
	TTargetSetup(const TTargetSetup& A)
	{
		memset(this, 0, sizeof(TTargetSetup));
		CopyContentsFrom(A);
	}

	void CopyContentsFrom(const TTargetSetup& A)
	{
		if (FDescription) FDescription->Clear();
		if (FCustomSettings) FCustomSettings->Clear();

		if (A.FDescription)
		{
			if (!FDescription) FDescription = new TStringList;
			FDescription->AddStrings(A.FDescription);
		}
		if (A.FCustomSettings)
		{
			if (!FCustomSettings) FCustomSettings = new TStringList;
			FCustomSettings->AddStrings(A.FCustomSettings);            
		}
		FTargetDef = A.FTargetDef;
		FUseScenario = A.FUseScenario;
	}

	// description to be shown in setup list
	TStringList* FDescription;
	// is this in use
	bool FIsDefined;
	// information about the target
	TTargetDefinition FTargetDef;
	// use scenario settings
	TParameterSet_Field_1 FUseScenario;
	// custom settings
	TStringList* FCustomSettings;
};


// this contains structs for 16 both ring and silhouette targets
// could implement getter/setter to dynamically allocate space
struct TAllTargets
{
	TAllTargets() {memset(this, 0, sizeof(TAllTargets));};

	// targets and silhouette infos
	// TRingTarget contains all 4 sets for each ring target
	TRingTarget FFixedRingTarget;
	TRingTarget FRingTargets[16];
	bool FRingTargetSet[16];

	TSilhouetteBitmapMatrix FSilhouetteBitmaps[16];
	TSilhouetteTarget FSilhouetteTargets[16];
	bool FSilhouetteSet[16];

	TRingTarget& GetRingTarget(int AIndex) { return FRingTargets[AIndex]; };
	void SetRingTarget(int AIndex, TRingTarget& ATarget) { memcpy(&FRingTargets[AIndex], &ATarget, sizeof(TRingTarget)); };

	TSilhouetteTarget& GetSilhouetteTarget(int AIndex) { return FSilhouetteTargets[AIndex]; };
	TSilhouetteBitmapMatrix& GetSilhouetteBitmapMatrix(int AIndex) { return FSilhouetteBitmaps[AIndex]; };
	void SetSilhouetteTarget(int AIndex, TSilhouetteTarget& ATarget) { memcpy(&FSilhouetteTargets[AIndex], &ATarget, sizeof(TSilhouetteTarget)); };
	void SetSilhouetteBitmapMatrix(int AIndex, TSilhouetteBitmapMatrix& ATarget) { memcpy(&FSilhouetteBitmaps[AIndex], &ATarget, sizeof(TSilhouetteBitmapMatrix)); };
};

//***************************

// typedefs for numbered sets
typedef struct TParameterSet_Field_1 TParameterSet_Field_UserScenario;

// ** Command frames sent to optical unit

// Actual command bytes, send to optical unit

const unsigned char CCMD_UploadPicture_Bitmap = 0x6A;
const unsigned char CCMD_UploadSilhouette_Bitmap = 0x66;
const unsigned char CCMD_DownloadSilhouette_Bitmap = 0x67;
const unsigned char CCMD_FieldEnterFunctionMode = 0x61;
const unsigned char CCMD_FieldGetDeviceInfo = 0x63;
const unsigned char CCMD_FieldGetParameterSet = 0x65;
const unsigned char CCMD_FieldSetParameterSet = 0x64;
const unsigned char CCMD_FieldGetShot = 0x62;
const unsigned char CCMD_FieldGetRingTarget = 0x6D;
const unsigned char CCMD_FieldSetRingTarget = 0x6C;
const unsigned char CCMD_FieldShowMessage = 0x73;

// ** Response frames received from optical unit

// Received frame identifiers;
//   const ending _0/1 means the frame ID is two-byte
//     _0 being first byte and _1 being second byte
//     NOTE that AckFrame and ErrorFrame define only first byte;
//     it is up to frame handler to detect is frame is actually one of subset frames (ParameterFrame is an AckFrame)

const unsigned char CFRAMEID_AckFrame = 0x59;
const unsigned char CFRAMEID_SampleFrameField_0 = 0xCA;
const unsigned char CFRAMEID_SampleFrameField_1 = 0x65;
const unsigned char CFRAMEID_ResultFrameField_1 = 0x6A;
const unsigned char CFRAMEID_InfoErrorFrame_1 = 0x63;
const unsigned char CFRAMEID_Field_ParameterSet_0 = 0x59;
const unsigned char CFRAMEID_Field_ParameterSet_1 = 0x65;
const unsigned char CFRAMEID_Field_DeviceInfo_0 = 0x59;
const unsigned char CFRAMEID_Field_DeviceInfo_1 = 0x63;
const unsigned char CFRAMEID_Field_DownloadSilhouette_0 = 0x59;
const unsigned char CFRAMEID_Field_DownloadSilhouette_1 = 0x67;
const unsigned char CFRAMEID_Field_GetRingTarget_1 = 0x6D;


struct TCommandFrame_Field_UploadSilhouetteBitmap
{
	unsigned char COMMANDBYTE;
  	union
    {
        unsigned char SH1_FRAME_CONTROL;   // which silhouette
        struct
        {
        	unsigned char SilhouetteNumber : 6; // 0..63, check mapping with sets
            unsigned char BitmaskFrame : 1;  // 1 = use the silhouette info struct
            unsigned char Reserved : 1;  // 1 = update device internal settings check byte
        };
    };
    union
    {
        unsigned char SH1_FRAME_NUMBER;    // which row in silhouette/byte block
        struct
        {
            unsigned char BlockNumber : 7;
            unsigned char UpdateCheckBit : 1;
        };
        struct
        {
            unsigned char SilhouetteTargetNameBlock : 1;
            unsigned char ReservedBits : 6;
            unsigned char UpdateCheckBit_ : 1;
        };
    };
    unsigned char DATA[16];
    unsigned char CHECKBYTE;
};

struct TCommandFrame_Field_UploadPictureBitmap
{
	unsigned char COMMANDBYTE;
  	union
    {
        unsigned char SH1_FRAME_CONTROL;   // which picture
        struct
        {
        	unsigned char PictureNumber : 6; // 0..63, check mapping with sets
            unsigned char PictureSizeGroup : 2;  // 1 = update device internal settings check byte
        };
    };
    unsigned char SH1_FRAME_NUMBER;    // which row in silhouette/byte block
    unsigned char DATA[16];
    unsigned char CHECKBYTE;
};

struct TCommandFrame_Field_DownloadSilhouetteBitmap
{
	unsigned char COMMANDBYTE;
  	union
    {
        unsigned char SH1_FRAME_CONTROL;   // which silhouette
        struct
        {
        	unsigned char SilhouetteNumber : 6; // 0..63, check mapping with sets
            unsigned char BitmaskFrame : 1;  // 1 = get bitmask of silhouette, 0 = get generic info 
            unsigned char Reserved : 1;  // 1 = update device internal settings check byte
        };
    };
    union
    {
        unsigned char SH1_FRAME_NUMBER;    // which row in silhouette/byte block
        struct
        {
            unsigned char BlockNumber : 7;
            unsigned char UpdateCheckBit : 1;
        };
        struct
        {
            unsigned char SilhouetteTargetNameBlock : 1;
            unsigned char ReservedBits : 6;
            unsigned char UpdateCheckBit_ : 1;
        };
    };
    unsigned char CHECKBYTE;
};

struct TCommandFrame_Field_ShowMessage
{
	unsigned char COMMANDBYTE;
	union
	{
		unsigned char CONTROL_DATA;
		struct
		{
			unsigned char ShowMessageMode : 2; // 00 = not final row, 01 = after next shot, 10 = when sleep, 11 = immediately
			unsigned char Reserved2 : 1;
			unsigned char ClearBufferFirst : 1;
			unsigned char TextBlinking : 1;
			unsigned char TextInversed : 1;
			unsigned char Reserved6 : 1;
			unsigned char SoundSignal : 1;
		};
	};
	unsigned char LINE_NUMBER;
	unsigned char CHARACTER[10];
	unsigned char RESERVED0_5[6];
	unsigned char CHECKBYTE;
};

struct TCommandFrame_Field_SetRingTarget
{
	unsigned char COMMANDBYTE;
    unsigned char RING_TARGET_NUMBER;
  	union
    {
        unsigned char TARGET_BLOCK_BYTE;
        struct
        {
        	unsigned char TargetBlock : 7; // 0..127, check mapping with sets
            unsigned char UpdateCheckBit : 1;  // 1 = update device internal settings check byte
        };
    };
    unsigned char SETBYTE[16];
    unsigned char CHECKBYTE;
};

struct TCommandFrame_Field_SetParameterSet
{
	unsigned char COMMANDBYTE;
  	union
    {
		unsigned char PARAMETERSETNUMBERBYTE;
        struct
        {
        	unsigned char ParameterSetNumber : 7; // 0..127, check mapping with sets
            unsigned char UpdateCheckBit : 1;  // 1 = update device internal settings check byte
        };
    };
    unsigned char SETBYTE[16];
    unsigned char CHECKBYTE;
};

struct TCommandFrame_Field_GetParameterSet
{
	unsigned char COMMANDBYTE;
  	union
    {
		unsigned char PARAMETERSETNUMBERBYTE;
        struct
        {
        	unsigned char ParameterSetNumber : 7; // 0..127, check mapping with sets
            unsigned char UpdateCheckBit : 1;  // 1 = update device internal settings check byte
        };
    };
    unsigned char CHECKBYTE;
};

struct TCommandFrame_Field_SetFunctionMode
{
	unsigned char COMMANDBYTE;
    unsigned char OPTICALUNITSTATE;
    unsigned char RESERVED2;
    unsigned char RESERVED3; // RS_CTRL_BYTE actually, 0x27
    unsigned char CHECKBYTE;
};

struct TCommandFrame_Field_GetShot
{
	unsigned char COMMANDBYTE;
    unsigned char MEMORY_PAGE_NUMBER;
    unsigned char SHOT_INDEX;
    unsigned char CHECKBYTE;
};

struct TCommandFrame_Field_GetRingTarget
{
    unsigned char COMMANDBYTE; // 6D
    unsigned char RING_TARGET_NUMBER;
    union
    {
        unsigned char RING_TARGET_FRAME;
        struct
        {
            unsigned char RingTargetFrame : 7;
            unsigned char VerifyCheckBit : 1;
		};
	};
	unsigned char CHECKBYTE;
};

// Frame type enumeration
enum EFRAMEID_FrameTypes
{
	KFID_None,
	KFID_NotIdentified,
	KFID_StartupByte,
	KFID_AckFrame,
	KFID_ErrorFrame,
	KFID_SampleFrameField,
	KFID_ResultFrameField,
	KFID_SilhouetteBitmapFrame,
	KFID_FieldParameterFrame,
	KFID_FieldDeviceInfoFrame,
	KFID_StatusFrame,
	KFID_FieldShotCountFrame,
	KFID_InfoErrorFrame,
	KFID_RingTargetFrame
};

struct TResponseFrame_DownloadSilhouetteBitmapFrame
{
 union
 {
  struct
  {
	unsigned char SYNCBYTE1;
	unsigned char SYNCBYTE2;
	unsigned char SH1_FRAME_CONTROL;    // which silhouette
	unsigned char SH1_FRAME_NUMBER;    // which row in silhouette/byte block
	unsigned char DATA[16];
	unsigned char CHECKBYTE;
  };
  unsigned char RAW_BYTE[21];
 };

 bool IsSilhouetteSettingsFrame() { return (SH1_FRAME_NUMBER & 0x80); }
};

struct TResponseFrame_FieldSampleFrame          // CA65 (CA 65)
{
 TResponseFrame_FieldSampleFrame() { memset(this,0,sizeof(TResponseFrame_FieldSampleFrame));}

 union
 {
  struct
  {
	unsigned char SYNCBYTE1;
	unsigned char SYNCBYTE2;
	unsigned char SAMPLE_COUNTER;

	union
	{
		unsigned char CONTROLS;
		struct
		{
			unsigned char LocalSummingMode:1;
			unsigned char StoreTracking:1;
			unsigned char ReducedTX:1;
			unsigned char Reserved_3:1;
			unsigned char RangeEvaluation:1;
			unsigned char AngleCheck:1;
			unsigned char FSICheck:1;
			unsigned char Reserved_7:1;
		};
	};

	unsigned char X_MSB;
	unsigned char X_LSB;
	unsigned char Y_MSB;
	unsigned char Y_LSB;
	unsigned char TILT_MSB;
	unsigned char TILT_LSB;
	unsigned char ROTATION_MSB;
	unsigned char ROTATION_LSB;

	union
	{
		unsigned char STATUS;
		struct
		{
			unsigned char LastAfterTriggerSample:1;
			unsigned char AfterTriggerSample:1;
			unsigned char Reserved__2:1;
			unsigned char Reserved__3:1;
			unsigned char LowBattery:1;
			unsigned char TriggerLine:1;
			unsigned char ChargeLine:1;
			unsigned char PowerFailure:1;
		};
	};

	union
	{
		unsigned char LEVELS;
		struct
		{
			unsigned char LevelsHigh:4;
			unsigned char LevelsLow:4;
		};
	};

	unsigned char DISTANCE_MSB;
	unsigned char DISTANCE_LSB;

	union
	{
		unsigned char FAILUREBITS;
		struct
		{
			unsigned char LowSignal:1;
			unsigned char HighSignal:1;
			unsigned char Reserved____2:1;
			unsigned char Reserved____3:1;
			unsigned char MeasureErrorFOV:1;
			unsigned char AngleError:1;
			unsigned char FSIError:1;
			unsigned char Reserved____7:1;
		};
	};

	unsigned char RESERVED_17;
	unsigned char RESERVED_18;
	unsigned char CHECK_BYTE;
  };
  unsigned char RAW_BYTE[20];
 };
	// helper functions for get/set >8 bit info
	short GetX() { return (short)( (X_MSB<<8) | X_LSB); }
	short GetY() { return (short)( (Y_MSB<<8) | Y_LSB); }
	short GetTilt() { return (short)( (TILT_MSB<<8) | TILT_LSB); }
	short GetRotation() { return (short)( (ROTATION_MSB<<8) | ROTATION_LSB); }

	double GetDist_m() { return 0.1 * (short)((DISTANCE_MSB << 8) | DISTANCE_LSB); }

	bool UnitNAU()    { return (!CONTROLS) & 1; }
	bool UnitMetric() { return CONTROLS & 1; }

	void SetX(short A) { Set16ToMSBLSB(&X_MSB, A); }
	void SetY(short A) { Set16ToMSBLSB(&Y_MSB, A); }
};

struct TResponseFrame_FieldResultFrame          // CA6A (CA 6A)
{
 TResponseFrame_FieldResultFrame() { memset(this,0,sizeof(TResponseFrame_FieldResultFrame));}

 union
 {
  struct
  {
	unsigned char SYNCBYTE1;
	unsigned char SYNCBYTE2;
	unsigned char SHOT_NUMBER;

	union
	{
		unsigned char CONTROLS;
		struct  // not used here - (used in sample frame)
		{
			unsigned char LocalSummingMode:1;
			unsigned char StoreTracking:1;
			unsigned char ReducedTX:1;
			unsigned char Reserved_3:1;
			unsigned char RangeEvaluation:1;
			unsigned char AngleCheck:1;
			unsigned char FSICheck:1;
			unsigned char Reserved_7:1;
		};
	};

	unsigned char X_MSB;
	unsigned char X_LSB;
	unsigned char Y_MSB;
	unsigned char Y_LSB;
	unsigned char TILT_MSB;
	unsigned char TILT_LSB;
	unsigned char ROTATION_MSB;
	unsigned char ROTATION_LSB;

	union
	{
		unsigned char STATUS;
		struct
		{
			unsigned char CloseRangeHit : 1;
			unsigned char HitIndicatorBit : 1;
			unsigned char ReducedTXPower : 1;
			unsigned char Status_Reserved_3 : 1;
			unsigned char LowBattery : 1;
			unsigned char TriggerLine : 1;
			unsigned char ChargeLine : 1;
			unsigned char PowerFailureLine : 1;
		};
	};

	union
	{
		unsigned char LEVELS;
		struct
		{
			unsigned char LevelsHigh:4;
			unsigned char LevelsLow:4;
		};
	};

	unsigned char RING_TARGET;
	unsigned char SILHOUETTE_TARGET;
	unsigned char FUNCTION_MODE;
	unsigned char STARTUP_COUNTER;
	unsigned char TIME_HH;
	unsigned char TIME_H;
	unsigned char TIME_L;
	unsigned char TIME_SUB;
	unsigned char FSI_X_MSB;
	unsigned char FSI_X_LSB;
	unsigned char RING_RESULT;

	union
	{
		unsigned char RING_SECTOR_BYTE;
		struct
		{
			unsigned char RingSector:4;
			unsigned char Unused:3;
			unsigned char XResult:1;
		};
	};

	unsigned char SILHOUETTE_RESULT;

	union
	{
		unsigned char FAILUREBITS;
		struct
		{
			unsigned char LowSignal:1;
			unsigned char HighSignal:1;
			unsigned char Reserved____2:1;
			unsigned char Reserved____3:1;
			unsigned char MeasureErrorFOV:1;
			unsigned char AngleError:1;
			unsigned char FSIError:1;
			unsigned char NoValidSamples:1;
		};
	};

	unsigned char DISTANCE_1;
	unsigned char DISTANCE_2;
	unsigned char MM_X_MSB;
	unsigned char MM_X_LSB;
	unsigned char MM_Y_MSB;
	unsigned char MM_Y_LSB;
	unsigned char RESERVED_34;
	unsigned char CHECK_BYTE;
  };
  unsigned char RAW_BYTE[36];
 };
	// helper functions for get/set >8 bit info
	short GetX() { return (short)( (X_MSB<<8) | X_LSB); }
	short GetY() { return (short)( (Y_MSB<<8) | Y_LSB); }
	double GetXmm() { return 0.1 * (short)( (MM_X_MSB<<8) | MM_X_LSB); }
	double GetYmm() { return 0.1 * (short)( (MM_Y_MSB<<8) | MM_Y_LSB); }
	short GetXmm_raw() { return (short)( (MM_X_MSB<<8) | MM_X_LSB); }
	short GetYmm_raw() { return (short)( (MM_Y_MSB<<8) | MM_Y_LSB); }
	short GetTilt() { return (short)( (TILT_MSB<<8) | TILT_LSB); }
	short GetRotation() { return (short)( (ROTATION_MSB<<8) | ROTATION_LSB); }
	short GetFSI() { return (short)( (FSI_X_MSB<<8) | FSI_X_LSB); }
	short GetDist() { return (short)( (DISTANCE_1<<8) | DISTANCE_2); }

	void SetX_raw(int A) { Set16ToMSBLSB(&X_MSB, A); }
	void SetY_raw(int A) { Set16ToMSBLSB(&Y_MSB, A); }
	void SetXmm_raw(int A) { Set16ToMSBLSB(&MM_X_MSB, A); }
	void SetYmm_raw(int A) { Set16ToMSBLSB(&MM_Y_MSB, A); }

	unsigned int GetTime() { return (TIME_HH << 16) | (TIME_H << 8) |  (TIME_L); }
};

struct TResponseFrame_Field_AckFrame
{
 union
 {
  struct
  {
   unsigned char SYNCBYTE1;
   unsigned char SYNCBYTE2;
   unsigned char ACKBYTE;
   unsigned char INFOBYTE;
   unsigned char CHECKBYTE;
  };
  unsigned char RAW_BYTE[5];
 };
};

struct TResponseFrame_ErrorFrame
{
 union
 {
  struct
  {
   unsigned char SYNCBYTE1;
   unsigned char SYNCBYTE2;
   unsigned char ERRORBYTE;
   unsigned char CHECKBYTE;
  };
  unsigned char RAW_BYTE[4];
 };
};

struct TResponseFrame_Field_ParameterSetFrame
{
 union
 {
  struct
  {
   unsigned char SYNCBYTE1;
   unsigned char SYNCBYTE2;
  	union
    {
		unsigned char PARAMETERSETNUMBERBYTE;
        struct
        {
        	unsigned char ParameterSetNumber : 7; // 0..127, check mapping with sets
            unsigned char UpdateCheckBit : 1;  // 1 = update device internal settings check byte
        };
    };
   unsigned char PARAMETERBYTE[16];
   unsigned char CHECKBYTE;
  };
  unsigned char RAW_BYTE[20];
 };
};

struct TResponseFrame_Field_DeviceInfoFrame
{
 union
 {
  struct
  {
   unsigned char SYNCBYTE1;
   unsigned char SYNCBYTE2;
   unsigned char PAGE_NUMBER;
   unsigned char INFOBYTE[16];
   unsigned char CHECKBYTE;
  };
  unsigned char RAW_BYTE[20];
 };
};

struct TResponseFrame_Field_ShotCountFrame // 596320 (59 63 20)
{
 union
 {
  struct
  {
   unsigned char SYNCBYTE1;
   unsigned char SYNCBYTE2;
   unsigned char STATUS;
   struct
   {
    unsigned char BLOCK_FUNCTION_MODE;
    unsigned char BLOCK_SHOT_COUNT;
   } MemoryPages[7];
   unsigned char RESERVED_17;
   unsigned char RESERVED_18;
   unsigned char CHECKBYTE;
  };
  unsigned char RAW_BYTE[20];
 };
 int GetNumberOfShots() { int ret = 0;
                          for (int i=0;i<7;i++) ret += MemoryPages[i].BLOCK_SHOT_COUNT;
                          return ret; }
};

struct TResponseFrame_Field_GetShotFrame // 596320 (59 63 20)
{
 union
 {
  struct
  {
   unsigned char SYNCBYTE1;
   unsigned char SYNCBYTE2;
   unsigned char STATUS;
   struct
   {
    unsigned char BLOCK_FUNCTION_MODE;
    unsigned char BLOCK_SHOT_COUNT;
   } MemoryPages[6];
   unsigned char RESERVED_15;
   unsigned char RESERVED_16;
   unsigned char RESERVED_17;
   unsigned char RESERVED_18;
   unsigned char CHECKBYTE;
  };
  unsigned char RAW_BYTE[20];
 };
 int GetNumberOfShots() { int ret = 0;
                          for (int i=0;i<6;i++) ret += MemoryPages[i].BLOCK_SHOT_COUNT;
                          return ret; }
};

struct TResponseFrame_Field_InfoErrorFrame
{
 union
 {
  struct
  {
   unsigned char SYNCBYTE1;
   unsigned char SYNCBYTE2;
   unsigned char INFOBYTE;
   unsigned char STATUSBYTE;
   unsigned char OPTICALUNIT;
   unsigned char ERRORCODE;
   unsigned char ERR0_COMMAND;
   unsigned char ERR1_VALUE;
   unsigned char ERR2_VALUE;
   unsigned char ERR3_VALUE;
   unsigned char OTHERBYTES[5];
   unsigned char CHECKBYTE;
  };
  unsigned char RAW_BYTE[16];
 };
};

struct TResponseFrame_Field_RingTargetFrame // 59 6D - 596d
{
 union
 {
  struct
  {
   unsigned char SYNCBYTE1;
   unsigned char SYNCBYTE2;
   unsigned char RING_TARGET_NUMBER;
  	union
    {
        unsigned char TARGET_BLOCK_BYTE;
        struct
        {
        	unsigned char TargetBlock : 7; // 0..127, check mapping with sets
            unsigned char UpdateCheckBit : 1;  // 1 = update device internal settings check byte
        };
    };
   unsigned char DATABYTES[16];
   unsigned char CHECKBYTE;
  };
  unsigned char RAW_BYTE[21];
 };
 // we have multiple payload types for this frame; TARGET_FRAME actually should be checked but at this time we leave it to "user"
 TParameterSet_Field_RingTarget* GetAsRingTarget() { return (TParameterSet_Field_RingTarget*)DATABYTES; }
 TParameterSet_Field_RingTarget_block2* GetAsRingTarget_block2() { return (TParameterSet_Field_RingTarget_block2*)DATABYTES; }
 TParameterSet_Field_RingTarget_block3* GetAsRingTarget_block3() { return (TParameterSet_Field_RingTarget_block3*)DATABYTES; }
 TParameterSet_Any16bytes* GetAsRingTarget_Name() { return (TParameterSet_Any16bytes*)DATABYTES; }
};


// Pack of recent frames; stores to overall state of device by keeping in memory the parameter and info frames;
// Also stores state of received frames and bits to indicate whether each frames have been handled or checked by the engine (UI)
// volatile so that optimizing compiler don't make nasty things
volatile struct TDeviceStatusFrames
{
    // which frame was received last? (ack, err or possibly some other)
    EFRAMEID_FrameTypes FLastReceivedFrameType;

	// parameter sets are stored when frames are received from the device;
	// also, when updating sets in the device, these sets are used to build the update frame.
	// field parameter sets (without silhouettes)
	TParameterSet_Field_1            FFieldParameterSet1; // use case settings
	TParameterSet_Field_2            FFieldParameterSet2; // technical settings
	//TParameterSet_Field_3        	 FFieldParameterSet3; // not used (reserved for factory use)
	TParameterSet_Field_4            FFieldParameterSet4; // device serial number

	struct TResponseFrame_Field_DeviceInfoFrame FFieldDeviceInfoPages[8];
	struct TResponseFrame_Field_DeviceInfoFrame FFieldUsageInfoFrame;
	struct TResponseFrame_Field_DeviceInfoFrame FFieldErrorCounterFrame;

	// recent received frames
	struct TResponseFrame_Field_AckFrame          FLastAckFrame;
	struct TResponseFrame_Field_ShotCountFrame    FLastShotCountFrame;
	struct TResponseFrame_ErrorFrame        FLastErrorFrame;
	struct TResponseFrame_FieldResultFrame  FLastResultFrameField;
	struct TResponseFrame_FieldSampleFrame  FLastSampleFrameField;
	struct TResponseFrame_DownloadSilhouetteBitmapFrame FLastSilhouetteBitmapFrame;
	struct TResponseFrame_Field_ParameterSetFrame   FLastFieldParameterSetFrame;
	struct TResponseFrame_Field_DeviceInfoFrame FLastFieldDeviceInfoFrame;
	struct TResponseFrame_Field_InfoErrorFrame   FLastInfoErrorFrame;
    struct TResponseFrame_Field_RingTargetFrame       FLastRingTargetFrame;
    struct TParameterSet_Field_SilhouetteSettings FLastSilhouetteSettings;

    // boolean indicate that a frame was received but not yet handled by the state machine
    bool FAckFrameReceived;
    bool FAckFrameHandled;
    bool FErrorFrameReceived;
    bool FErrorFrameHandled;
    bool FStatusFrameReceived;
    bool FStatusFrameHandled;
    bool FSampleFrameReceived;
	bool FSampleFrameHandled;
	bool FResultFrameReceived;
    bool FDownloadSilhouetteBitmapFrameReceived;
    bool FDownloadSilhouetteBitmapFrameHandled;
    bool FParameterSet_1_FrameReceived;
    bool FParameterSet_1_FrameHandled;
    bool FParameterSet_2_FrameReceived;
    bool FParameterSet_2_FrameHandled;
    bool FDeviceInfoFrameReceived;
    bool FDeviceInfoFrameHandled;
	bool FShotCountFrameReceived;

	// quick for checking if anything worth noticing has happened
	bool AnyFrameReceived()
	{
		return FAckFrameReceived | FErrorFrameReceived | FSampleFrameReceived |
			   FParameterSet_1_FrameReceived | FParameterSet_2_FrameReceived |
			   FDeviceInfoFrameReceived | FStatusFrameReceived;
	}
	// reset received bits; typically when start wait for ack or any other response
	void ResetReceivedHandled()
	{
		FAckFrameReceived = FAckFrameHandled =
		FErrorFrameReceived = FErrorFrameHandled =
		FStatusFrameReceived = FStatusFrameHandled =
		FSampleFrameReceived = FSampleFrameHandled =
		FParameterSet_1_FrameReceived = FParameterSet_1_FrameHandled =
		FParameterSet_2_FrameReceived = FParameterSet_2_FrameHandled =
		FDeviceInfoFrameReceived = FDeviceInfoFrameHandled =
		FShotCountFrameReceived = false;
	}

	// clear all sets and bits
	void Reset() { memset(this, 0, sizeof(TDeviceStatusFrames)); };
};

#endif

