//---------------------------------------------------------------------------
// SerialHandler.cpp
//
// Noptel/FN Expert, NOS Pro
// C. Noptel Oy, Oulu Finland, 2010-2016
//
// Serial communications components: Sourceforge ComPort Library, COMPortLib, (2010)
//
//---------------------------------------------------------------------------

// base 'visual component library'
#include <vcl.h>

// serial port object (we use a Delphi implementation; basic 'API' makes it easy to change implementation)
// CPortFromDelphi contains abstract factory for com port object
#include "CPortFromDelphi.hpp"
// CPort includes API translated from Delphi implementation; auto-generated in this case
#include "CPort.hpp"

// pre-compiled header stop
#pragma hdrstop

// class definition and other includes
#include "SerialHandler.h"
// incoming data frame structures and parameter sets
#include "FramesAndSets.h"
// the PDM device data handler; buffers and parses incoming bytes
#include "DataHandler.h"

// to measure latencies; 1 so that wont get zero divides
__int64 GPerformanceCounterFrequency=1;

// terminal form
//#include "UnitTerminalForm.h"

//---------------------------------------------------------------------------

// strings object has com ports
Classes::TStrings* FPortList;

// return pointer to list
TStrings* TSerialHandler::GetPortList()
{
    if (!FPortList)
        FPortList = new TStringList();

    EnumComPorts(FPortList);

    return FPortList;
}

// ** Constructor
// - Allocate incoming data buffer
// - Create com port object and set bauds (but keep it closed)
TSerialHandler::TSerialHandler(TDataHandler& ADataHandler)
{
	memset(this, 0, sizeof(*this) );

	CreateSerialPort();

	FDataHandler = &ADataHandler;

	LARGE_INTEGER LInt;
	QueryPerformanceFrequency(&LInt);
	GPerformanceCounterFrequency = LInt.QuadPart;

	// For debug purposes, enabled logging at start until disabled after scan.
	//SetLogEnabled(true);
}

void TSerialHandler::CreateSerialPort()
{
    // this function calls the delphi-constructor of the CPort class.
	FCOMPort = CreateCPort();
    // make serial traffic important
    FCOMPort->EventThreadPriority = tpNormal;
	// not connected until requested
	FCOMPort->Connected = false;
	// device starts up in 9600
    FCOMPort->BaudRate = br9600;
    // enough buffer
    FCOMPort->Buffer->InputSize = 4096;

    FCOMPort->SyncMethod = smNone; // our PDMDevice/DataHandler is "threadsafe"
    FCOMPort->DataBits = dbEight;
    FCOMPort->StopBits = sbOneStopBit;
    FCOMPort->Parity->Bits = prNone;
    FCOMPort->DiscardNull = false;

	// DTR and RTS down on open; set up manually
    //FCOMPort->SetDTR(true);
    //FCOMPort->SetRTS(true);

	// function which handles incoming data
    FCOMPort->OnRxChar = DataAvailableFromComPort;
}

// ** Destructor
// - close and free com port object
// - free incoming data buffer
TSerialHandler::~TSerialHandler()
{
	// try
	if (FCOMPort)
	{
		FCOMPort->Connected = false;
		// free com port object
		delete FCOMPort;
	}

    //if (FPortList)
	//    delete FPortList;

	// disable log
	SetLogEnabled(false);

}

// ****
// **** COM Port control functions and event handlers
// ****

// ** set port and connect
bool TSerialHandler::OpenCOMPort(int aComPort)
{
	// only assume success if opening doesn't throw an exception 
	bool LSuccess = false;
	try
    {
		// our port implementation wants a string
		AnsiString portname = "COM";
		portname += aComPort;
		//FCOMPort->Port = portname;
		FCOMPort->SetPortString(portname);
		FCOMPort->Connected = true;
		LSuccess = true;
    }
    catch (...)
    {
    	// any exception is assumed failure.
    };
	return LSuccess;
}

// ** close connection
void TSerialHandler::CloseCOMPort()
{
	// assume closing is ok always.
	try
	{
		if (FCOMPort)
		{
			FCOMPort->Connected = false;
            // no longer delete if close
			//delete FCOMPort;
		}
	}
	catch (Exception& E)
	{
		// now instead of deleting the thing we just ignore it is still there...
		FCOMPort = NULL;
	}
}

// has a successful open been done
bool TSerialHandler::IsPortOpen()
{
	return FCOMPort->Connected;
}

// ** write given bytes to port
void TSerialHandler::Write(void* APtrToData, int ALength)
{
	if (!(FCOMPort && FCOMPort->Connected))
	{
		throw new Exception("Write to closed port");        
	}

	try
	{
		FCOMPort->Write( (const char *) APtrToData, ALength);
	}
	catch (Exception& e)
	{
		// notify port is dead
		//int i=0;
	}

	LARGE_INTEGER LInt;
	QueryPerformanceCounter(&LInt);
	FLastSendTime = LInt.QuadPart;

	// If logging is enabled, find end and add new message to the end of the list
	if (FLoggingEnabled)
	{
		// Find the last entry of the write list
		TLogRow* LLastEntry = FWriteLogEntryList;
		while (LLastEntry->FPtrToNext != NULL)
		{
			LLastEntry = LLastEntry->FPtrToNext;
		}

		TLogRow* LNewRow = new TLogRow(LLastEntry);
		LLastEntry->FPtrToNext = LNewRow;
		LNewRow->FPtrToPrev = LLastEntry;

		LNewRow->SetHexMessage((char *)APtrToData, ALength);
		LNewRow->FTimestamp = LInt.QuadPart;
	}
}

// ** read requested number of bytes, or fail on timeout
// - returns number of bytes read
int TSerialHandler::Read(void* APtrToData, int ALength)
{
	// TODO add logging to here too
	return FCOMPort->Read( (char *)APtrToData, ALength);
}

void TSerialHandler::SetBaudRate(int BaudRate)
{
    AnsiString str;
    str = BaudRate;
    FCOMPort->BaudRate = StrToBaudRate(str);
}

// ** set DTR and RTS to serial port; used to control the device
void TSerialHandler::SetDTRRTS(bool DTR, bool RTS)
{
	FCOMPort->SetDTR(DTR);
	FCOMPort->SetRTS(RTS);
}

// ** event handler linked to com port's incoming data event
// - wrapper function which forwards read bytes to incoming data buffer and check for frames
// - made simple so can be changed easily if another com port implementation is used
// - "fastcall" so that the function can be linked to com port objects data received -event
void __fastcall TSerialHandler::DataAvailableFromComPort(TObject* ASender, int ACount)
{
	char LReceiveBuffer[256];
	int  LReceiveBufferBytesReceived;

	LARGE_INTEGER LInt;
	QueryPerformanceCounter(&LInt);
	FLastLatency = (LInt.QuadPart - FLastSendTime) * (1.0 / GPerformanceCounterFrequency);

	TComPort* LPort = (TComPort*)ASender;
	// while we have stuff coming...
    int LBytesComing = ACount;
	while (LBytesComing > 0)
    {
		// we read 256 bytes at time; could well be the 4k but that much should not ever be needed
		if (LBytesComing > 256) LBytesComing = 256;

		// read the actual bytes
		LReceiveBufferBytesReceived = LPort->Read(LReceiveBuffer, LBytesComing);

		// If logging is enabled, find end and add new message to the end of the list
		if (FLoggingEnabled)
		{

			// Find the last entry of the READ list
			TLogRow* LLastEntry = FReadLogEntryList;
			while (LLastEntry->FPtrToNext != NULL)
			{
				LLastEntry = LLastEntry->FPtrToNext;
			}

			TLogRow* LNewRow = new TLogRow(LLastEntry);
			LLastEntry->FPtrToNext = LNewRow;
			LNewRow->FPtrToPrev = LLastEntry;

			LNewRow->SetHexMessage(LReceiveBuffer, LReceiveBufferBytesReceived);
			// timestamp is taken before message is decoded/handled
			LNewRow->FTimestamp = LInt.QuadPart;
		}

        // move bytes upward to be buffered and handled
        FDataHandler->ReceiveIncomingBytes(LReceiveBuffer, LReceiveBufferBytesReceived);

		// update bytes to receive
	    LBytesComing = LPort->InputCount();
    }
}

void TSerialHandler::SetLogEnabled(bool AEnable)
{
	FLoggingEnabled = AEnable;
    // prepare logs, make initial entries having current datetime
	if (AEnable)
	{
		QueryPerformanceCounter(&FLoggingStarted);

		AnsiString StartMessage;
		StartMessage = "Log started " +  TDateTime::CurrentDateTime().DateTimeString();

 		// set initial entries with timestamp
		TLogRow* StartLogEntry = new TLogRow(0);
		StartLogEntry->SetMessage(StartMessage.c_str(), StartMessage.Length());
		FReadLogEntryList = StartLogEntry;

		// create another copy for the write log too
		StartLogEntry = new TLogRow(0);
		StartLogEntry->SetMessage(StartMessage.c_str(), StartMessage.Length());
		FWriteLogEntryList = StartLogEntry;
	}
}

// reset log start times for example when starting programming
void TSerialHandler::ResetLogStartTimes()
{
	if (FLoggingEnabled)
	{
		QueryPerformanceCounter(&FLoggingStarted);
		FReadLogEntryList->FTimestamp = FLoggingStarted.QuadPart;
		FWriteLogEntryList->FTimestamp = FLoggingStarted.QuadPart;
	}
}
