//---------------------------------------------------------------------------
// UnitFormSelectSetup.cpp
//
// Noptel/FN Expert, NOS Pro
// C. Noptel Oy, Oulu Finland, 2011-2017
//
// Training Setup dialog form with
//  -Setup List and Setup List Organization buttons
//  -Expert device Setup (stand-alone) & Start & Configure buttons
//
// Reference: Noptel Doc# N41839: Noptel Expert Application - User's Guide 2.1
//            Chapters: 3.3 Real Training & 3.4 Simulated Training
//
//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "UnitFormSelectSetup.h"

// to get a com port
#include "UnitFormPrepareDownload.h"

// shooter form
#include "UnitFormShooter.h"

// device
#include "Device.h"

// min
#include "Math.hpp"

// setup file handler
#include "SetupData.h"

// shot data to create shooter
#include "ShotData.h"

// device manager
#include "DeviceManager.h"

// main panel
#include "UnitFormShooterPanel.h"

// frames
#include "FramesAndSets.h"

// back to main menu
#include "UnitFormStartMenu.h"

// editor
#include "UnitFormEditSetup.h"

// translate
#include "Globals.h"

#include "MiniXMLParser.h"

#include "UnitFormNetworkControl.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TFormSelectSetup *FormSelectSetup;

//---------------------------------------------------------------------------
__fastcall TFormSelectSetup::TFormSelectSetup(TComponent* Owner)
	: TForm(Owner)
{
	FCurrentSetup = NULL;
	
	FTargetBitmap = new Graphics::TBitmap();
	FTargetBitmap->Width = 64;
	FTargetBitmap->Height = 64;

    FTargetRenderer.SetBitmapDimensions(64,64);
	FTargetRenderer.FBitmapTarget = FTargetBitmap;

	DrawGridSetups->ColWidths[1] = DrawGridSetups->Width - DrawGridSetups->ColWidths[0] - 26;

    FSList = new TStringList;

	OpenDialogSetupFile->InitialDir = GSetupsPath;
	FormEditSetup->OpenDialogLoad->InitialDir = GSetupsPath;
	FormEditSetup->SaveDialogSave->InitialDir = GSetupsPath;

	// load default setup. The path is set in AppInitialize which is run before this.
	// TODO if given another setup file as startup parameter?
	if (FileExists(GDefaultSetupPath) )
	{
		bool LOK = FSetupHandler.LoadFromFile(GDefaultSetupPath);
		if (!LOK)
		{
			WideString LStr = WGetTranslation("Error loading setup file");
			WideString LStrCap = WGetTranslation("Error");
			//Application->MessageBox("Error loading setup file","Error",MB_OK);
			Application->MessageBox(LStr.c_bstr(),LStrCap.c_bstr(),MB_OK);
			throw new Exception("Error loading setup file");
		}

		DrawGridSetups->Row = 0;
		DrawGridSetups->RowCount = FSetupHandler.FSetupCount; 
		DrawGridSetups->Invalidate();
		DrawGridSetupsClick(NULL);
	}

	// to get translation
	WGetTranslation("Ring or silhouette not selected");
}

void __fastcall TFormSelectSetup::FormDestroy(TObject *Sender)
{
	delete FTargetBitmap;
	delete FSList;
}
//---------------------------------------------------------------------------

void __fastcall TFormSelectSetup::FormClose(TObject *Sender, TCloseAction &Action)
{
	FormStartMenu->Show();
	FormStartMenu->BringToFront();
}
//---------------------------------------------------------------------------

void TFormSelectSetup::SortSetups(bool ARingsFirst)
{
	// go through setups, raise ring/silhouette until no ring is above it... (bubble sort)
	for (int i = 1; i < FSetupHandler.FSetupCount; i++)
	{
		TTargetSetup& LA = FSetupHandler.GetSetup(i-1);
		TTargetSetup& LB = FSetupHandler.GetSetup(i);

		// if B is sorted higher than A (previous), move B upwards until find correct place
		if ( (LB.FTargetDef.FUsingRing == ARingsFirst) &&
			 (LA.FTargetDef.FUsingRing != ARingsFirst) )
		{
			FSetupHandler.SwapAtIndexes(i-1, i);
			// move back to first (i++ = 1)
			i = 0;
		}
	}
}

//---------------------------------------------------------------------------
void __fastcall TFormSelectSetup::DrawGridSetupsDrawCell(TObject *Sender, int ACol, int ARow,
	  TRect &Rect, TGridDrawState State)
{
	// row 0 = header?
	// col 0 is the icon of discipline
	if ( (ARow < 0) || (ARow >= FSetupHandler.FSetupCount) ) return;

	TTargetSetup& LSetup = FSetupHandler.GetSetup(ARow);

	// draw background color
	if (State.Contains(gdSelected) )
	{
    	TColor LSelectedRowColor = (TColor)RGB(224,224,224); 
		DrawGridSetups->Canvas->Brush->Style = bsSolid;
		DrawGridSetups->Canvas->Brush->Color = LSelectedRowColor;
		DrawGridSetups->Canvas->FillRect(Rect);
	}

	// col 0 = target bitmap
	if (ACol == 0)
	{
		FTargetRenderer.FTargetDefinition.FRingTarget = LSetup.FTargetDef.FRingTarget;
		FTargetRenderer.FTargetDefinition.FSilTarget = LSetup.FTargetDef.FSilTarget;
		FTargetRenderer.FTargetDefinition.FSilhouette = LSetup.FTargetDef.FSilhouette;
		FTargetRenderer.FTargetDefinition.FUsingRing = LSetup.FTargetDef.FUsingRing;
		FTargetRenderer.FTargetDefinition.FUsingSilhouette = LSetup.FTargetDef.FUsingSilhouette;

		// we dont get division by zeros
		int LDiameter = LSetup.FTargetDef.GetWidth_mm(); 
		FTargetRenderer.SetDrawAreaIn_mm(LDiameter);

		FTargetRenderer.DrawTarget();
		DrawGridSetups->Canvas->Draw(Rect.left, Rect.top, FTargetBitmap);
	}
	// col 1 is the name of the discipline and other text
	else if (ACol == 1)
	{
		FSList->Clear();
		if (LSetup.FDescription)
			FSList->AddStrings(LSetup.FDescription);

    	FSList->Add(LSetup.FTargetDef.GetTargetName() );

		if (DrawGridSetups->RowHeights[ARow] < FSList->Count * Canvas->TextHeight("0") )
		{
			DrawGridSetups->RowHeights[ARow] = FSList->Count * Canvas->TextHeight("0") + 2;
		}

		for (int i = 0; i < FSList->Count; i++)
		{
			if (i == 0)
			{
				DrawGridSetups->Canvas->Font->Size = 10;
				DrawGridSetups->Canvas->Font->Style = TFontStyles() << fsBold;
				DrawGridSetups->Canvas->Font->Color = (TColor)RGB(255,0,0);
			}
			else
			{
				DrawGridSetups->Canvas->Font->Style = TFontStyles();
            	DrawGridSetups->Canvas->Font->Color = (TColor)RGB(0,0,0);
			}

			//DrawGridSetups->Canvas->TextOutA(Rect.left + 2, Rect.top + i * Canvas->TextHeight("0"), FSList->Strings[i]);
			DrawGridSetups->Canvas->TextOutA(Rect.left + 8, 2 + Rect.top + i * 1.2 * Canvas->TextHeight("0"), FSList->Strings[i]);
		}

	}
}
//---------------------------------------------------------------------------

void __fastcall TFormSelectSetup::ButtonOpenSetupFileClick(TObject *Sender)
{

	if (OpenDialogSetupFile->Execute())
	{
		UnicodeString LFileName = OpenDialogSetupFile->FileName;

		if (!FSetupHandler.LoadFromFile(LFileName) )
		{
			Application->MessageBox(WGetTranslation("Error loading setup file").c_str(),WGetTranslation("Error").c_str(),MB_OK);
			throw new Exception("Error loading setup file");
        }

		DrawGridSetups->RowCount = FSetupHandler.FSetupCount;
		DrawGridSetups->Row = 0;

		// sort setups according to selected mode (classic=rings first, dynamic=sils first)
		// ALWAYS sils first
		SortSetups(false);
		//SortSetups(GSelectedMode == KClassicTraining);

		DrawGridSetups->Invalidate();

		// update selection
		DrawGridSetupsClick(NULL);
	}
}
//---------------------------------------------------------------------------

void __fastcall TFormSelectSetup::ButtonOpenDefaultSetupFileClick(TObject *Sender)
{

	UnicodeString LStr = WGetTranslation("Load factory default Setup list?");

	int LRes = Application->MessageBox(LStr.c_str(), WGetTranslation("Confirm").c_str(), MB_OKCANCEL );
	if (LRes == IDOK )
	{
		FSetupHandler.LoadFromFile(GOriginalDefaultSetupPath);

		// force redraw targets
		DrawGridSetups->RowCount = FSetupHandler.FSetupCount;
		DrawGridSetups->Row = 0;
		DrawGridSetups->Invalidate();

		// update selection
		DrawGridSetupsClick(NULL);
	}
}
//---------------------------------------------------------------------------


void __fastcall TFormSelectSetup::ButtonStartClick(TObject *Sender)
{
	// pick the selected target definition
	TTargetDefinition& LT = FCurrentSetup->FTargetDef;

	if (! (LT.FUsingRing || LT.FUsingSilhouette) )
	{
		Application->MessageBox(WGetTranslation("Ring or silhouette not selected").c_str(),WideString("Error").c_bstr(),MB_OK);
		return;
	}

	Beep();

	TDeviceManager* LDevMan = GetDeviceManager();

	// now we must have devices connected
	if (LDevMan->GetDeviceCount() == 0)
	{
		Application->MessageBox(WideString("No device connected. Shooting not possible.").c_bstr(),WideString("Insert Optical Units").c_bstr(), MB_OK);
		return;
	}

	PanelShowProgress->Show();
	Application->ProcessMessages();
	TimerProgressBar->Enabled = true;

	// hourglass or what ever it is
	AddBusyCount();
    DisableUI();

	// we wait until all the devices have UploadTargetsDone
	FDeviceCountAtStart = LDevMan->GetDeviceCount();

	// clicked "start"
	FSetOrStartGoing = 2;

	// start status poll timer
	TimerProgrammingPoller->Enabled = true;

	bool LIsDynamic = CheckBoxModeDynamic->Checked;

    // PKy 21.3.2013. Classic mode vain jos rengastauluja laitteen ollessa johdolla kiinni
	LIsDynamic = !LT.FUsingRing;

	// 19.11.2014. Update classic/dynamic selection
	UpdateDynamicClassicSelection();

	LIsDynamic = (GSelectedMode == KDynamicTraining);

	// clear session before starting a new one, go through devices and destroy shooters
	for (int i = 0; i < LDevMan->GetDeviceCount(); i++)
	{
		// delete "old" shooter, create new one
		TDevice* LDevice = LDevMan->GetDevice(i);
		if (LDevice->FShooter)
		{
			GSession->FShooters->Extract(LDevice->FShooter);
			delete LDevice->FShooter;
			LDevice->FShooter = NULL;
		}
	}

	// AND if session still has (disconnected) shooters, delete them too
	while (GSession->GetShooterCount() > 0)
	{
		TShooter* LSh = GSession->GetShooter(0);
		GSession->FShooters->Extract(LSh);
		delete LSh;
    }

	// for each created shooter...
	for (int i = 0; i < LDevMan->GetDeviceCount(); i++)
	{
		// device should be without shooter by default (?)
		TDevice* LDevice = LDevMan->GetDevice(i);

		// clear zeroing
        LDevice->ClearZeroing();

		// delete "old" shooter, create new one
		if (LDevice->FShooter)
		{
			GSession->FShooters->Extract(LDevice->FShooter);
			delete LDevice->FShooter;
			LDevice->FShooter = NULL;
		}

		// open COM port if we had downloaded stuffs and closed port after it.
		// TODO this is no longer used
		if (LDevice->FCOMPort < 0)
		{
			// open port and put bauds, and DTR RTS up
			LDevice->OpenComPort(-LDevice->FCOMPort, true);
			Sleep(1000);
		}

		// prepare async upload information
		TDeviceAsyncDownloadInfo& LInfo = LDevice->FAsyncInfo;
        memset(&LInfo, 0, sizeof(LInfo));
		LInfo.FUploadTargetsGoing = true;
		LInfo.FUploadTargetsDone = false;
		LInfo.FUploadUserScenarioAndErase = true;
		// update
		LInfo.FLastAsyncProcessed = GetTickCount();

		// create new shooter
		TShooter* LShooter = new TShooter();
		GSession->FShooters->Add(LShooter);
        LShooter->FSerialNumber = LDevice->FSerialNumber;

		// selected target
		LShooter->FTargetDefinition = LT;
		// assign shooter to device
		LDevice->FShooter = LShooter;

		// set startup mode accordingly
		LDevice->GetUserScenarioSet().STARTUP_FUNCTION_MODE = LIsDynamic ? 2 : 1;

		// TRANSLATE IMAGE ORDER TO GUN MOUNTING
		int LMount;
		if (SpeedButtonMounting->Tag == 0) LMount = 0;
		if (SpeedButtonMounting->Tag == 1) LMount = 1;
		if (SpeedButtonMounting->Tag == 2) LMount = 3;
		if (SpeedButtonMounting->Tag == 3) LMount = 2;

		// Apply selected gun mounting.
		LDevice->GetUserScenarioSet().DeviceMounting = LMount;
		//LDevice->GetUserScenarioSet().DeviceMounting = FCurrentSetup->FUseScenario.DeviceMounting;

		// update classic or dynamic mode settings?
		TModeSettings& LMode = LIsDynamic ? LDevice->GetUserScenarioSet().DYNAMIC_MODE_SETTINGS : LDevice->GetUserScenarioSet().CLASSIC_MODE_SETTINGS;

		// copy setup settings to device
		LMode = LIsDynamic ? FCurrentSetup->FUseScenario.DYNAMIC_MODE_SETTINGS : FCurrentSetup->FUseScenario.CLASSIC_MODE_SETTINGS;
		LDevice->GetUserScenarioSet().Arrow_buttons = FCurrentSetup->FUseScenario.Arrow_buttons;

		// no target in use until set...
		LMode.RING_TARGET_NUMBER = 255;
		LMode.SILH_TARGET_NUMBER = 255;
		LMode.EvaluateSilhouette = false;
		LMode.CountRingScore = false;

		int LProgressMax = 14;
		LInfo.FProgressNow = 0;

		if (LT.FUsingRing)
		{
			LMode.CountRingScore = true;
			LMode.RING_TARGET_NUMBER = 0;

			LDevice->FAllTargets.SetRingTarget(0, LT.FRingTarget);

			LInfo.FUploadRingTarget = true;
			LInfo.FUploadRingTargetIndex = 0;
			LInfo.FUploadRingTargetBlock = 0;

			LProgressMax += 4;
		}

		if (LT.FUsingSilhouette)
		{
			LMode.EvaluateSilhouette = true;
			LMode.SILH_TARGET_NUMBER = 0;

			LDevice->FAllTargets.SetSilhouetteTarget(0, LT.FSilTarget);
			LDevice->FAllTargets.SetSilhouetteBitmapMatrix(0, LT.FSilhouette);

			LInfo.FUploadSilhouetteTarget = true;
			LInfo.FUploadSilhouetteTargetIndex = 0; // silhouette #0
			LInfo.FUploadSilhouetteTargetBlock = 0; // silhouette #0
			LInfo.FUploadSilhouetteTargetRow = 0;

			LProgressMax += 66;
		}

		LInfo.FProgressMax = LProgressMax;
        ProgressBarSetup->Max = LProgressMax;

		// start the upload
		LDevice->ProcessAsyncUploadTargets();

		// copy device use scenario to shooter
		LShooter->FUseScenario = LDevice->GetUserScenarioSet();
	}

	// start time
	GSession->FDateTime = FormatDateTime("ddddd t", Now());

}
//---------------------------------------------------------------------------

void __fastcall TFormSelectSetup::ButtonSetupClick(TObject *Sender)
{
	// pick the selected target definition
	TTargetDefinition& LT = FCurrentSetup->FTargetDef;

	if (! (LT.FUsingRing || LT.FUsingSilhouette) )
	{
		Application->MessageBox(WGetTranslation("Ring or silhouette not selected").c_str(),WGetTranslation("Error").c_str(),MB_OK);
		return;
	}

    Beep();

	TDeviceManager* LDevMan = GetDeviceManager();

	// now we must have devices connected
	if (LDevMan->GetDeviceCount() == 0)
	{
		Application->MessageBox(WGetTranslation("No device connected. Shooting not possible.").c_str(),WGetTranslation("Insert Optical Units").c_str(), MB_OK);
		return;
	}

	PanelShowProgress->Show();
	TimerProgressBar->Enabled = true;
	
    // hour glass
	AddBusyCount();
	DisableUI();

	// we wait until all the devices have UploadTargetsDone
	FDeviceCountAtStart = LDevMan->GetDeviceCount();

	// "set" going
	FSetOrStartGoing = 1;

	// enable timeout timer
	TimerProgrammingPoller->Enabled = true;

	// MHi 1.7.2013 Bug fix tryout: force here to classic by default
	// was always dynamic when using the following line
	//bool LIsDynamic = CheckBoxModeDynamic->Checked;
	// ... so, we try this instead:
	//bool LIsDynamic = false;

	// MHi 11.12.2013 Prev change was not ok - was classic in stand-alone with silhouettes too!
	// we'll go back to how it was
	bool LIsDynamic = CheckBoxModeDynamic->Checked;

	// 19.11.2014. Update classic/dynamic selection
	UpdateDynamicClassicSelection();

	LIsDynamic = (GSelectedMode == KDynamicTraining);

	// for each created shooter...
	for (int i = 0; i < LDevMan->GetDeviceCount(); i++)
	{
		// device should be without shooter by default (?)
		TDevice* LDevice = LDevMan->GetDevice(i);

		// open COM port if we had downloaded stuffs and closed port after it
		if (LDevice->FCOMPort < 0)
		{
			LDevice->OpenComPort(-LDevice->FCOMPort, true);
			Sleep(1000);
		}

		// set startup mode accordingly
		LDevice->GetUserScenarioSet().STARTUP_FUNCTION_MODE = LIsDynamic ? 2 : 1;

		// TRANSLATE IMAGE ORDER TO GUN MOUNTING
		int LMount = 0;
		if (SpeedButtonMounting->Tag == 0) LMount = 0;
		if (SpeedButtonMounting->Tag == 1) LMount = 1;
		if (SpeedButtonMounting->Tag == 2) LMount = 3;
		if (SpeedButtonMounting->Tag == 3) LMount = 2;

		// Apply selected gun mounting.
		LDevice->GetUserScenarioSet().DeviceMounting = LMount;
    	//LDevice->GetUserScenarioSet().DeviceMounting = FCurrentSetup->FUseScenario.DeviceMounting;

		// update classic or dynamic mode settings?
		TModeSettings& LMode = LIsDynamic ? LDevice->GetUserScenarioSet().DYNAMIC_MODE_SETTINGS : LDevice->GetUserScenarioSet().CLASSIC_MODE_SETTINGS;

		// update mode according to which mode was selected on first screen
		LMode = LIsDynamic ? FCurrentSetup->FUseScenario.DYNAMIC_MODE_SETTINGS : FCurrentSetup->FUseScenario.CLASSIC_MODE_SETTINGS;
		// arrows are not of use when connected anyway...?
		LDevice->GetUserScenarioSet().Arrow_buttons = FCurrentSetup->FUseScenario.Arrow_buttons;

		// prepare async upload information
		TDeviceAsyncDownloadInfo& LInfo = LDevice->FAsyncInfo;
        memset(&LInfo, 0, sizeof(LInfo));
		LInfo.FUploadTargetsGoing = true;
		LInfo.FUploadTargetsDone = false;
   		LInfo.FUploadUserScenarioAndErase = true;

		// start stop programming, send user scenario
		int LProgressMax = 14;
		LInfo.FProgressNow = 0;

		if (LT.FUsingRing)
		{
			LDevice->FAllTargets.SetRingTarget(LMode.RING_TARGET_NUMBER, LT.FRingTarget);

			// update desired mode to use this set
			LMode.CountRingScore = true;

			LInfo.FUploadRingTarget = true;
			LInfo.FUploadRingTargetIndex = LMode.RING_TARGET_NUMBER;
			LInfo.FUploadRingTargetBlock = 0;

			LProgressMax += 4;
		}

		if (LT.FUsingSilhouette)
		{
			LDevice->FAllTargets.SetSilhouetteTarget(LMode.SILH_TARGET_NUMBER, LT.FSilTarget);
			LDevice->FAllTargets.SetSilhouetteBitmapMatrix(LMode.SILH_TARGET_NUMBER, LT.FSilhouette);

			LMode.EvaluateSilhouette = true;

			LInfo.FUploadSilhouetteTarget = true;
			LInfo.FUploadSilhouetteTargetIndex = LMode.SILH_TARGET_NUMBER; // silhouette #0
			LInfo.FUploadSilhouetteTargetBlock = 0; // silhouette #0
			LInfo.FUploadSilhouetteTargetRow = 0;

			LProgressMax += 66;
		}

		// set maximum number of operations as "progress max"
		LInfo.FProgressMax = LProgressMax;
		ProgressBarSetup->Max = LProgressMax;

		// start the upload
		LDevice->ProcessAsyncUploadTargets();
	}

}
//---------------------------------------------------------------------------

// 19.11.2014. Update classic dynamic selection upon target
void TFormSelectSetup::UpdateDynamicClassicSelection()
{

	// TODO jos ei ole valittu mitn...
	if (FCurrentIndex < 0) return;

	if (!FCurrentSetup->FTargetDef.FUsingRing)
	{
		GSelectedMode = KDynamicTraining;
	}
	else
	{
		GSelectedMode = KClassicTraining;
	}
}

void __fastcall TFormSelectSetup::BitBtnEditClick(TObject *Sender)
{
	// TODO jos ei ole valittu mitn...
	if (FCurrentIndex < 0) return;

	// 19.11.2014. Update classic dynamic selection upon target
	UpdateDynamicClassicSelection();

	// TODO Just one fsetup, no need for FInEdit?
	FormEditSetup->FSetup.CopyContentsFrom(*FCurrentSetup);
	FormEditSetup->UpdateViewsFromSetup();

	// set to update setup directly in handler, not "CurrentSetup"
	FormEditSetup->FWhereToApply = &FSetupHandler.GetSetup(FCurrentIndex);

	if (FormEditSetup->ShowModal() == mrOk)
	{
        // Setup is updated when OK is clicked on edit form (using pointer to FCurrentSetup)
		DrawGridSetups->Invalidate();
		DrawGridSetupsClick(NULL);
	}
}
//---------------------------------------------------------------------------

void TFormSelectSetup::DeleteSelected()
{
	if (FCurrentIndex < 0 || FSetupHandler.FSetupCount < 1)
		return;

	// delete selected index
	FSetupHandler.DeleteAtIndex(FCurrentIndex);
    // update list
	DrawGridSetups->RowCount = FSetupHandler.FSetupCount;
	DrawGridSetups->Invalidate();
	DrawGridSetupsClick(NULL);
}

void TFormSelectSetup::MoveSelectedToTop()
{
	// "bubble swap" selected to top of list
	for (int i = FCurrentIndex-1; i >= 0; i--)
		FSetupHandler.SwapAtIndexes(i, i+1);

	FCurrentIndex = 0;
	DrawGridSetups->Row = 0;
	// update list
	DrawGridSetups->RowCount = FSetupHandler.FSetupCount;
	DrawGridSetups->Invalidate();
	DrawGridSetupsClick(NULL);
}

void TFormSelectSetup::AddToList(TTargetSetup& ASetup)
{
	FSetupHandler.AddSetup(ASetup);
	DrawGridSetups->RowCount = FSetupHandler.FSetupCount;
	DrawGridSetups->Row = DrawGridSetups->RowCount - 1;
	DrawGridSetups->Invalidate();

	// update current selection
	DrawGridSetupsClick(NULL);
}

void __fastcall TFormSelectSetup::DrawGridSetupsClick(TObject *Sender)
{
	FCurrentIndex = DrawGridSetups->Row;
	FCurrentSetup = &FSetupHandler.GetSetup(FCurrentIndex);
}
//---------------------------------------------------------------------------

void __fastcall TFormSelectSetup::BitBtnCloseClick(TObject *Sender)
{
	// show main menu again
	Hide();

	FormStartMenu->Show();	
}
//---------------------------------------------------------------------------

void __fastcall TFormSelectSetup::CheckBoxSortListClick(TObject *Sender)
{
	SortSetups(CheckBoxSortList->Checked);
	DrawGridSetups->Row = 0;
	DrawGridSetups->Invalidate();
}
//---------------------------------------------------------------------------

void __fastcall TFormSelectSetup::FormShow(TObject *Sender)
{
	// sort setups according to selected mode (classic=rings first, dynamic=sils first)
	SortSetups(GSelectedMode == KClassicTraining);

	// show classic/dynamic mode also in caption
	WideString LTxt = WGetTranslation("Training Setup");

	//LTxt += ": ";
	//LTxt += WGetTranslation(GSelectedMode == KClassicTraining ? "Classic Operation" : "Dynamic Operation");

	Caption = LTxt;

	// restore gun mounting and update picture
	UnicodeString LMountStr = GetAppSettingsOptionValueStr(CAppSettingsGunMounting);
	int LMount = 0;
	if (LMountStr == CMountingTop) LMount = 0;
	if (LMountStr == CMountingRight) LMount = 1;
	if (LMountStr == CMountingLeft) LMount = 3;
	if (LMountStr == CMountingBelow) LMount = 2;
	SpeedButtonMounting->Tag = LMount;

	UpdateMountingPicture();
}
//---------------------------------------------------------------------------

void TFormSelectSetup::EnableUI()
{
	ButtonSetup->Enabled = true;
	ButtonStart->Enabled = true;
	SpeedButtonMounting->Visible = true;
	BitBtnEdit->Enabled = true;
	ButtonOpenSetupFile->Enabled = true;

	// We keep close button available even when others are locked
	BitBtnClose->Enabled = true;

}

void TFormSelectSetup::DisableUI()
{
	ButtonSetup->Enabled = false;
	ButtonStart->Enabled = false;
	SpeedButtonMounting->Visible = false;
	// We keep close button available even when others are locked
	//BitBtnEdit->Enabled = false;
	//ButtonOpenSetupFile->Enabled = false;
	//BitBtnClose->Enabled = false;
}

void __fastcall TFormSelectSetup::TimerProgressBarTimer(TObject *Sender)
{
	TDeviceManager* LDM = GetDeviceManager();
	int LPro = ProgressBarSetup->Max;
	int LCount = LDM->GetDeviceCount();
	for (int i = 0; i < LCount; i++)
	{
		LPro = Min(LDM->GetDevice(i)->FAsyncInfo.FProgressNow, LPro);
	}

	ProgressBarSetup->Position = LPro;
	ProgressBarSetup->Update();
	PanelShowProgress->Update();
}
//---------------------------------------------------------------------------

void __fastcall TFormSelectSetup::TimerProgrammingPollerTimer(TObject *Sender)
{
	bool LContWait/* = true*/;
	bool LError = false;

	TDeviceManager* LDevMan = GetDeviceManager();

	LContWait = false;
	for (int i = 0; i < LDevMan->GetDeviceCount(); i++)
	{
		LContWait |= !LDevMan->GetDevice(i)->FAsyncInfo.FUploadTargetsDone;
		LError |= LDevMan->GetDevice(i)->FAsyncInfo.FAsyncError; //26.1.2017 MHi ToDo: this line causes frequent errors when starting the session
		LError |= LDevMan->GetDevice(i)->CheckAsyncTimeout();
	}

	// if a device leaves during programming, something is "failed" and need retry
	if (LDevMan->GetDeviceCount() != FDeviceCountAtStart) LError = true;

	// if things are going, we just wait
	if (LContWait && !LError) return;

	//Beep();

	// unlock buttons
	DecBusyCount();
	EnableUI();

	// we won't come here any more
	TimerProgrammingPoller->Enabled = false;

	// no longer "setting up devices"
	PanelShowProgress->Hide();
	TimerProgressBar->Enabled = false;

	if (LError)
	{
		FSetOrStartGoing = 0;
		WideString LStr = WGetTranslation("Error during programming");
		WideString LStrCap = WGetTranslation("Error");

		Application->MessageBox(LStr.c_bstr(), LStrCap.c_bstr(), MB_OK);

		// lets put this here so that if we meet an error user was wait on message box
		GetDeviceManager()->StopAllAsync();

		return;
	}

	// if we were on set, we are done now. if start, we create shooter forms and go on
	if (FSetOrStartGoing == 1)
	{
		FSetOrStartGoing = 0;
		return;
	}

	FSetOrStartGoing = 0;

	// hide this form
	Hide();

	bool LIsDynamic = CheckBoxModeDynamic->Checked;

	// PKy. Ring Target connected = classic mode
	LIsDynamic = !LDevMan->GetDevice(0)->FShooter->FTargetDefinition.FUsingRing;

	// TODO Refactor
	bool LStartAutomatically = (GAppSettings->GetNode(CAppSettingsOptionsNode) && (GAppSettings->GetNode(CAppSettingsOptionsNode)->GetNodeText(CAppSettingsOptionsStartAutomatically).LowerCase() == "true") );

	// for each created shooter...
	for (int i = 0; i < LDevMan->GetDeviceCount(); i++)
	{
		// device should be without shooter by default (?)
		TDevice* LDevice = LDevMan->GetDevice(i);

		// just anyone form owns it
		TFormShooter* LForm = new TFormShooter(NULL);
		GFormShooters->Add(LForm);

		// classic=1, dynamic=2
		LDevice->FMeasuringMode = LIsDynamic ? 2 : 1;
		LForm->FStartupMode = LDevice->FMeasuringMode;

		// realtime shooting i.e. don't copy device targets
		LForm->UseDeviceShooter(LDevice, LDevice->FShooter, true);
		// show shooter window and render target
		LForm->Show();
		LForm->FTargetRenderer.DrawTarget();

		if (LStartAutomatically)
		{
			// todo refactor?
			LForm->ButtonStartStopClick(NULL);
		}
	}

	// make shooter panel to organize windows
	FormShooterPanel->ShowAllShooters();

	Beep();

	if (GetNetworkMode() == CAppSettingsNetworkNodeModeNode)
	{
		FormNetworkControl->FClientSocket->Host = GAppSettings->GetNode(CAppSettingsNetworkNode)->GetAttributeValue(CAppSettingsNetworkNodeIP);
		FormNetworkControl->FClientSocket->Port = GAppSettings->GetNode(CAppSettingsNetworkNode)->GetAttributeValueDef(CAppSettingsNetworkNodeMonitorPort,14351);

		FormNetworkControl->FClientSocket->Active = true;

		// now we just need to wait and see whether we get ready.
	}
}
//---------------------------------------------------------------------------

void __fastcall TFormSelectSetup::BitBtnMoreClick(TObject *Sender)
{

	bool LNewVisiblity = !ButtonOpenSetupFile->Visible;

	// advanced use - view all setup list edit buttons
	if(IsAdvancedSettingsEnabled() == true)
	{
		BitBtnCopy->Visible = LNewVisiblity;
		BitBtnDelete->Visible = LNewVisiblity;

		BitBtnMoveUp->Visible = LNewVisiblity;
		BitBtnMoveDown->Visible = LNewVisiblity;

		ButtonOpenSetupFile->Left = 311;
		ButtonOpenSetupFile->Visible = LNewVisiblity;

		ButtonSaveSetupFile->Visible = LNewVisiblity;

		ButtonOpenDefaultSetupFile->Left = 365;
		ButtonOpenDefaultSetupFile->Visible = LNewVisiblity;
	}
	else // normal use - view open/default buttons only
	{
		ButtonOpenSetupFile->Left = 95;
		ButtonOpenSetupFile->Visible = LNewVisiblity;

		ButtonOpenDefaultSetupFile->Left = 149;
		ButtonOpenDefaultSetupFile->Visible = LNewVisiblity;
	}
}
//---------------------------------------------------------------------------

void __fastcall TFormSelectSetup::BitBtnCopyClick(TObject *Sender)
{
	// TODO jos ei ole valittu mitn...
	if (FCurrentIndex < 0) return;

	// copy target...
	TTargetSetup LSetup;
	LSetup.CopyContentsFrom(*FCurrentSetup);

	if (LSetup.FDescription->Count > 0)
	{
		UnicodeString LString = LSetup.FDescription->Strings[0];
		LString = "Copy of " + LString;
		LSetup.FDescription->Strings[0] = LString;
	}

	AddToList(LSetup);

	// Setup is updated when OK is clicked on edit form (using pointer to FCurrentSetup)
	DrawGridSetups->Row = DrawGridSetups->RowCount - 1;
	DrawGridSetups->Invalidate();
	ButtonSaveSetupFile->Enabled = true;
}
//---------------------------------------------------------------------------

void __fastcall TFormSelectSetup::BitBtnDeleteClick(TObject *Sender)
{
	DeleteSelected();
	ButtonSaveSetupFile->Enabled = true;
}
//---------------------------------------------------------------------------


void __fastcall TFormSelectSetup::BitBtnMoveUpClick(TObject *Sender)
{
	if (DrawGridSetups->Row > 0)
	{
		FSetupHandler.SwapAtIndexes(DrawGridSetups->Row, DrawGridSetups->Row - 1);

		DrawGridSetups->Row = DrawGridSetups->Row - 1;
		DrawGridSetups->Invalidate();
		ButtonSaveSetupFile->Enabled = true;
	}
}
//---------------------------------------------------------------------------

void __fastcall TFormSelectSetup::BitBtnMoveDownClick(TObject *Sender)
{
	if (DrawGridSetups->Row < FSetupHandler.FSetupCount - 1)
	{
		FSetupHandler.SwapAtIndexes(DrawGridSetups->Row, DrawGridSetups->Row + 1);

		DrawGridSetups->Row = DrawGridSetups->Row + 1;
		DrawGridSetups->Invalidate();
		ButtonSaveSetupFile->Enabled = true;
	}
}
//---------------------------------------------------------------------------


void __fastcall TFormSelectSetup::ButtonSaveSetupFileClick(TObject *Sender)
{
	UnicodeString LErrorMessage = WGetTranslation("Default setup cannot be overwritten");

	if(FormEditSetup->SaveDialogSave->Execute() )
	{
		UnicodeString LFilename = FormEditSetup->SaveDialogSave->FileName;

		if ( ExtractFileName(LFilename).LowerCase() == "default.setup.xml")
		{
			Application->MessageBox( LErrorMessage.c_str(), WGetTranslation("Error").c_str(),MB_OK);
			return;
		}
		else
		{
			FSetupHandler.SaveToFile(LFilename);
			Close();
        }
    }

}
//---------------------------------------------------------------------------

void __fastcall TFormSelectSetup::SpeedButtonMountingClick(TObject *Sender)
{
	int i = SpeedButtonMounting->Tag;
	i++;
	if (i == 4) i = 0;

	SpeedButtonMounting->Tag = i;

	UpdateMountingPicture();

}

void TFormSelectSetup::UpdateMountingPicture()
{
	int i = SpeedButtonMounting->Tag;
	if (i == 0) SpeedButtonMounting->Glyph = ImageMounting1->Picture->Bitmap;
	if (i == 1) SpeedButtonMounting->Glyph = ImageMounting2->Picture->Bitmap;
	if (i == 2) SpeedButtonMounting->Glyph = ImageMounting3->Picture->Bitmap;
	if (i == 3) SpeedButtonMounting->Glyph = ImageMounting4->Picture->Bitmap;

	if (i == 0) SetAppSettingsOptionValueStr(CAppSettingsGunMounting, CMountingTop);
	if (i == 1) SetAppSettingsOptionValueStr(CAppSettingsGunMounting, CMountingRight);
	if (i == 2) SetAppSettingsOptionValueStr(CAppSettingsGunMounting, CMountingBelow);
	if (i == 3) SetAppSettingsOptionValueStr(CAppSettingsGunMounting, CMountingLeft);

}

//---------------------------------------------------------------------------



